/**************************************************************************
* The finite height tree automata abstraction based on MONA.
* Tomas Vojnar
**************************************************************************/

#include <stdlib.h>
#include <stdio.h>

#include "gta.h"
#include "../BDD/bdd_external.h"
#include "../BDD/bdd.h"
#include "../Mem/mem.h"

#include "modify_shape_gta.h"
#include "collaps2.h"
#include "finite_up_abstr.h"
#include "initial_partition.h"

#define SS1             1

// !!! These values are a pure guess...

#define MIN_BDDs_CACHE		        100
#define INCR_IN_BDDs_CACHEs             50
#define SPARSE_PLANE_SIZE_GUESS         3

/**************************************************************************
* The extern variables 
**************************************************************************/

extern Guide guide;

/**************************************************************************
* Sparse planes of the accummulated bdd cubes
* -- based on dynamically resizable arrays inspired by Mona sequential lists
**************************************************************************/

struct s_s_cls_pred {
        unsigned cL;  // a left class of predecessor states of a given state q
        unsigned cR;  // a right class of predecessor states of q
        union u_acc_bdd {
          bdd_handle hnd_acc_bdd; // via which input symbols one can get to q
          bdd_ptr ptr_acc_bdd; //  ptr derived from hnd once all bdd ops are finished
        } acc_bdd;
};

typedef struct s_s_cls_pred s_cls_pred;

// can be used for sorting arrays of s_cls_pred with hnd_acc_bdd already replaced by ptr_acc_bdd
int compare_s_cls_pred(const void *av, const void *bv) {
        s_cls_pred *a = (s_cls_pred *) av;
        s_cls_pred *b = (s_cls_pred *) bv;
        
        if (a->cL < b->cL) return -1;
        if (a->cL > b->cL) return 1;
        if (a->cR < b->cR) return -1;
        if (a->cR > b->cR) return 1;
        if (a->acc_bdd.ptr_acc_bdd < b->acc_bdd.ptr_acc_bdd) return -1;
        if (a->acc_bdd.ptr_acc_bdd > b->acc_bdd.ptr_acc_bdd) return 1;
        return 0;
}

struct s_s_cls_pred_array {
        s_cls_pred *array;
        unsigned length;
        unsigned index;
};

typedef struct s_s_cls_pred_array s_cls_pred_array;

void init_cls_pred_array(s_cls_pred_array *arr,unsigned size) {
        arr->array = (s_cls_pred *) mem_alloc((size_t)size * sizeof(s_cls_pred));
        arr->length = size;
        arr->index = 0;
}
        
void free_cls_pred_array(s_cls_pred_array *arr) {
        mem_free(arr->array);
        arr->length = 0;
        arr->index = 0;
}

// we add an accummulated bdd handle, which we at the end change to a bdd ptr
void add_to_cls_pred_array(s_cls_pred_array *arr,unsigned cL,unsigned cR,bdd_handle hnd) {
        if (!(arr->index < arr->length - 1)) {
		arr->length = arr->length*2;
                arr->array = (s_cls_pred *) mem_resize(arr->array,(size_t)((arr->length) * (sizeof (s_cls_pred))));
        }
        arr->array[arr->index].cL = cL;
        arr->array[arr->index].cR = cR;
        arr->array[arr->index].acc_bdd.hnd_acc_bdd = hnd;
        arr->index = arr->index+1;
}

// the function replaces bdd handles by pointers and sorts the array
void transf_and_sort_cls_pred_array(s_cls_pred_array *arr,bdd_manager *bdd_mng) {
        for (unsigned i=0;i < arr->index;i++)
                arr->array[i].acc_bdd.ptr_acc_bdd = BDD_ROOT(bdd_mng,arr->array[i].acc_bdd.hnd_acc_bdd);       
        if (arr->index) qsort(arr->array,(size_t) arr->index,sizeof(s_cls_pred),&compare_s_cls_pred);
}

// can be used for sorting arrays of arrays of s_cls_pred with hnd_acc_bdd already replaced by ptr_acc_bdd
int compare_s_cls_pred_array(s_cls_pred_array *a,s_cls_pred_array *b) {
	int elm_comp;

	if (a->index != b->index) return ((a->index)<(b->index))?-1:1;
	for (unsigned i=0;i<(a->index);i++)
		if ((elm_comp=compare_s_cls_pred((a->array)+i,(b->array)+i))!=0) return elm_comp;
	return 0;
}

/**************************************************************************
* Sorting the sparse bdd cube while tracking what changes are being done
* so that one can realize which states stay in the same class and which not.
**************************************************************************/

// the array st_order says which state is currently at the i-th position
int compare_states_in_bdd_cube(unsigned a,unsigned b,unsigned *classes,s_cls_pred_array *bdd_cls_cube,unsigned *st_order) {
	unsigned qa = st_order[a];
	unsigned qb = st_order[b];

	if (classes[qa]!=classes[qb]) return (classes[qa]<classes[qb])?-1:1;
	return compare_s_cls_pred_array((bdd_cls_cube+qa),(bdd_cls_cube+qb));
}

void swap_states_in_bdd_cube(unsigned a,unsigned b,unsigned *st_order) {
	unsigned qa = st_order[a];

	st_order[a] = st_order[b];
	st_order[b] = qa;
}

void quicksort_bdd_cube(unsigned from,unsigned to,unsigned *classes,s_cls_pred_array *bdd_cls_cube,unsigned *st_order) {
  if (from < to) {
    unsigned i = from, j = to, v = to;
    do {
      while (i < to && compare_states_in_bdd_cube(i,v,classes,bdd_cls_cube,st_order) < 0)
	i++;
      while (j > from && compare_states_in_bdd_cube(j,v,classes,bdd_cls_cube,st_order) > 0)
	j--;
      if (i <= j) {
	swap_states_in_bdd_cube(i,j,st_order);
	if (v == i)
	  v = j;
	else if (v == j)
	  v = i;
	i++;
	j--;
      }
    } while (i <= j);
    quicksort_bdd_cube(from,j,classes,bdd_cls_cube,st_order);
    quicksort_bdd_cube(i,to,classes,bdd_cls_cube,st_order);
  }
}

unsigned *sort_bdd_cube(unsigned num_states,unsigned *classes,s_cls_pred_array *bdd_cls_cube) {
	unsigned *st_order;

	st_order = (unsigned *) mem_alloc(sizeof(unsigned)*num_states);
	for (unsigned i=0;i<num_states;i++) st_order[i]=i;
	quicksort_bdd_cube(0,num_states-1,classes,bdd_cls_cube,st_order); 
	return st_order;
}


/**************************************************************************
* The finite height abstraction -- we mean wrt. L^{\leq k}(M,q)...
**************************************************************************/

unsigned cube_target_state;  // the state which we consider the target of another step when computing the finite height abstraction

// we mark the designated target with 1, the rest with 0
unsigned mark_cube_target(unsigned q) {
  return (q == cube_target_state) ? 1 : 0;
}

// unite two bdds in the bdd cube
unsigned bbd_cube_entry_union(unsigned a, unsigned b) {
  return (a || b) ? 1 : 0;
}

GTA *abstract_fin_height(GTA *g1,int height) {
        GTA *g, *g2;
        unsigned num_states;
        unsigned *classes, *st_order;
        unsigned num_classes; // i.e. the number of equivalence classes
        s_cls_pred_array *bdd_cls_cube; // the sparse cube of accummulated BDDs Q_LEFT_CLS x RIGHT_CLS --> BDD saying
                                        // by which letters one can get from a left and a right predecessor class of states to a state
        bdd_manager *bdd_cube_mng=NULL;
        unsigned bdd_cube_mng_tab_size;

	// printf(">>> We are at the beginning of abs_fin_h!\n");

	g=gtaMinimize(g1);
	num_states=g->ss[SS1].size;

	// printf(">>> The number of states: %d\n",num_states);

        if (guide.numSs!=3) {
	        // printf("ERROR: The finite height abstraction works for 3 statespaces only!\n");
	        exit(0);
        }

        // do the initial partition of s.s. 1 (the main one)
        classes=initial_partition_of_ss1(g);
        num_classes=MEMN_CLS+1;

	// printf(">>> We are after the initial partitioning!\n");

        // refine the initial classes ``height'' times
        for (unsigned refnm_no=0; refnm_no<height; refnm_no++) { // we do the loop as many times as the height to consider is

		// printf(">>> We begin iteration %d of the refinement loop!\n",refnm_no);
		// printf(">>> Creating the bdd_manager!\n");
		
                // initialize the bdd_manager for the cube
	        if (bdd_cube_mng!=NULL) {
                        bdd_cube_mng_tab_size=bdd_cube_mng->table_elements;
                        bdd_kill_manager(bdd_cube_mng); // !!! I hope this releases all the memory allocated for the BDDs
                } else bdd_cube_mng_tab_size=g->ss[SS1].bddm->table_elements;
                bdd_cube_mng = bdd_new_manager(bdd_cube_mng_tab_size,INCR_IN_BDDs_CACHEs);
                  // !!! the parameters above are a pure guess -- are they efficient? ask authors of Mona...
                
		// printf(">>> Creating the bdd_cache -- size, incr: %d, %d!\n",bdd_cube_mng->table_elements,INCR_IN_BDDs_CACHEs);

                // prepare the cache for uniting bdds in the cube
		bdd_kill_cache(bdd_cube_mng);
		if ((bdd_cube_mng->table_elements) > MIN_BDDs_CACHE)
                	bdd_make_cache(bdd_cube_mng,bdd_cube_mng->table_elements,INCR_IN_BDDs_CACHEs);
		else 
                	bdd_make_cache(bdd_cube_mng,MIN_BDDs_CACHE,INCR_IN_BDDs_CACHEs);
                  	  // !!! the parameters above are a pure guess -- are they efficient? ask authors of Mona...
                  	  // !!! can the cache be used all the time in all the below iterations?
			
		// printf(">>> Creating the bdd cube!\n");

                // initialize the cube of accummulated bdds
                bdd_cls_cube = (s_cls_pred_array *) mem_alloc(sizeof(s_cls_pred_array)*num_states);

                for (cube_target_state=0; cube_target_state<num_states; cube_target_state++) { // for every target state
                  bdd_handle hnd_for_q_qL_qR, hnd_for_q_cL_cR;
                  bdd_ptr ptr_for_q_qL_qR, ptr_for_q_cL_cR;
                  char no_cL_cR_pred_found;

		  // printf(">>>>>>> In iteration %d computing the plane for state %d!\n",refnm_no,cube_target_state);
                  
                  // prepare the sparse plane of the cube for q implemented as a list of s_cls_pred
                  init_cls_pred_array((bdd_cls_cube+cube_target_state),SPARSE_PLANE_SIZE_GUESS);
                  
                  for (unsigned cL=0; cL<num_classes; cL++) // for every left predecessor class
                    for (unsigned cR=0; cR<num_classes; cR++) { // for every right predecessor class
                      no_cL_cR_pred_found=1;
                            
                      for (unsigned qL=0; qL<num_states; qL++) // for every left predecessor
                        if (classes[qL]==cL)  
                          for (unsigned qR=0; qR<num_states; qR++) // for every right predecessor
                            if (classes[qR]==cR) {
	                        
				// printf(">>> Before bdd_prepare_apply1\n");
				
		  		bdd_prepare_apply1(g->ss[SS1].bddm);  
                    		  // !!! is this necessary? mark_cube_target stays the same, but cube_target_state changes...
	                    	  // ... ask the Mona guys...
				
		  		// printf(">>> before bdd_apply1 for q, cL(qL), cR(qR): %d, %d(%d), %d(%d)!\n",cube_target_state,cL,qL,cR,qR);
				
				ptr_for_q_qL_qR=bdd_apply1(g->ss[SS1].bddm, 
			                          BDD_ROOT(g->ss[SS1].bddm,BEH(g->ss[SS1],qL,qR)),
			                          bdd_cube_mng,
			                          &mark_cube_target);

		  		// printf(">>> after bdd_apply1!\n");

                                if (bdd_is_leaf(bdd_cube_mng,ptr_for_q_qL_qR))
                                  if (bdd_leaf_value(bdd_cube_mng,ptr_for_q_qL_qR)==0) { // no transition from qL,qR to q -> delete the result
                                        POP_SEQUENTIAL_LIST(bdd_cube_mng->roots,bdd_handle,hnd_for_q_qL_qR);
                                        continue;
                                  }

                                if (no_cL_cR_pred_found) // we have not found any path from cL,cR to q
                                        no_cL_cR_pred_found=0; // the last root stays in the bdd_cube_mng as the accummulated bdd root
				else {
                                        POP_SEQUENTIAL_LIST(bdd_cube_mng->roots,bdd_handle,hnd_for_q_qL_qR); // we will unite it with the rest
					hnd_for_q_cL_cR=BDD_LAST_HANDLE(bdd_cube_mng);
					ptr_for_q_cL_cR=BDD_ROOT(bdd_cube_mng,hnd_for_q_cL_cR);

					if (ptr_for_q_qL_qR!=ptr_for_q_cL_cR) {
                                          POP_SEQUENTIAL_LIST(bdd_cube_mng->roots,bdd_handle,hnd_for_q_cL_cR); // we will unite it with the above 
					
		  			  // printf("~~~ before bdd_apply2 -- q_ql_qR, q_cL_cR: %d, %d!\n",ptr_for_q_qL_qR,ptr_for_q_cL_cR);

                                          bdd_apply2_hashed(bdd_cube_mng,ptr_for_q_qL_qR,
                                                            bdd_cube_mng,ptr_for_q_cL_cR,
                                                            bdd_cube_mng,&bbd_cube_entry_union);

		  			  // printf("~~~ after bdd_apply2!\n");
					}
                                } 

                          } // end of for every right predecessor (of course via the if)

                      if (!no_cL_cR_pred_found)
                        add_to_cls_pred_array((bdd_cls_cube+cube_target_state),cL,cR,BDD_LAST_HANDLE(bdd_cube_mng));

                    } // end of for every right predecessor class
                    
                } // for every target state

		// printf("++++++++ Cube has been computed!\n");

		// we have computed the cube, now we have to sort its planes and then the queue itself
                for (cube_target_state=0; cube_target_state<num_states; cube_target_state++) // for every target state
			transf_and_sort_cls_pred_array((bdd_cls_cube+cube_target_state),bdd_cube_mng);
		st_order=sort_bdd_cube(num_states,classes,bdd_cls_cube);

		// printf("++++++++ Cube has been sorted!\n");
		
		// we now refine the classes
		num_classes = 0;
		for (unsigned q = 0; q < num_states-1; q++)
		  if (compare_states_in_bdd_cube(q,q+1,classes,bdd_cls_cube,st_order) != 0) // the value can only be -1 or 0 after the above sorting
		    classes[st_order[q]] = num_classes++;
		  else
	    	    classes[st_order[q]] = num_classes;
	    	classes[st_order[num_states-1]] = num_classes++;

		// printf("++++++++ Classes have been refined!\n");

		// we don't need the cube any more -- we clean it up
                for (cube_target_state=0; cube_target_state<num_states; cube_target_state++) // for every target state
			free_cls_pred_array(bdd_cls_cube+cube_target_state);
                mem_free(bdd_cls_cube);
                mem_free(st_order);

        } // for all the considered heights

	// printf("######## The partition has been computed -- we are about to collaps the states!\n");

        // create the resulting automaton by collapsing the appropriate states
        g2 = colaps2(g,(int *)classes); // !!! why the classes are defined via the signed int?

	// printf("######## States have been collapsed!\n");

        // cleaning up all the remaining mess
        if (bdd_cube_mng) bdd_kill_manager(bdd_cube_mng); // !!! I hope this releases all the memory allocated for the BDDs
        mem_free(classes);
	gtaFree(g);
        
	// printf("######## We are after the final cleaning up -- we are finished!\n");

        return g2; 
}
