/* Adam Rogalewicz
 *
 * This file contains two possible abstractions:
 * neighbour_abstraction
 *    - collapsed can be just memory nodes, which are neighbours
 *
 * finite height abstraction with respect to 'pointer structure'
 *    - collapsed can be just memory nodes
 * */


#include "neighbour-abstract.h"
#include "modify_shape_gta.h"
#include "../BDD/bdd_external.h"
#include "../BDD/bdd.h"
#include "../Mem/mem.h"

#include "initial_partition.h"

#include "main.h"

#define COMPARE_BY_CLASSES

/* Gloval variables of this module */
reachability_cache *rcache;
bdd_manager *rcache_bdd_mng;
unsigned *na_classes;
unsigned target_state;

bdd_ptr get_ptr(bdd_handle handle) {
	return BDD_ROOT(rcache_bdd_mng,handle);
}

/* this function compare two elements of the type cache_type by values of 'bddlink.ptr' */
int compare_cache_type(const void *av, const void *bv) {
	cache_type *a = (cache_type *) av;
        cache_type *b = (cache_type *) bv;
	/* compare by bdd pointers */
	if (get_ptr(a->handle) > get_ptr(b->handle)) return 1;
	if (get_ptr(a->handle) < get_ptr(b->handle)) return -1;
	/* compare by classes */
#ifdef COMPARE_BY_CLASSES
	if (na_classes[a->right] > na_classes[b->right]) return 1;
	if (na_classes[a->right] < na_classes[b->right]) return -1;
	if (na_classes[a->left] > na_classes[b->left]) return 1;
	if (na_classes[a->left] < na_classes[b->left]) return -1;
#endif
	return 0;
}

// the function replaces bdd handles by pointers and sorts the array by bdd_pointers values
void sort_cache_by_ptrs(unsigned size) {
	unsigned i;
	int j; /* rcache[i].no can be -1 !!! */
        for (i=0;i < size;i++) {
		/* sort rcache[i].from by qsort */
		if (rcache[i].no>1) /* not necessary to call qsort for 0 or 1 element */
			qsort(rcache[i].from, (size_t)rcache[i].no, sizeof(cache_type), &compare_cache_type);
	}
}


/* init the reachability chache */
void init_cache(GTA *G) {
	unsigned tabsize,states,i;
	states=G->ss[SS1].size;
	rcache=(reachability_cache *) mem_alloc (sizeof(reachability_cache)*states);
	/* set -1 for all automata states - state not processed */
	for (i=0;i<states;i++) { rcache[i].no=-1; rcache[i].alocated=0; }

	tabsize=G->ss[SS1].bddm->table_elements;
        rcache_bdd_mng = bdd_new_manager(tabsize,INCR_IN_BDDs_CACHEs);
        // !!! the parameters above are a pure guess -- are they efficient? ask authors of Mona...
	// prepare the cache for uniting bdds in the cube
/*	bdd_kill_cache(rcache_bdd_mng);
	if ((rcache_bdd_mng->table_elements) > MIN_BDDs_CACHE)
               	bdd_make_cache(rcache_bdd_mng,rcache_bdd_mng->table_elements,INCR_IN_BDDs_CACHEs);
	else 
               	bdd_make_cache(rcache_bdd_mng,MIN_BDDs_CACHE,INCR_IN_BDDs_CACHEs);
*/

}

/* destroy the reachability cache */
void destroy_cache(GTA *G) {
	unsigned states,i;
	states=G->ss[SS1].size;
	for (i=0;i<states;i++)
		/* clean used cache */
		if (rcache[i].alocated>0) mem_free(rcache[i].from);
	mem_free(rcache);
	bdd_kill_manager(rcache_bdd_mng);
}

void cache_add(unsigned position,unsigned left, unsigned right, bdd_handle hnd) {
	if (rcache[position].no==-1) {
		rcache[position].no=0;
		rcache[position].from=(cache_type *)mem_alloc(sizeof(cache_type)*ALLOC_STEP);
		rcache[position].alocated=ALLOC_STEP;
	} else if(rcache[position].no==rcache[position].alocated) {
		rcache[position].from=(cache_type *)mem_resize(rcache[position].from,
				sizeof(cache_type)*(rcache[position].alocated+ALLOC_STEP));
		rcache[position].alocated=rcache[position].alocated+ALLOC_STEP;
	}
	rcache[position].from[rcache[position].no].left=left;		
	rcache[position].from[rcache[position].no].right=right;		
	rcache[position].from[rcache[position].no].handle=hnd;		
	rcache[position].no++;
}

// we mark the designated target with 1, the rest with 0
unsigned mark_target(unsigned q) {
  return (q == target_state) ? 1 : 0;
}



void compute_predecessors(GTA *G, int actual) {
	/* fill cache for state i with predecessor touples */
	unsigned i,j,states;
	bdd_ptr ptr;
	bdd_handle hnd;
	if (rcache[actual].no!=-1) {
		printf("Warning: runing again for the same state - aborted\n");
		return;
	}
	states=G->ss[SS1].size;
	target_state=actual;

	bdd_prepare_apply1(G->ss[SS1].bddm);  

	for(i=0;i<states;i++)
		for(j=0;j<states;j++) {
			//bdd_prepare_apply1(G->ss[SS1].bddm);  
			ptr=bdd_apply1(G->ss[SS1].bddm, 
		                         BDD_ROOT(G->ss[SS1].bddm,BEH(G->ss[SS1],i,j)),
		                         rcache_bdd_mng,
		                         &mark_target);
			if (bdd_is_leaf(rcache_bdd_mng,ptr))
                     		if (bdd_leaf_value(rcache_bdd_mng,ptr)==0) { 
					// no transition from i,j to q -> delete the result
        	                       	POP_SEQUENTIAL_LIST(rcache_bdd_mng->roots,bdd_handle,hnd);
                                       	continue;
                       		}
			cache_add(actual,i,j,BDD_LAST_HANDLE(rcache_bdd_mng));
		}
	/* -1 in the field rcache[actual].no means 'not precessed by get_cache'
	 * if there is now -1 => no transition points to this state -> we change it to '0' */
	if (rcache[actual].no==-1) rcache[actual].no=0;
}

int eq_classes(unsigned state1,unsigned state2) {
	if ((na_classes[state1]==INITIAL_CLS)||(na_classes[state1]==INITIAL_CLS2)) {
		if  ((na_classes[state2]!=INITIAL_CLS)&&(na_classes[state2]!=INITIAL_CLS2))
			return 0;
		else return 1;
	}
	if (na_classes[state1]!=na_classes[state2]) return 0;
	return 1;
}

int is_init_class(unsigned state1) {
	if ((na_classes[state1]!=INITIAL_CLS)&&(na_classes[state1]!=INITIAL_CLS2)) return 0;	
	return 1;
}

int compare_memory_nodes(GTA *G ,unsigned state1, unsigned state2, unsigned height);

int compare_next_pointer_nodes(GTA *G, unsigned state1, unsigned state2, unsigned height) {
	unsigned i,sort=0;
	unsigned states=G->ss[SS1].size;
	/* check, if state1, and state2 are in the equal class */
	if (na_classes[state1]!=na_classes[state2]) return 0;
	/* look, if predecessors are computed in cache, if not, compute it */	
	if (rcache[state1].no==-1) { compute_predecessors(G,state1); sort=1; }
	if (rcache[state2].no==-1) { compute_predecessors(G,state2); sort=1; }
	if (sort) sort_cache_by_ptrs(states);
	/* if there are different numbers of predecessor touples, exit */
	if (rcache[state1].no!=rcache[state2].no) return 0;
	/* check, if predecessors are in the equal classes, and bdd_pointers are equal */
	for(i=0;i<rcache[state1].no;i++) {
		if (!eq_classes(rcache[state1].from[i].right,rcache[state2].from[i].right)) return 0;
		if (!eq_classes(rcache[state1].from[i].left,rcache[state2].from[i].left)) return 0;
			 
		/* check, if you can reach state1 and state2 by equal symbols -> compare bdd pointers */
		if (get_ptr(rcache[state1].from[i].handle)!=get_ptr(rcache[state2].from[i].handle)) return 0;
	}
	/* OK, in this node everything matches, fork the computation */
	for(i=0;i<rcache[state1].no;i++) {
		if ((i>0)&&(rcache[state1].from[i-1].right==rcache[state1].from[i].right)&&
			(rcache[state2].from[i-1].right==rcache[state2].from[i].right)) {
			if ((!is_init_class(rcache[state1].from[i].right))&&
					(!compare_next_pointer_nodes(G,rcache[state1].from[i].right,rcache[state2].from[i].right,height)))
				return 0;
			/* !!! dokoncit pokracovani pres memory nody se snizovanim height!!!*/
			if ((height)&&(!is_init_class(rcache[state1].from[i].left))) {
				printf("+\n");
				if(!compare_memory_nodes(G,rcache[state1].from[i].left,rcache[state2].from[i].left,height-1))
					return 0;
			}
		}
	}
	return 1;
}


int compare_memory_nodes(GTA *G ,unsigned state1, unsigned state2, unsigned height) {
	unsigned i,sort=0;
	unsigned states=G->ss[SS1].size;
	/* check, if state1, and state2 are memory nodes */
	if ((na_classes[state1]!=MEMN_CLS)||
			(na_classes[state1]!=MEMN_CLS)) return 0;
	/* look, if predecessors are computed in cache, if not, compute it */	
	if (rcache[state1].no==-1) { compute_predecessors(G,state1); sort=1; }
	if (rcache[state2].no==-1) { compute_predecessors(G,state2); sort=1; }
	if (sort) sort_cache_by_ptrs(states);
	/* if there are different numbers of predecessor touples, exit */
	if (rcache[state1].no!=rcache[state2].no) return 0;
	/* check, if predecessors are in the equal classes, and bdd_pointers are equal */
	for(i=0;i<rcache[state1].no;i++) {
		if (na_classes[rcache[state1].from[i].left]!=na_classes[rcache[state2].from[i].left]) return 0;
		/*if (na_classes[rcache[state1].from[i].left]>=NEXTPTR_CLS) return 0;*/
		if (!is_init_class(rcache[state1].from[i].right)) return 0;
		if (!is_init_class(rcache[state2].from[i].right)) return 0;
		/* check, if you can reach state1 and state2 by equal symbols -> compare bdd pointers */
		if (get_ptr(rcache[state1].from[i].handle)!=get_ptr(rcache[state2].from[i].handle)) return 0;
	}
	/* OK, in this node everything matches, fork the computation */
	for(i=0;i<rcache[state1].no;i++) {
		if ((i<=0)||(!rcache[state1].from[i-1].left==rcache[state1].from[i].left)||
			(!rcache[state2].from[i-1].left==rcache[state2].from[i].left))
			continue;
		if (!compare_next_pointer_nodes(G,rcache[state1].from[i].left,rcache[state2].from[i].left,height))
			return 0;
	}
	return 1;
}

/*---------------------------------------------------------------------------*/

/* This function returns list of neighbours for 'state'
 * Neighbour is nearest upper memory node */
int *get_neighbours(GTA *G, unsigned state) {
	/* list is terminated by -1 => it is the reason for signed int */
	char bit_vector[4]; // Used for describing the bit prefix of a symbol when selecting a certain sub-BDD.
	int *tmp,*tmp2,*res;
	unsigned states=G->ss[SS1].size;
	unsigned i,j,k,mark,res_no,init,init2,reject_st;
	/* in todo_done will be placed information about processed pointer nodes
	 *  0 nothing to do (now) with this state
	 *  1 to be processed by read of inner next pointer
	 *  2 to be processed by read memory node 
	 *  3 - reject state - dont process in future */
	int *todo_done;
	todo_done=(int *) mem_alloc(sizeof(int)*states);
	for(i=0;i<states;i++) todo_done[i]=0; /* initial state */
	/* search fo 'init' && 'init2' && reject states */
	init=-1;
	init2=-1;
	reject_st=-1;
	for(i=0;i<states;i++){
		if(na_classes[i]==INITIAL_CLS) init=i; /* there must exist one node in this class */
		if(na_classes[i]==INITIAL_CLS2) init2=i;
		if(na_classes[i]==REJECT_CLS) reject_st=i;
	}

	/* search for next pointer nodes, which are reachable from memory node 'state' */
	bit_vector[0]=0; bit_vector[1]=1; // next pointer node
	mark=0;
	for(i=0;i<states;i++) 
		if ((is_init_class(i))||(na_classes[i]>=NEXTPTR_CLS)) {
			tmp=search_for_states(G,state,i,states,2,bit_vector);
			if (tmp[0]==-1) continue;
			tmp2=result_merge(tmp,todo_done,states);
			mem_free(tmp);
			mem_free(todo_done);
			todo_done=tmp2;
		}
	/* do not process reject state in future */
	todo_done[reject_st]=3;
/*	for(i=0;i<states;i++)printf("%i:%i, ",i,todo_done[i]);
	printf("\n");*/
	mark=1;
	while (mark) {
		mark=0;
		/* process states with '1' in todo_done list */
		for(i=0;i<states;i++)
			if (todo_done[i]==1) {
				todo_done[i]=2; /* mark state by 'processed' label - '2' */
				for(j=0;j<states;j++)
					if ((na_classes[j]==MEMN_CLS)||(is_init_class(j))){
						tmp=search_for_states(G,j,i,states,2,bit_vector);
						if (tmp[0]!=-1)
							for(k=0;k<states;k++)
								if ((tmp[k])&&(todo_done[k]==0)) {
									todo_done[k]=1;
									mark=1;
								}
						mem_free(tmp);
					}
			}
	}
/*	for(i=0;i<states;i++)printf("%i:%i, ",i,todo_done[i]);
	printf("\n");*/
	res=(int *)mem_alloc(sizeof(int)*states+1);
	res_no=0;
	bit_vector[0]=1; bit_vector[1]=0; // next memory node
	for (i=0;i<states;i++) 
		if (todo_done[i]==2) {
			tmp=search_for_states(G,i,init,states,2,bit_vector);	
			if(init2!=-1) tmp2=search_for_states(G,i,init2,states,2,bit_vector);
			else tmp2=tmp;
			for(j=0;j<states;j++)
				if ((((tmp[0]!=-1)&&(tmp[j])) || ((tmp2[0]!=-1)&&(tmp2[j]))) 
						&& (na_classes[j]==MEMN_CLS)
						&& (!is_inside(j,res,res_no)))  {
					/* pridat do listu */
					res[res_no]=j;
					res_no++;
				}
			mem_free(tmp);
			if(init2!=-1) mem_free(tmp2);
		}
	res[res_no]=-1;
	/*printf("RESULT for state %i: ",state);
	for(i=0;res[i]!=-1;i++) printf("%i ,",res[i]);
	printf("\n");*/
	
	mem_free(todo_done);
	return res;
}

/* Find all neigbour states in automaton G */
unsigned *find_neighbours(GTA *G) {
	int *nb;
	unsigned *touples;
	unsigned i,j,states=G->ss[SS1].size,tpl_ptr=0,tpl_alloc,tpl_alloc_step;
	/* initial size of touples field is a pure guess. I thing states*2 will be enaugh in almost all cases */
	tpl_alloc_step=states*2;
	tpl_alloc=tpl_alloc_step;
	touples=(unsigned *) mem_alloc (sizeof(unsigned)*tpl_alloc_step);
	for (i=0;i<states;i++)
		/* if state i is a in class 'memory node' */
		if (na_classes[i]==MEMN_CLS) {
			/* compute neighbours for i, received list is terminated by '-1' */
			/*printf("processing %i ...\n",i);*/
			nb=get_neighbours(G,i);
			for (j=0;nb[j]!=-1;j++) {
				if (tpl_ptr>tpl_alloc-3) { 
					/* not enaugh alocated memory - there must be room for 2 terminal symbols*/
					tpl_alloc=tpl_alloc+tpl_alloc_step;
					touples=(unsigned *) mem_resize(touples,sizeof(unsigned)*tpl_alloc);
				}
				/* we are not interested in touples (x,x) */
				if (i!=nb[j]) {
					/* add touple into a list */
					touples[tpl_ptr]=i;
					touples[tpl_ptr+1]=nb[j];
					tpl_ptr=tpl_ptr+2;
				}
			}
			mem_free(nb);
		}
	/* at the end of touple list will be two symbols '0' 
	 * There is guaranted enaugh alocated room for this terminal symbols */
	touples[tpl_ptr]=0;
	touples[tpl_ptr+1]=0;
	return touples;
}

int *abstract_neighbour(GTA *G) {
	unsigned *nb,i,j,old_class,new_class;
	unsigned states=G->ss[SS1].size;
	//unsigned *classes;
	int *new_classes;
	na_classes=initial_partition_of_ss1(G);
	new_classes=(int *) mem_alloc (sizeof(int)*states);
	for(i=0;i<states;i++) new_classes[i]=i;
	init_cache(G); 
	/* find neighbours touples */
	nb=find_neighbours(G);
	/* go thrue nb list and compare touples of states */
	for(i=0;(nb[i])||(nb[i+1]);i=i+2) {
		old_class=new_classes[nb[i+1]];
		new_class=new_classes[nb[i]];
		if ((compare_memory_nodes(G ,nb[i], nb[i+1], 0)) && (new_class!=old_class))
			/* merge classes represented by nb[i] && nb[i+1] */
			for (j=0;j<states;j++)
				if (new_classes[j]==old_class) new_classes[j]=new_class;
	}
	/* do collapsing */
	/*for(i=0;i<states;i++) printf("%i ",new_classes[i]);
	printf("\n");*/
	//res=collaps2(G,new_classes);
	/* leave memory free */
	destroy_cache(G);	
	mem_free(nb);
	mem_free(na_classes);
	return new_classes;
}

/*--------------------------------------------------------------------
 * Other implementation of finite height abstraction
 * - not generic
 * - take care about structure of pointers
 * - collapsing only memory nodes  
 * -------------------------------------------------------------------  */

int *abstract_fin_height_follow_pointer_str(GTA *aut1,int height) {
	unsigned states,i,j,k;
	int *new_classes;
	int j_class,i_class;
	if (height<0) {
		printf("ERROR: abstract_fin_height_follow_pointer_str: given height is <1\n");
		exit(1);
	}
	/* it is necessary to run minimalization before everything, otherwice seq. fault */
	states=aut1->ss[SS1].size;
	init_cache(aut1); 
	/* this partition is used inside function compare_memory_nodes for node comparision */
	na_classes=initial_partition_of_ss1(aut1);
	/* this partition is created on-the-fly, and then is used for real collapsing */
	new_classes=(int *) mem_alloc (sizeof(int)*states);
	for(i=0;i<states;i++) new_classes[i]=i;

	/* try all touples of memory nodes */
	for(i=0;i<(states-1);i++) {
		if ((na_classes[i])!=MEMN_CLS) continue;
		for(j=i+1;j<states;j++) {
			if ((na_classes[j])!=MEMN_CLS) continue;
			i_class=new_classes[i];
			j_class=new_classes[j];
			if ((compare_memory_nodes(aut1 ,i, j, height)) && (i_class!=j_class))
				/* merge classes represented by j && i */
				for (k=0;k<states;k++)
					if (new_classes[k]==j_class) new_classes[k]=i_class;
	
		}
	}
	//res=collaps2(aut1,new_classes);
	destroy_cache(aut1);	
	mem_free(na_classes);
	return new_classes;
}


