/*
 * Main.java
 */

import java.io.* ;

import mona.* ;
import parser.* ;
import tree.* ;
import automata.* ;
import bdd.* ;


/**
 * Only a main.
 */
public class Main {

  // path of the program file
  private static String PROGRAM_PATH = "program.py" ;
  
  // path of the typedefs files
  private static String TYPEDEF_PATH = "typedefs" ;
  
  // verbose output
  private static boolean verbose = true ;
  
  
  /**
   * Main method.
   */
  public static void main(String[] args) throws Exception {
    
    if (args.length > 1) {
      System.err.println("too few args") ;
      System.exit(1) ;
    }
    
    if (args[0].equals("generate"))
      generate() ;
    
    else if (args[0].equals("descriptor")) {
      verbose = false ;
      descriptor() ;
    }
    
    else {
      System.err.println("wrong command") ;
      System.exit(1) ;
    }
  }
  
  /**
   * Generate BDDs
   */
  private static void generate()
  throws ParseException,PathException,BddException,IOException {
    
    ProgramParser program = parseProgram(PROGRAM_PATH) ;
    
    Typedef[] typedefs = parseTypedef(TYPEDEF_PATH) ;
    
    PatternBuilder patternBuilder = buildPatterns(typedefs,program,true) ;
    
    Pattern[] patterns = patternBuilder.getPatterns() ;
    RouteIndexer routes = patternBuilder.getRouteIndexer() ;
    
    Rule[] rules = buildRules(patterns,routes,program) ;
    
    TypedefBddBuilder bddBuilder = buildBdd(rules,program.getNodeWidth()) ;
    Bdd bdd = bddBuilder.getBdd() ;
    StateIndexer states = bddBuilder.getStateIndexer() ;
    
    writeOutput(bdd,states,new FileOutputStream("init.gta")) ;
    writeProjectBits(program.getNodeWidth()*2,bdd.getMaxBitIndex(),"bit-init.gta") ;
    
    for (int i=0;i<routes.getRouteCount();i++) {
      Route route = routes.getRoute(i) ;
      writeTransducer(i+1,route,program) ;
    }
  }
  
  
  /**
   * Find the number of needed descriptors.
   */
  private static void descriptor()
  throws ParseException,PathException,IOException{
    
   
    Typedef[] typedefs = parseTypedef(TYPEDEF_PATH) ;
    
    PatternBuilder patternBuilder = buildPatterns(typedefs,null,false) ;
    
    Pattern[] patterns = patternBuilder.getPatterns() ;
    RouteIndexer routes = patternBuilder.getRouteIndexer() ;
    
    int max = routes.getRouteCount() ;
    for (Pattern pattern : patterns) {
      int[] marks = pattern.getMarks() ;
      for (int i=0;i<marks.length;i++) {
        max = Math.max(max,marks[i]+1) ;
      }
    }
    System.out.println(max);
  }
  
  
  /**
   * Parse the program definition.
   * 
   * @param path The input file.
   * 
   * @return The parser.
   */
  private static ProgramParser parseProgram(String path)
  throws IOException,ParseException {
    
    if (verbose) {
      System.err.println("Parsing program.....") ;
      System.err.println("Input: "+path) ;
    }
    
    FileReader reader = new FileReader(path) ;
    
    ProgramParser parser = new ProgramParser(reader) ;
    parser.parse() ;
    reader.close() ;
    
    if (verbose) {
      System.err.println("Done !") ;
      System.err.println("Node width: "+parser.getNodeWidth());
      System.err.println("Descriptor count: "+parser.getDescriptorCount()) ;
      System.err.println("Data width: "+parser.getDataWidth());
      System.err.println("Variables:") ;
      for (int i=0;i<parser.getVariableCount();i++)
        System.err.println("\t"+i+": "+parser.getVariable(i));
      System.err.println("Pointers:") ;
      for (int i=0;i<parser.getMemberCount();i++)
        System.err.println("\t"+i+": "+parser.getMember(i));
      System.err.println() ;
    }
    
    return parser ;
  }
  
  
  /**
   * Parse the typedef definition.
   * 
   * @param path The input file.
   * 
   * @return The parsed typedefs.
   */
  private static Typedef[] parseTypedef(String path)
  throws IOException,ParseException,PathException {
    
    if (verbose) {
      System.err.println("Parsing typedefs.....") ;
      System.err.println("Input: "+path) ;
    }
    
    
    FileReader reader = new FileReader(path) ;
    
    TypedefParser parser = new TypedefParser(reader) ;
    parser.parse() ;
    reader.close() ;
    
    Typedef[] typedefs = parser.getTypedefs() ;
    
    if (verbose) {
      System.err.println("Done !") ;
      System.err.println("Types:") ;
      for (Typedef typedef : typedefs) {
        System.err.print("\t"+typedef.getName()) ;
        System.err.print("\t(") ;
        String[] roots = typedef.getRootVariables() ;
        if (roots == null)
          System.err.print("no root") ;
        else {
          System.err.print("roots: ");
          for (String root : roots)
            System.err.print(root+' ');
        }
        
        System.err.println(')');
      }
      System.err.println() ;
    }
    
    return typedefs ;
  }
  
  
  /**
   * Build the tree patterns.
   * 
   * @param typedefs The parsed typedefs.
   * @param program The program parser.
   */
  private static PatternBuilder buildPatterns(Typedef[] typedefs,
      ProgramParser program,boolean filter)
  throws PathException {
    
    if (verbose) {
      System.err.println("Building tree patterns.....") ;
    }
    
    PatternBuilder builder = new PatternBuilder(program) ;
    builder.build(typedefs,filter) ;
    
    Pattern[] patterns = builder.getPatterns() ;
    String[] inputs    = builder.getInputNames() ;
    RouteIndexer routeIndexer = builder.getRouteIndexer() ;
    
    if (verbose) {
      System.err.println("Done !") ;
      
      System.err.println("Present inputs:");
      for (int i=0;i<inputs.length;i++) 
        System.err.println("\t"+i+": "+inputs[i]);
        
      System.err.println("Routes: "+routeIndexer.getRouteCount()) ;
      
      for (int i=0;i<patterns.length;i++) {
        System.err.println("Pattern "+i) ;
        System.err.println("\tType: "+patterns[i].getOutput());
        System.err.print("\tRoots: ") ;
        for (String root : patterns[i].getRootVariables())
          System.err.print(root+" ");
        System.err.println() ;
        System.err.print("\tMarks: ");
        for (int mark : patterns[i].getMarks())
          System.err.print(mark+" ") ;
        System.err.println();

        for (Input input : patterns[i].getInputs()) {
          System.err.print("\t"+input.getName()) ;
          System.err.print(" (") ;
          if (input.isNull())
            System.err.print("null ") ;
          if (input.isPrimitive())
            System.err.print("primitive") ;
          if (input.isUndefined())
            System.err.print("undefined") ;
          System.err.println(')');
        }
        
        
      }
      System.err.println() ;
    }
    
    return builder ;
  }
  
  
  /**
   * Build the automata rules.
   * 
   * @param patterns The tree patterns.
   * 
   * @exception PathException The patterns are invalid for the context.
   */
  private static Rule[] buildRules(Pattern[] patterns,RouteIndexer routes,ProgramParser program)
  throws PathException {
    
    if (verbose) {
      System.err.println("Building automata rules.....") ;
    }
    
    RuleBuilder ruleBuilder = new RuleBuilder(routes,program) ;
    ruleBuilder.build(patterns) ;
    Rule[] rules = ruleBuilder.getRules() ;
    
    if (verbose) {
      System.err.println("Done !") ;
      for (Rule rule : rules) {
        System.err.print("\tleft:\t"+rule.getLeftState());
        if (rule.getLeftState().length() < 8)
          System.err.print('\t');
        System.err.print("\tright:\t"+rule.getRightState());
        if (rule.getRightState().length() < 8)
          System.err.print('\t');
        System.err.print("\tinput:\t"+rule.getInput());
        if (rule.getInput().length() < 8)
          System.err.print('\t');
        System.err.print("\tstate:\t"+rule.getState());
        System.err.println();
      }
      System.err.println();
    }
    
    return rules ;
  }
  
  
  /**
   * Build the BDD.
   * 
   * @param rules The automata rules.
   * @param nodeWodth The allowed binary size.
   */
  private static TypedefBddBuilder buildBdd(Rule[] rules,int nodeWidth)
  throws BddException {
    
    if (verbose) {
      System.err.println("Building BDD nodes.....") ;      
    }
    
    TypedefBddBuilder builder = new TypedefBddBuilder(nodeWidth) ;
    builder.build(rules) ;
    Bdd bdd = builder.getBdd() ;
    StateIndexer states = builder.getStateIndexer() ;
    
    if (verbose) {
      System.err.println("Done !") ;
      System.err.println("Node count: "+bdd.getNodeCount()) ;
      System.err.println("States: ");
      for (int i=0;i<states.getStateCount();i++)
        System.err.println("\t"+i+":\t"+states.getState(i));
      System.err.println();
    }
    
    return builder ;
  }
  
  
  /**
   * Write the mona output.
   * 
   * @param bdd The BDD to write.
   * @param state The automata states.
   * @param stream The mona output.
   */
  private static void writeOutput(Bdd bdd,StateIndexer states,OutputStream stream)
  throws IOException {
    
    if (verbose) {
      System.err.println("Writing mona output.....") ;
    }
    
    Writer writer = new OutputStreamWriter(stream) ;
    
    MonaBuilder monaBuilder = new MonaBuilder(states) ;
    monaBuilder.writeBdd(bdd,writer) ;
    
    writer.close() ;
    
    if (verbose) {
      System.err.println("Done !") ;
      System.err.println();
    }
  }
  
  
  /**
   * Create and write a transducer.
   * 
   * @param index The route index.
   * @param route The route to write.
   * @param program The program parser.
   */
  private static void writeTransducer(int index,Route route,ProgramParser program)
  throws PathException,BddException,IOException {
    
    String file = "descr-"+index+".gta" ;
    
    if (verbose) {
      System.err.println("Creating transducer "+index+"...");
    }
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    TransducerRule[] rules = builder.getRules() ;
    
    if (verbose) {
      System.err.println("Done.");
      for (TransducerRule rule : rules) {
        System.err.print("\tleft:\t"+rule.getLeftState());
        if (rule.getLeftState().length() < 8)
          System.err.print('\t');
        System.err.print("\tright:\t"+rule.getRightState());
        if (rule.getRightState().length() < 8)
          System.err.print('\t');
        System.err.print("\tinput:\t"+rule.getInput());
        if (rule.getInput().length() < 8)
          System.err.print('\t');
        System.err.print("\tstate:\t"+rule.getState());
        System.err.print("\toutput:\t"+rule.getOutput());
        System.err.println();
      }
    }
    
    if (verbose) {
      System.err.println("Building transducer BDD") ;
    }
    
    TransducerBddBuilder bddBuilder = new TransducerBddBuilder(program.getNodeWidth()) ;
    bddBuilder.build(rules) ;
    Bdd bdd = bddBuilder.getBdd() ;
    StateIndexer states = bddBuilder.getStateIndexer() ;
    states.setAcceptState(builder.getAcceptState()) ;
    
    if (verbose) {
      System.err.println("Done !");
    }
    
    if (verbose) {
      System.err.println("Writing mona output to "+file+"....") ;
    }
       
    MonaBuilder monaBuilder = new MonaBuilder(states) ;
    FileWriter writer = new FileWriter(file) ;
    monaBuilder.writeBdd(bdd,writer) ;
    writer.close() ;
    if (verbose) {
      System.err.println("Done !") ;
      System.err.println();
    }
    
    writeProjectBits(program.getNodeWidth()*2,bdd.getMaxBitIndex(),"bit-"+index+".gta") ;
    
  }
  
  
  /**
   * Write list of bits to project.
   */
  private static void writeProjectBits(int start,int end,String file)
  throws IOException {
    
    FileWriter writer = new FileWriter(file) ;
    
    for (int i=start; i<=end; i++) {
      writer.write(String.valueOf(i)) ;
      writer.write(' ') ;
    }
    writer.close() ;
  }
  
}
