/*
 * Rule.java
 */
package automata;

/**
 * Tree automata rule.
 * 
 * @author Pierre Salagnac
 */
public class Rule {
  
  
  /**
   * This state is used for the top rule.
   */
  public final static String ACCEPT_STATE = "A_accept" ;
  
  /**
   * This state is used for the bottoms.
   */
  public final static String BOTTOM_STATE = "A_bottom" ;
  
  /**
   * This state is used for an error in the BDD (reject).
   */
  public final static String ERROR_STATE = "A_error" ;
  
  
  // rule contents
  private String input ;
  private String leftState ;
  private String rightState ;
  private String state ;
  
  
  /**
   * Creates a new instance.
   * 
   * @param input The binary input.
   * @param leftState The left state of the automata.
   * @param rightState The right state of the automata.
   * @param state The output state.
   */
  public Rule(String input,String leftState,String rightState,String state) {
    
    // a state cannot be null
    if (leftState == null)
      throw new IllegalArgumentException("left state is null") ;
    if (rightState == null)
      throw new IllegalArgumentException("right state is null") ;
    if (state == null)
      throw new IllegalArgumentException("output state is null") ;
    
    this.input      = input ;
    
    this.leftState  = leftState ;
    this.rightState = rightState ;
    
    this.state      = state ;
  }
  
  
  /**
   * Returns the binary input of the rule.
   * 
   * @return The binary input.
   */
  public String getInput() {
    return input ;
  }
  
  
  /**
   * Returns the right input state.
   * 
   * @return The right input.
   */
  public String getRightState() {
    return rightState ;
  }
  
  
  /**
   * Returns the left input state.
   * 
   * @return The left state.
   */
  public String getLeftState() {
    return leftState ;
  }
  
  
  /**
   * Returns the output state of this rule.
   * 
   * @return The output state.
   */
  public String getState() {
    return state ;
  }
  
  
  /**
   * Returns a string describing this rule.
   * 
   * @return A string representation of the object.
   */
  @Override
  public String toString() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    buffer.append(getClass().getName()) ;
    buffer.append('[') ;
    
    buffer.append("left=") ;
    buffer.append(leftState) ;
    
    buffer.append(",right=") ;
    buffer.append(rightState) ;
    
    buffer.append(",input=") ;
    buffer.append(input) ;
    
    buffer.append(",state=") ;
    buffer.append(state) ;
    
    buffer.append(']') ;
    
    return buffer.toString() ;
  }
  
}
