/*
 * RuleBuilder.java
 */
package automata ;

import java.util.* ;

import tree.* ;
import parser.* ;

/**
 * Transforms tree patterns into rules.
 * 
 * @author Pierre Salagnac
 */
public class RuleBuilder extends AutomataBuilder {
  
  // indexer de routes
  private RouteIndexer routes ;
  
  // program content
  private ProgramParser program ;
  
  // builtRules
  private List<Rule> rules ;
  
  
  /**
   * Creates a new instance.
   * 
   * @param routes The route indexer.
   * @param dataLength The length of data.
   */
  public RuleBuilder(RouteIndexer routes,ProgramParser program) {
    
    this.routes  = routes ;
    this.program = program ;
    
    rules = new ArrayList<Rule>() ;
    
  }
  
  
  /**
   * Build rules for the specified patterns.
   * 
   * @param patterns The tree patterns.
   * 
   * @return The rules accepting the patterns.
   * 
   * @exception PathException The patterns cannot be converted for the current
   * context.
   */
  public void build(Pattern[] patterns)
  throws PathException {
    
    buildOutRules() ;
    
    // we intercept the case if we have no root
    boolean noRoot = true ;
    for (Pattern pattern : patterns) {
      if (pattern.hasRootVariable())
        noRoot = false ;
    }
      
    if (noRoot) {
      buildZombieRules() ;
      buildNullRules() ;
      buildUndefinedRules(patterns) ;
      buildTopRules() ;
      return ;
    }
    
      
    // type 3 and 4
    buildPatternRules(patterns) ;
    
    // type 1 and 2
    buildNullRules() ;
    buildUndefinedRules(patterns) ;
    
    // top node
    buildTopRules() ;
  }
  
  
  /**
   * Returns all built rules.
   * 
   * @return All automata rules.
   */
  public Rule[] getRules() {
    
    Rule[] array = new Rule[rules.size()] ;
    
    rules.toArray(array) ;
    
    return array ;
    
  }
  
  
  /**
   * Build rules for a pattern.
   * 
   * @param pattern The tree pattern to transform.
   */
  private void buildPatternRules(Pattern[] patterns)
  throws PathException {
    
    // inputs
    String[] inputStates = new String[patterns.length] ;
    for (int i=0;i<patterns.length;i++)
      inputStates[i] = buildInputRules(patterns[i]) ;
    
    // root outputs
    Map<String,String> rootStates = new TreeMap<String,String>() ;
    for (Pattern pattern : patterns) {
      if (!rootStates.containsKey(pattern.getOutput()))
        rootStates.put(pattern.getOutput(),createState()) ;
    }
    
    // type 3 for rules regular node
    for (int i=0;i<patterns.length;i++) {
      Pattern pattern = patterns[i] ;
      String binary = buildType3Node(pattern,null) ;
      rules.add(new Rule(binary,inputStates[i],Rule.BOTTOM_STATE,pattern.getOutput())) ;
    }
    
    // type 3 rules for root nodes
    Map<String,String> rightStates = new TreeMap<String,String>() ;
    rightStates.put(patterns[0].getOutput(),Rule.BOTTOM_STATE) ;
    for (int i=0;i<patterns.length;i++) {
      String[] roots = patterns[i].getRootVariables() ;
      if (roots == null  ||  roots.length == 0)
        continue ;
      
      String binary = buildType3Node(patterns[i],roots) ;
      String state  = rootStates.get(patterns[i].getOutput()) ;
      String right  = rightStates.get(patterns[i].getOutput()) ;
      if (right == null) {
        right = rules.get(rules.size()-1).getState() ;
        rightStates.put(patterns[i].getOutput(),right) ;
      }
      Rule rule = new Rule(binary,inputStates[i],right,state) ;
      
      rules.add(rule) ; 
    }
    
  }
  
  
  /**
   * Creates all input rules for a pattern.
   * 
   * @param pattern The pattern witch contains the inputs.
   * 
   * @return The output state of the top input.
   */
  private String buildInputRules(Pattern pattern) {
    
    Input[] inputs = pattern.getInputs() ;
    
    // most right input
    Rule firstRule = buildInputRule(inputs[inputs.length-1],Rule.BOTTOM_STATE) ;
    rules.add(firstRule) ;
    String rightState = firstRule.getState() ;
    
    // others inputs
    for (int j=inputs.length-2;j>=0;j--) {
      Rule rule = buildInputRule(inputs[j],rightState) ;
      rightState = rule.getState() ;
      rules.add(rule) ;
    }
    
    return rightState ;
  }
  
  
  /**
   * Creates a rule for an input.
   * 
   * @param right The state at the right of the input (in the mona tree).
   */
  private Rule buildInputRule(Input input,String right) {
    
    String binary = buildType4Node(input) ;
    String left   = input.getType() ;
    String state  = createState() ;
    
    if (!input.isPrimitive())
      left = Rule.BOTTOM_STATE ;
    
    Rule rule = new Rule(binary,left,right,state) ;
    
    return rule ;
  }
  
  
  /**
   * Build rules for the null node.
   */
  private void buildNullRules() {
    
    String binary = buildType2Node() ;
    String left   = rules.get(rules.size()-1).getState() ;
    String right  = Rule.BOTTOM_STATE ;
    String state  = createState() ;
    
    Rule rule = new Rule(binary,left,right,state) ;
    rules.add(rule) ;
    
  }
  
  
  /**
   * Build rules for the undefined node.
   * 
   * @param patterns The tree patterns.
   */
  private void buildUndefinedRules(Pattern[] patterns) {
    
    String binary = buildType1Node(patterns) ;
    String left   = rules.get(rules.size()-1).getState() ;
    String right  = Rule.BOTTOM_STATE ;
    String state  = createState() ;
    
    Rule rule = new Rule(binary,left,right,state) ;
    rules.add(rule) ;
    
  }
  
  
  /**
   * Build rules for the nodes not in the tree.
   */
  private void buildOutRules() {
    
    String binary = "00" ;
    String left  = Rule.BOTTOM_STATE ;
    String right = Rule.BOTTOM_STATE ;
    String state = Rule.BOTTOM_STATE ;
    
    Rule rule = new Rule(binary,left,right,state) ;
    rules.add(rule) ;
    
  }
  
  /**
   * Build rules for the top node (program line).
   */
  private void buildTopRules() {
    
    String binary = program.getLineBinary(0) ;
    String left   = rules.get(rules.size()-1).getState() ;
    String right  = Rule.BOTTOM_STATE ;
    String state  = Rule.ACCEPT_STATE ;
    
    Rule rule = new Rule(binary,left,right,state) ;
    rules.add(rule) ;
    
  }
  
  
  /**
   * Build rules for an undefined node (with no child).
   */
  private void buildZombieRules() {
    
    String binary = buildZombieNode() ;
    String left   = Rule.BOTTOM_STATE ;
    String right  = Rule.BOTTOM_STATE ;
    String state  = createState() ;
    
    Rule rule = new Rule(binary,left,right,state) ;
    rules.add(rule) ;
  }
  
  
////////////////////////////////////////////////////////////////////////////////
//                                  BINARY                                    //
////////////////////////////////////////////////////////////////////////////////
  
  
  /**
   * Build the binary code for a type 4 node.
   * 
   * @param input The corresponding input.
   * 
   * @return The code of the node.
   */
  private String buildType4Node(Input input) {
    
    StringBuffer buffer = new StringBuffer() ;
    
    // type 4
    buffer.append("01") ;
    
    // undefined
    if (input.isUndefined()) {
      buffer.append("011") ;
    }
    
    // null
    else if (input.isNull()) {
      buffer.append("010") ;
    }
    
    // route
    else {
      buffer.append('1') ;
      
      int route = input.getRoute() ;
      for (int i=0;i<program.getDescriptorCount()-1;i++) {
        if (i == route)
          buffer.append('1') ;
        else
          buffer.append('0') ;
      }
    }
    
    return buffer.toString() ;
    
  }
  
  
  /**
   * Build the binary code for a type 3 node.
   * 
   * @param pattern The corresponding pattern.
   * @param roots Variables pointing on the node.
   * 
   * @return The code of the node.
   */
  private String buildType3Node(Pattern pattern,String[] roots)
  throws PathException {
    
    StringBuffer buffer = new StringBuffer() ;
    
    // type 3
    buffer.append('1') ;
    
    // root
    if (roots != null)
      buffer.append('1') ;
    else
      buffer.append('0') ;
    
    // no zombies
    buffer.append('0') ;
    
    // variables
    for (int i=0;i<program.getVariableCount();i++) {
      boolean found = false ;
      if (roots != null) {
        for (String root : roots) {
          if (program.getVariableIndex(root) == i)
            found = true ;
        }
      }
        
      if (found)
        buffer.append('1') ;
      else
        buffer.append('0') ;
    }
    
    // routing token
    buffer.append('0') ;
    
    // F marks
    for (int i=0;i<program.getDescriptorCount();i++) {
      if (pattern.containsMark(i))
        buffer.append('1') ;
      else
        buffer.append('0') ;
    }
    
    // data
    String data = pattern.getData() ;
    int dataWidth = program.getDataWidth() ;
    if (data != null) {
      if (data.length() != program.getDataWidth())
        throw new PathException("data length expected:"+dataWidth+", found:"+data.length()) ;
      buffer.append(data) ;
    }
    else {
      for (int i=0;i<dataWidth;i++)
        buffer.append('0') ;
    }
    
    return buffer.toString() ;
  }
  
  
  /**
   * Build the binary code for a type 2 (null) node.
   * 
   * @return The code of the node.
   */
  private String buildType2Node() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    // type 1 or 2
    buffer.append("00") ;
    
    // unused bit
    buffer.append('0') ;
    
    // variables
    for (int i=0;i<program.getVariableCount();i++)
      buffer.append('0') ;
    
    return buffer.toString() ;
    
  }
  
  
  /**
   * Build the binary code for a type 1 (undefined) node.
   * 
   * @param patterns The tree patterns.
   * 
   * @return The code of the node.
   */
  private String buildType1Node(Pattern[] patterns) {
    
    StringBuffer buffer = new StringBuffer() ;
    
    // type 1 or 2
    buffer.append("00") ;
    
    // unused bit
    buffer.append('0') ;
    
    // variables
    for (int i=0;i<program.getVariableCount();i++) {
      String name = program.getVariable(i) ;
      boolean found = false ;
      
      for (Pattern pattern : patterns) {
        if (pattern.containsRootVariable(name))
            found = true ;
      }
      if (found)
        buffer.append('0') ;
      else
        buffer.append('1') ;
    }
    
    return buffer.toString() ;
    
  }
  
  
  /**
   * Build to binary code for a deleted node (zombine).
   * 
   * @return The code of the node.
   */
  private String buildZombieNode() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    buffer.append("111") ;
    
    // lots of zero
    for (int i=0;i<program.getNodeWidth()-3;i++)
      buffer.append('0') ;
    
    return buffer.toString() ;
  }
  
}
