/*
 * TransducerBuilder.java
 */
package automata;

import java.util.* ;

import tree.* ;
import parser.* ;


/**
 * This class transforms a routing expression into a tree transducer.
 * 
 * <b>Note</b> This class assumes that route goes top then down.
 * 
 * @author Pierre Salagnac, hacked by Adam Rogalewicz
 */
public class TransducerBuilder extends AutomataBuilder {

  // result of program parse
  private ProgramParser program ;
  
  // built rules
  private List<TransducerRule> rules ;
  
  // last name, for merge nodes
  private Map<String,ArrayList<String>> mergePoints ;
  private int mp_index1 =0; /* index in mergepoints */
  private int mp_index2 =0; /* 2nd index in mergepoints */
  
  // top state
  private String acceptState ;
  
  
  /**
   * Creates a  new instance.
   * 
   * @param program The result of program parse.
   */
  public TransducerBuilder(ProgramParser program) {
    
    this.program = program ;
    
    rules       = new ArrayList<TransducerRule>() ;
    mergePoints = new TreeMap<String,ArrayList<String>>() ;
  }
  
  
  /**
   * Creates transducer rules.
   * 
   * @param route The route to transform.
   * 
   * @exception PathException The transformation into a transducer failed.
   */
  public void build(Route route) throws PathException {
    
    rules.clear();
    mergePoints.clear() ;
    acceptState = null ;
    
    Path path = route.getPath() ;
    
    buildType1EmptyRules() ;
    buildType3EmptyRules(Rule.BOTTOM_STATE) ;
    buildType4EmptyRules(Rule.BOTTOM_STATE) ;
    
    // only going down
    if (!hasUpMove(path)) {
      String end       = buildEndRule(route,Rule.BOTTOM_STATE,null) ;
      buildPathRules(path,end,false) ;
      String merge     = buildMergeRule(null) ;
      buildStartRule(route,merge,acceptState) ;
    }
    
    // only going up
    else if (!hasDownMove(path)) {
      String start    = buildStartRule(route,Rule.BOTTOM_STATE,null) ;
      buildPathRules(path,start,true) ;
      String end      = buildMergeRule(null) ;
      buildEndRule(route,end,acceptState) ;
    }
    
    // up and down
    else {
      String start = buildStartRule(route,Rule.BOTTOM_STATE,null) ;
      buildPathRules(path,start,true) ;
      
      String end   = buildEndRule(route,Rule.BOTTOM_STATE,null) ;
      buildPathRules(path,end,false) ;
    
      String node = buildMergeRule(null) ;
      buildNodeRule(node,acceptState) ;
    }
    
    // not in the path
    if (!acceptState.equals(Rule.BOTTOM_STATE)) {
      rules.add(new TransducerRule("",null,Rule.BOTTOM_STATE,acceptState,acceptState)) ;
      rules.add(new TransducerRule("",null,acceptState,Rule.BOTTOM_STATE,acceptState)) ;
    }
    
    else
      throw new PathException("accept invalid") ;
    
  }
  
  
  /**
   * Returns all built rules.
   * 
   * @return All automata rules.
   */
  public TransducerRule[] getRules() {
    
    TransducerRule[] array = new TransducerRule[rules.size()] ;
    
    rules.toArray(array) ;
    
    return array ;
    
  }
  
  
  /**
   * Return the accepting state in the rules.
   * 
   * @return The top state.
   */
  public String getAcceptState() {
    return acceptState ;
  }
  
  
  /**
   * Add a state, used for the merge rules just before the merging root node.
   * 
   * @param move The name of the primitive move.
   * @param state An automata state.
   */
  private void addMergePoint(String move,String state) {
    
    ArrayList<String> points = mergePoints.get(state) ;
    if (points == null) {
      points = new ArrayList<String>() ;
      mergePoints.put(state,points) ;
    }
    
    points.add(move) ;
    
  }
  
  
  /**
   * Creates the rules fot the type 1 not in the path.
   */
  private void buildType1EmptyRules() {
    
    String input = "00" ; 
    String state = Rule.BOTTOM_STATE ;
    
    rules.add(new TransducerRule(input,null,state,state,state)) ;
    
    
  }
  
  /**
   * Creates the rules for the type 3 not in the path.
   * 
   * @param state The state to transmit.
   */
  private void buildType3EmptyRules(String state) {
    
    int count = program.getVariableCount() ;
    StringBuffer input = new StringBuffer() ;
    input.append('1') ;          // type 3
    input.append('X') ;          // root
    input.append('X') ;          // zombie
    for (int i=0;i<count;i++)    // variables
      input.append('X') ;
    input.append('0') ;          // token
    
    rules.add(new TransducerRule(input.toString(),null,state,Rule.BOTTOM_STATE,state)) ;
  }
  
  
  /**
   * Creates the rules for the type 4 not in the path.
   * 
   * @param state The state to transmit.
   */
  private void buildType4EmptyRules(String state) {
    
    String input = "01" ; 
    
    rules.add(new TransducerRule(input,null,state,Rule.BOTTOM_STATE,state)) ;
    
    if (!state.equals(Rule.BOTTOM_STATE))
      rules.add(new TransducerRule(input,null,Rule.BOTTOM_STATE,state,state)) ;
    
  }
  
  
  /**
   * Creates the rule starting at the token.
   * 
   * @param route The route to transform into a transducer.
   * @param start The start state.
   * @param end The end state, or null to create one.
   * 
   * @return The state at the end.
   */
  private String buildStartRule(Route route,String start,String end) {
    
    int count = program.getVariableCount() ;
    
    StringBuffer buffer = new StringBuffer() ;
    buffer.append('1') ;        // type 3
    if (isRootStart(route))     // root
      buffer.append('1') ;
    else
      buffer.append('X') ;
    buffer.append('X') ;        // zombie
    for (int i=0;i<count;i++)   // variables
      buffer.append('X') ;
    buffer.append('1') ;        // token
    
    String input  = buffer.toString() ;
    buffer.setCharAt(buffer.length()-1,'0') ;
    String output = buffer.toString() ;
    String left   = start ;
    String right  = Rule.BOTTOM_STATE ;
    String state  = (end == null) ? (createState()) : (end) ;
    
    rules.add(new TransducerRule(input,output,left,right,state)) ;
    
    return state ;
  }
  
  
  /**
   * Creates the rule ending the route.
   * 
   * @param route The route to transform into a transducer.
   * @param start The start state.
   * @param end The end state, or null to create one.
   * 
   * @return The state at the end.
   * 
   */
  private String buildEndRule(Route route,String start,String end) {
    
    //TODO: est-ce possible qu'un pointeur revienne sur lui ??
    // si oui, mettre le token a X
    
    int vars = program.getVariableCount() ;
    int desc = program.getDescriptorCount() ;
    
    StringBuffer buffer = new StringBuffer() ;
    buffer.append('1') ;        // type 3
    if (isRootEnd(route))       // root
      buffer.append('1') ;
    else
      buffer.append('X') ;
    buffer.append('X') ;        // zombie
    for (int i=0;i<vars;i++)    // variables
      buffer.append('X') ;
    buffer.append('0') ;        // token
    int tokenIndex = buffer.length()-1 ;
    for (int i=0;i<desc;i++) {  // F marks
      if (route.containsMark(i))
        buffer.append('1') ;
      else
        buffer.append('X') ;
    }
    
    String input = buffer.toString() ;
    buffer.setCharAt(tokenIndex,'1') ;
    String output = buffer.toString() ;
    String left   = start ;
    String right  = Rule.BOTTOM_STATE ;
    String state  = (end == null) ? (createState()) : (end) ;
    
    rules.add(new TransducerRule(input,output,left,right,state)) ;
    
    return state ;
  }
  
  
  /**
   * Creates the rule witch try to merge the left and right parts of the
   * path.
   * 
   * @param end The end state, or null to create one.
   * 
   * @return The state at the end.
   * 
   * @exception PathException The merge points are invalids. 
   */
  private String buildMergeRule(String end) throws PathException {
    
    if (end == null)
      end = createState() ;
    
    Map<String,String> point = getNextMergePoint() ;

    /* There must be at leas one merge point */
    if (point==null) {
       throw new PathException("ERROR: No Mergepoints");
    }

    while(point!=null) {

      System.err.println("mergepoint:  ("+point+")") ;
      int index = Integer.MIN_VALUE ;
      for (String name : point.keySet()) {
        int nameIndex = program.getMemberIndex(name) ;
        index = Math.max(index,nameIndex) ;
      }
      
      String lastState = null ;
      
      for (int i=index;i>=0;i--) {
        String state,leftState,rightState ;
  
        // first
        if (i == index)
          rightState = Rule.BOTTOM_STATE ;
        else
          rightState = lastState;
        
        // last
        if (i == 0)
          state = end ;
        else
          state = createState() ;
        
        // input
        leftState = Rule.BOTTOM_STATE ;
        for (String name : point.keySet()) {
          int nameIndex = program.getMemberIndex(name) ;
          if (i == nameIndex)
            leftState = point.get(name) ;
        }
          
	System.err.println("Rule: "+leftState+" "+rightState+" -> "+state) ;	
        rules.add(new TransducerRule("01",null,leftState,rightState,state)) ;
        lastState = state ;
      }
      /* get net merge point to process */
      point = getNextMergePoint() ;
    }
    
    return end ;
  }
  
  
  
  /**
   * Creates a map with unused merge point.
   * 
   * @return A map with move names and automata states.
   */
  private Map<String,String> getNextMergePoint() throws PathException {
    
    /* This is only hack into the original code to make it run ...
     * It is not a nice piece of programming :-( 
     */
    if (mergePoints.isEmpty())
    	throw new PathException("no merge point") ;

    /* There can be 2 types of merge points:
     - single - only one automata state and one move 
       (The moves are stored in the first element of "mergePoints" && there is no second element)
     - double - 2 different automata states and 2 different moves 
       (The first move is taken from the first element of mergePoints, and the second from the second one)

     - variables mp_index[12] are used to point to the next mergepoint

       It is unpossible to have this two types in one transducer
     */
    Map<String,String> point = new TreeMap<String,String>() ;
    String move=null;
    String move2=null;
    int i=1;
    /* The following for loop is executed at most 2 times */
    for (String state : mergePoints.keySet()) {
    	if (i==3) 
		throw new PathException("Mergepoints: the case of 3 different states is not correctly handled -> STOPED") ;

	if (i==1) {
		ArrayList<String> moves = mergePoints.get(state) ;
		if (mp_index1==moves.size()) return null;
        	move=moves.get(mp_index1); 
    		point.put(move,state) ;
	} else {
		ArrayList<String> moves = mergePoints.get(state) ;
        	move2=moves.get(mp_index2); 
    		point.put(move2,state) ;
		mp_index2=mp_index2+1;
		if (mp_index2==moves.size()) {
			mp_index1=mp_index1+1;
			mp_index2=0;
		}
	}
    	i=i+1;
    }
    if (i==2) { 
    	mp_index1=mp_index1+1;
	mp_index2=0;
    }
    /* the case, where the two moves are equal is not a correct mergepoint
     * => return the next one
     */
    if (move.equals(move2)) 
    	return getNextMergePoint();

    return point ;
  }
  
  
  /**
   * Build rules for an unique-direction path.
   * 
   * @param path The path to transform.
   * @param start The start state.
   * @param reversed false for up, true for down.
   * 
   * @return The state at the end.
   * 
   * @exception PathException The conversion into rules failed.
   */
  private String buildPathRules(Path path,String start,boolean reversed)
  throws PathException {
    
    // disjunction
    if      (path instanceof Disjunction) {
      Disjunction disjunction = (Disjunction)path ;
      String end = buildDisjunctionRules(disjunction,start,reversed) ;
      if (disjunction.isMultiple())
        acceptState = start ;
      else
        acceptState = createState() ;
      return end ;
    }
    
    // sequence
    else if (path instanceof Sequence) {
      Sequence sequence = (Sequence)path ;
      String end = buildSequenceRules(sequence,start,reversed) ;
      if (sequence.isMultiple())
        acceptState = start ;
      else
        acceptState = createState() ;
      return end ;
    }
    
    // single move
    else if (path instanceof Move) {
      Move move = (Move)path ;
      if (move.isReversed() != reversed)
        throw new PathException("wrong direction") ;
      addMergePoint(move.getName(),start) ;
      if (move.isMultiple())
        acceptState = start ;
      else
        acceptState = createState() ;
      return start ;
    }
    
    // error
    else
      throw new PathException("wrong path type: "+path) ;

  }
  
  
  /**
   * Creates the rules for a sequence.
   * 
   * @param sequence The sequence to transform.
   * @param start The start state.
   * @param reversed false for up, true for down.
   * 
   * @return The state at the end.
   * 
   * @exception PathException The conversion into rules failed.
   */
  private String buildSequenceRules(Sequence sequence,String start,boolean reversed)
  throws PathException {
    
    // only one element
    if (sequence.getChildrenCount() <= 1)
      throw new PathException("minimization needed, sequence with one child") ;
    
    int i = 0 ;
    boolean first = true ;
    String nextEnd = null ;
    
    while (true) {
      
      Path child = getSubPath(sequence,i,reversed) ;
      
      // end of sequence
      if (child == null)
        return start ;
      
      // no node at the start
      if (first)
        first = false ;
      else
        start = buildNodeRule(start,nextEnd) ;
      
      // disjunction
      if      (child instanceof CompoundPath) {
        Disjunction disjunction = (Disjunction)child ;
        start = buildDisjunctionRules(disjunction,start,reversed) ;
        if (isLast(disjunction,reversed))
          return start ;
      }
      
      // single move
      else if (child instanceof Move) {
        Move move = (Move)child ;
        if (move.isReversed() != reversed)
          throw new PathException("wrong direction") ;
        
        if (isLastChild(sequence,i,reversed)) {
          addMergePoint(move.getName(),start) ;
          return start ;
        }
        else {
          nextEnd = (move.isMultiple()) ? (start) : (null) ;
          start = buildMoveRules(move,start,null) ;
        }
      }
      
      // error
      else
        throw new PathException("wrong path type: "+child) ;
      
      
      i++ ;
    }
  }
  
  
  /**
   * Creates the rules for a disjunction.
   * 
   * @param disjunction The disjunction to transform.
   * @param start The start state.
   * @param end The end state, or null to create one.
   * 
   * @return The state at the end.
   * 
   * @exception PathException The conversion into rules failed.
   */
  private String buildDisjunctionRules(Disjunction disjunction,String start,boolean reversed)
  throws PathException {
    
    int count = disjunction.getChildrenCount() ;
    String mergeEnd = createState() ;
    
    // only one element
    if (count <= 1)
      throw new PathException("minimization needed, disjunction with one child") ;
    
    
    // children
    for (int i=0;i<count;i++) {
            
      Path child = disjunction.getChild(i) ;
      
      if      (child instanceof Move) {
        Move move = (Move)child ;
        if (move.isReversed() != reversed)
          throw new PathException("wrong direction") ;
        
        if (isLast(disjunction,reversed))
          addMergePoint(move.getName(),start) ;
        else
          buildMoveRules(move,start,mergeEnd) ;
        
        
      }
      
      else if (child instanceof Sequence) {
        Sequence sequence = (Sequence)child ;
        buildSequenceRules(sequence,start,reversed) ;
      }
      
      else if (child instanceof Disjunction)
        throw new PathException("minimization needed (compound disjunction)") ;
      
      else
        throw new PathException("wrong path type: "+child) ;
    }
    
    // end node
    if (!isLast(disjunction,reversed)) {
      return mergeEnd ;
    }
    else
      return start ;
    
  }
  
  
  /**
   * Creates the rules for a single move. The type 4 rules are added.
   * 
   * @param move The move to transform.
   * @param start The start state.
   * @param end The end state, or null to create one.
   * 
   * @return The state at the end.
   * 
   * @exception PathException The start and end states are invalid for the
   * move.
   */
  private String buildMoveRules(Move move,String start,String end)
  throws PathException {
    
    // end state
    if (end == null)
      end = createState() ;
    
    int index = program.getMemberIndex(move.getName()) ;
    String lastState = null ;
    
    for (int i=index;i>=0;i--) {
      String state,leftState,rightState ;

      // first
      if (i == index) {
        rightState = Rule.BOTTOM_STATE ;
        leftState  = start ;
      }
      else {
        rightState = lastState;
        leftState  = Rule.BOTTOM_STATE ;
      }
      
      // last
      if (i == 0)
        state = end ;
      else
        state = createState() ;
      
      rules.add(new TransducerRule("01",null,leftState,rightState,state)) ;
      lastState = state ;
    }
    
    return end ;
    
  }
  
  
  /**
   * Build a rule for a type 3 memory node.
   * 
   * @param left The left state of the node (input).
   * @param end The end state, or null to create one.
   * 
   * @return The state at the end.
   */
  private String buildNodeRule(String left,String end) {
    
    if (end == null)
      end = createState() ;
    
    int count = program.getVariableCount() ;
    
    StringBuffer buffer = new StringBuffer() ;
    buffer.append('1') ;       // type 3
    buffer.append('X') ;       // root
    buffer.append('X') ;       // zombie
    for (int i=0;i<count;i++)  // variables
      buffer.append('X') ;
    buffer.append('0') ;       // token
    
    rules.add(new TransducerRule(buffer.toString(),null,left,Rule.BOTTOM_STATE,end)) ;
    
    return end ;
  }
  
  
  /**
   * Returns the top-level subpath for the given index.
   * 
   * @param sequence The parent sequence.
   * @param index The index to search.
   * @param reversed Start at the end.
   * 
   * @return The path, or null if no path was found.
   * 
   * @exception PathException The specified path is invalid, or needs a
   * minimization.
   */
  private Path getSubPath(Sequence sequence,int index,boolean reversed)
  throws PathException {
    
    int count = sequence.getChildrenCount() ;
    if (index >= count)
      return null ;
    
    if (!reversed)
      index = count-index-1 ;
    
    Path child = sequence.getChild(index) ;
    if (child instanceof Sequence)
      throw new PathException("minimization nedded (compound sequence)") ;
    
    if (child instanceof Move) {
      Move move = (Move)child ;
      if (move.isReversed() != reversed)
        throw new PathException("no more move (index:"+index+")") ;
    }
    
    return child ;
    
  }
  
  
  /**
   * Returns if the move is the last.
   * 
   * @param path The compound path.
   * @param index The child index.
   * @param reversed The route part direction.
   * 
   * @return true if the child is the last.
   */
  private boolean isLastChild(Path path,int index,boolean reversed)
  throws PathException {
    
    // single move
    if (path instanceof Move) {
      if (index == 0)
        return true ;
      else
        throw new PathException("wrong move") ;
    }
    
    // compound
    else {
      CompoundPath compound = (CompoundPath)path ;
      
      // going up
      if (reversed) {
        if (index+1 == compound.getChildrenCount())
          return true ;
        Path next = compound.getChild(index+1) ;
        return !isFirstReversed(next) ;
      }
      
      // going down
      else {
        index = compound.getChildrenCount()-index-1 ;
        if (index == 0)
          return true ;
        Path next = compound.getChild(index-1) ;
        return isFirstReversed(next) ;
      }
    }
    
  }
  
  
  /**
   * Returns if the specified disjunction is the last in its parent sequence.
   * 
   * @param disjunction The disjunction to test.
   * @param reversed The path direction.
   * 
   * @return true if the disjunction is the last.
   * 
   * @exception PathException The disjunction parent is invalid.
   */
  private boolean isLast(Disjunction disjunction,boolean reversed)
  throws PathException {
    
    Path parent = disjunction.getParent() ;
    
    if (parent == null)
      return true ;
    
    if (parent instanceof Sequence) {
      Sequence sequence = (Sequence)parent ;
      
      int index = -1 ;
      for (int i=0;i<sequence.getChildrenCount();i++) {
        Path child = sequence.getChild(i) ;
        if (child.equals(disjunction)) {
          index = i ;
        }
      }
      
      if (index == -1)
        throw new PathException("not found in parent") ;
      
      if (reversed) {
        if (index == sequence.getChildrenCount()-1)
          return true ;
        return !isFirstReversed(sequence.getChild(index+1)) ;
      }
      else {
        if (index == 0)
          return true ;
        return isFirstReversed(sequence.getChild(index-1)) ;
      }
    }
    
    // error
    else
      throw new PathException("wrong path structure") ;
    
  }
  
  
  /**
   * Returns if the first primitive move in the path is reversed.
   * 
   * @param path The path to test.
   * 
   * @retur true If the first move has the direction.
   * 
   * @exception PathException A disjunction has both directions.
   */
  private boolean isFirstReversed(Path path)
  throws PathException {
    
    // single move
    if (path instanceof Move) {
      Move move = (Move)path ;
      return move.isReversed() ;
    }
    
    // sequence
    else if (path instanceof Sequence) {
      Sequence sequence = (Sequence)path ;
      Path child = sequence.getChild(0) ;
      return isFirstReversed(child) ;
    }
    
    // disjunction
    else if (path instanceof Disjunction) {
      Disjunction disjunction = (Disjunction)path ;
      boolean first = isFirstReversed(disjunction.getChild(0)) ;
      for (int i=1;i<disjunction.getChildrenCount();i++) {
        if (first != isFirstReversed(disjunction.getChild(i)))
          throw new PathException("disjunction with two directions") ;
      }
      return first ;      
    }
    
    // error
    else
      throw new PathException("wrong path type") ;
    
  }
  
  
  /**
   * Returns if the specified path has a going up move.
   * 
   * @param path The path, can be compound.
   * 
   * @return true if there is a reversed move.
   */
  private boolean hasUpMove(Path path) {
    return hasMove(path,true) ;
  }
  
  
  /**
   * Returns if the specified path has a going down move.
   * 
   * @param path The path, can be composed.
   * 
   * @return true if the is a direct move.
   */
  private boolean hasDownMove(Path path) {
    return hasMove(path,false) ;
  }
  
  
  /**
   * Returns if there is a primitive move with the specified direction.
   * 
   * @param path The path to test.
   * @param direction The direction to search.
   * 
   * @return true if the direction was found.
   */
  private boolean hasMove(Path path,boolean direction) {
    
    // single move
    if (path instanceof Move) {
      Move move = (Move)path ;
      boolean reversed = move.isReversed() ;
      return (reversed == direction) ;
    }
    
    // compound
    else {
      CompoundPath compound = (CompoundPath)path ;
      for (int i=0;i<compound.getChildrenCount();i++) {
        Path child = compound.getChild(i) ;
        if (hasMove(child,direction))
          return true ;
      }
      return false ;
    }
    
  }
  
  
  
  /**
   * Tests if the route starts at the root absolute point.
   * 
   * @param route The route to test.
   * 
   * @return true if the route starts at the root.
   */
  private boolean isRootStart(Route route) {
    
    Path path = route.getPath() ;
    String point = path.getStartPoint() ;
    
    if (point != null  &&  point.equals("root"))
      return true ;
    else
      return false ;
    
  }
  
  
  /**
   * Tests if the route ends at the root absolute point.
   * 
   * @param route The route to test.
   * 
   * @return true if the route ends at the root.
   */
  private boolean isRootEnd(Route route) {
    
    Path path = route.getPath() ;
    String point = path.getEndPoint() ;
    
    if (point != null  &&  point.equals("root"))
      return true ;
    else
      return false ;
    
  }
  
}
