/*
 * BddBuilder.java
 */
package bdd ;

import java.util.* ;

import automata.* ;


/**
 * Creates a BDD from automata rules.
 * 
 * @author Pierre Salagnac
 */
abstract class BddBuilder<T extends Rule> {
  
  // the build BBDs
  private List<BddTree> trees ;
  private List<BddNode> nodes ;
  
  // all automata states
  private StateIndexer states ;
  
  // max node width
  private int nodeWidth ;
  
  
  /**
   * Creates a new instance.
   * 
   * @param nodeWidth The max length of a node binary.
   */
  public BddBuilder(int nodeWidth) {
    
    this.nodeWidth = nodeWidth ;
    
    trees = new ArrayList<BddTree>() ;
    nodes = new ArrayList<BddNode>() ;
    
  }
  
  
  /**
   * Build some BDDs for the automata rules.
   * 
   * @param rules The automata rules to transform.
   * 
   * @exception BddException The rules cannot be converted into a BDD.
   */
  public void build(T[] rules) throws BddException {
    
    // create state list
    states = new StateIndexer(rules) ;
    states.setBottomState(Rule.BOTTOM_STATE) ;
    states.setErrorState(Rule.ERROR_STATE) ;
    states.setAcceptState(Rule.ACCEPT_STATE) ;
    
    buildNodes(rules) ;
    
    
    // expand nodes
    for (BddNode node : nodes) {
      node.expandNotLeaf(nodeWidth*2) ;
      node.expandMultiple(nodeWidth*2) ;
    }
    for (BddTree tree : trees)
      tree.expand(nodeWidth*2) ;
    
    // create index of nodes
    generateIndexes() ;
    
  }
  
  
  /**
   * Build the BDD nodes.
   * 
   * @param rules The automata rules.
   */
  abstract void buildNodes(T[] rules) throws BddException ;
  
  
  /**
   * Returns the build BDD.
   * 
   * @return The complete BDD.
   */
  public Bdd getBdd() {
    
    BddTree[] ts = new BddTree[trees.size()] ;
    trees.toArray(ts) ;
    
    BddNode[] ns = new BddNode[nodes.size()] ;
    nodes.toArray(ns) ;

    return new Bdd(ts,ns) ;
  }
  
  
  /**
   * Returns the automata states.
   * 
   * @return The indexed states.
   */
  public StateIndexer getStateIndexer() {
    return states ;
  }
  
  
  /**
   * Returns the node with of this builder.
   * 
   * @return The max node with.
   */
  int getNodeWidth() {
    return nodeWidth ;
  }
  
  
  /**
   * Adds a single node (not connected) in the bdd.
   * 
   * @param node The node to add.
   */
  void addNode(BddNode node) {
    nodes.add(node) ;
  }
  
  
  /**
   * Returns the BDD of the specified input states.
   * 
   * @param left The left input state.
   * @param right The right input state.
   */
  BddTree getTree(String left,String right) {
    
    BddTree tree = null ;
    
    // search
    for (BddTree t : trees) {
      String l = t.getLeftState() ;
      String r = t.getRightState() ;
      
      if (equalStates(left,right,l,r)) {
        tree = t ;
        break ;
      }
    }
    
    
    // new tree
    if (tree == null) {
      tree = new BddTree(left,right) ;
      trees.add(tree) ;
      return tree ;
    }
    
    // tree found
    else
      return tree ;
    
  }
  
  
  /**
   * Returns if two pairs of automata states are equals.
   * 
   * @param l1 The left state of the first pair.
   * @param r1 The right state of the first pair.
   * @param l2 The left state of the second pair.
   * @param r2 The right state of the second pair.
   * 
   * @return true If the two pair are equals.
   */
  private boolean equalStates(String l1,String r1,String l2,String r2) {
    
    // left
    boolean left ;
    if      (l1 == null  &&  l2 == null)
      left = true ;
    else if (l1 == null  ||  l2 == null)
      left = false ;
    else
      left = l1.equals(l2) ;
    
    // right
    boolean right ;
    if      (r1 == null  &&  r2 == null)
      right = true ;
    else if (r1 == null  ||  r2 == null)
      right = false ;
    else
      right = r1.equals(r2) ;
    
    return left & right ;
  }
  
  
  /**
   * Generates indexes fo all nodes.
   */
  private void generateIndexes() {
    
    int index = 0 ;
    
    for (BddNode node : nodes)
      index = generateIndexes(node,index) ;
    
    for (BddNode node : trees)
      index = generateIndexes(node,index) ;
    
  }
  
  
  /**
   * Generates indexes a tree.
   * 
   * @param node The root node.
   * @param index The first index to use.
   * 
   * @return The next index to use.
   */
  private int generateIndexes(BddNode node,int index) {
    
    if (node.getIndex() != -1)
      return index ;
    
    node.setIndex(index) ;
    index++ ;
    
    BddNode trueChild  = node.getChild(true) ;
    BddNode falseChild = node.getChild(false) ;
    
    // left recursivity
    if (trueChild != null)
      index = generateIndexes(trueChild,index) ;
    
    // right recursivity
    if (falseChild != null)
      index = generateIndexes(falseChild,index) ;
    
    return index ;
    
  }
  
}
