/*
 * BddNode.java
 */
package bdd ;

import java.util.* ;


/**
 * Simple node of a BDD tree.
 * 
 * @author Pierre Salagnac
 * 
 * methods "add_list_of_states" and "get_all_states" added by Adam Rogalewicz
 */
public class BddNode {
  
  // children
  private BddNode trueChild ;
  private BddNode falseChild ;
  
  // position in the binary input
  private int bitIndex ;
  
  // unique index in the bdds
  private int index ;
  
  // automata states
  private List<Integer> states ;
  
  
  /**
   * Create a new instance.
   * 
   * @param bitIndex The index in the input.
   */
  BddNode(int bitIndex) {
    
    this.bitIndex = bitIndex ;
    
    index = -1 ;
    states = new ArrayList<Integer>() ;
  }
  
  
  /**
   * Determine a child.
   * 
   * @param side The position of the child.
   * @param child The child to add.
   */
  void setChild(boolean side,BddNode child) {
    
    if (side)
      trueChild = child ;
    else
      falseChild = child ;
    
  }
  
  
  /**
   * Return a child of this node.
   * 
   * @param side The position of the child.
   */
  public BddNode getChild(boolean side) {
    
    if (side)
      return trueChild ;
    else
      return falseChild ;
    
  }
  
  
  /**
   * Return if this node is a leaf.
   * 
   * @return true if the node is a leaf.
   */
  boolean isLeaf() {
    
    return (trueChild == null)  &&  (falseChild == null) ;  
    
  }
  
  
  /**
   * Set the bdd index.
   * 
   * @param index The node index.
   */
  public void setIndex(int index) {
    this.index = index ;
  }
  
  
  /**
   * Return the index of this node in the Mona data.
   * 
   * @return The node index.
   */
  public int getIndex() {
    return index ;
  }
  
  
  /**
   * Return the index of the bit in the binary input.
   * 
   * @return The input index.
   */
  public int getBitIndex() {
    return bitIndex ;
  }
  
  
  /**
   * Add an automata state.
   * 
   * @param state The automata state.
   */
  void addState(int state) {
    
    Integer integer = Integer.valueOf(state) ;
    
    states.add(integer) ;
    
  }
  
  
  /**
   * Return the state of the node. An exception is thrown if the node doesn't
   * have exactly one state.
   * 
   * @return The automata state, ot -1 if the node has no associated state.
   */
  public int getState() {
    
    // no state
    if (states.isEmpty())
      return -1 ;
    
    // error (expand needed)
    if (states.size() > 1)
      throw new IllegalStateException("expand before ("+this+")") ;
    
    return states.get(0) ;
  }
  
  /**
   * Retur a copy of list of states
   */
  public List<Integer> get_all_states(){
    	List<Integer> export_states = new ArrayList<Integer>();
    	for (int i=0;i<states.size();i++)
		export_states.add(states.get(i));
	return export_states;	
  	
  }
  /**
   * Add an list of automata states
   */
  public void add_list_of_states(List<Integer> in_list) {
  	for (int i=0;i<in_list.size();i++)
		states.add(in_list.get(i));
  }
  
  /**
   * Return the children depth.
   */
  public int getMaxBitIndex() {
    
    int index = bitIndex ;
    
    if (falseChild != null)
      index = Math.max(index,falseChild.getMaxBitIndex()) ;
    
    if (trueChild != null)
      index = Math.max(index,trueChild.getMaxBitIndex()) ;
    
    return index ;
  }
  
  /**
   * Return a string describing this node.
   * 
   * @return A string representation of the object.
   */
  @Override
  public String toString() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    buffer.append(getClass().getName()) ;
    buffer.append('[') ;
    
    // bitIndex
    buffer.append("bit=") ;
    buffer.append(bitIndex) ;
    
    // index
    buffer.append(",index=") ;
    buffer.append(index) ;
    
    // states
    buffer.append(",states={") ;
    for (int i=0;i<states.size();i++) {
      if (i != 0)
        buffer.append(',') ;
      buffer.append(states.get(i)) ;
    }
    buffer.append('}') ;
    
    // false child
    buffer.append(",false=") ;
    if (falseChild == null)
      buffer.append("null") ;
    else
      buffer.append(falseChild.getIndex()) ;
    
    // true child
    buffer.append(",true=") ;
    if (trueChild == null)
      buffer.append("null") ;
    else
      buffer.append(trueChild.getIndex()) ;
    
    buffer.append(']') ;
    
    return buffer.toString() ;
  } 

  
////////////////////////////////////////////////////////////////////////////////
//                               EXPAND                                       //
////////////////////////////////////////////////////////////////////////////////
  
  /**
   * Expand the BDD to move all automata states to leaves.
   * 
   * @param bit The next available bit index.
   */
  void expandNotLeaf(int nextBit) {

    // no change for a leaf
    if (isLeaf())
      return ;
    
    // no change if we have no state
    if (states.isEmpty()) {
      if (falseChild != null)
        falseChild.expandNotLeaf(nextBit) ;
      if (trueChild != null)
        trueChild.expandNotLeaf(nextBit) ;
      return ;
    }
    
    
    int[] movedStates = new int[states.size()] ;
    for (int i=0;i<states.size();i++)
      movedStates[i] = states.get(i).intValue();
    states.clear() ;
    
    int bit = Math.max(bitIndex+1,nextBit) ;
    
    if (falseChild == null)
      falseChild = new BddNode(bit) ;
    if (trueChild  == null)
      trueChild  = new BddNode(bit) ;
    
    for (int state : movedStates) {
      falseChild.addState(state) ;
      trueChild.addState(state) ;
    }
    
    // recursion
    falseChild.expandNotLeaf(nextBit) ;
    trueChild.expandNotLeaf(nextBit) ;
    
  }
  
  
  /**
   * Expand the BDD to duplicate nodes with several automata states.
   * 
   * Here we are sure that states are on leaves.
   * 
   * @param bit The next available bit index.
   * 
   * @exception BddException The expansion process is wrong.
   */
  void expandMultiple(int nextBit)
  throws BddException {
    
    // not state, recusive call without increment
    if (states.isEmpty()) {
      if (trueChild != null)
        trueChild.expandMultiple(nextBit) ;
      if (falseChild != null)
        falseChild.expandMultiple(nextBit) ;
      return ;
    }
    
    // we have states in a not-leaf
    else if (!isLeaf())
      throw new BddException("expand a multiple node, but not leaf") ;
    
    
    int size = states.size() ;
    int bit = Math.max(bitIndex,nextBit) ;
    
    // two leaf children
    if (size == 2) {
      int falseState = states.get(0) ;
      int trueState  = states.get(1) ;
      states.clear();
      
      bitIndex = bit ;
      falseChild = new BddNode(bit+1) ;
      trueChild  = new BddNode(bit+1) ;
      
      falseChild.addState(falseState) ;
      trueChild.addState(trueState) ;
    }
    
    // lots of states !
    else if (size > 2) {
      int middle = size/2 ;
      int[] falseStates = new int[middle] ;
      int[] trueStates  = new int[size-middle] ;
      for (int i=0;i<middle;i++)
        falseStates[i] = states.get(i).intValue() ;
      for (int i=middle;i<size;i++)
        trueStates[i-middle] = states.get(i).intValue() ;
      states.clear() ;
      
      bitIndex = bit ;
      falseChild = new BddNode(bit+1) ;
      trueChild  = new BddNode(bit+1) ;
      
      for (int state : falseStates)
        falseChild.addState(state) ;
      for (int state : trueStates)
        trueChild.addState(state) ;
    }
    
    
    // true side recursivity
    if (trueChild != null)
      trueChild.expandMultiple(nextBit+1) ;
    
    // false side recursivity
    if (falseChild != null)
      falseChild.expandMultiple(nextBit+1) ;
    
  }
  
}
