/*
 * StateIndexer.java
 */
package bdd ;

import java.util.* ;

import automata.* ;


/**
 * This class is an index for all automata states.
 * 
 * @author Pierre Salagnac
 */
public class StateIndexer {
  
  // states array
  private List<String> states ;
  
  // special states
  private String bottomState ;
  private String errorState ;
  private String acceptState ;
  
  
  /**
   * Create a new instance.
   * 
   * @param rules All automata rules.
   */
  StateIndexer(Rule[] rules) {
    
    states = new ArrayList<String>() ;
    
    createStates(rules) ;
    
  }
  
  
  /**
   * Create a list with all automata states.
   * 
   * @param rules All automata rules.
   */
  private void createStates(Rule[] rules) {
    
    // fixed states
    states.add(Rule.ERROR_STATE) ;
    states.add(Rule.BOTTOM_STATE) ;
    states.add(Rule.ACCEPT_STATE) ;
    
    for (Rule rule : rules) {
      
      String out   = rule.getState() ;
      String left  = rule.getLeftState() ;
      String right = rule.getRightState() ;
      
      if (!states.contains(out))
        states.add(out) ;
      
      if (left != null  &&  !states.contains(left))
        states.add(left) ;
      
      if (right != null &&  !states.contains(right))
        states.add(right) ;
    }
    
    
  }
  
  
  /**
   * Return the state of the specified index.
   * 
   * @param index The state index.
   * 
   * @return The state name.
   */
  public String getState(int index) {
    
    return states.get(index) ;
    
  }
  
  
  /**
   * Return the index of the specified state.
   * 
   * @param name The state name.
   * 
   * @return The state index.
   */
  public int getStateIndex(String name) {
    
    return states.indexOf(name) ;
    
  }
  
  /**
   * Set the Bottom set for the BDD.
   * 
   * @param state The new bottom state.
   */
  public void setBottomState(String state) {
    bottomState = state ;
  }
  
  /**
   * Return the 'bottom' state in the BDD.
   * 
   * @return The index of the bottom state.
   */
  public int getBottomState() {
    return getStateIndex(bottomState) ;
  }
  
  
  /**
   * Set the error state for the BDD.
   * 
   * @param state The new error state.
   */
  public void setErrorState(String state) {
    errorState = state ;
  }
  
  /**
   * Return the 'error' state in the BDD.
   * 
   * @return The index of the error state.
   */
  public int getErrorState() {
    return getStateIndex(errorState) ;
  }
  
  
  /**
   * Set the accept state for the BDD.
   * 
   * @param state The new acccept state.
   */
  public void setAcceptState(String state) {
    acceptState = state ;
  }
  
  /**
   * Return the 'accept' state in the BDD.
   * 
   * @return The index of the accept state.
   */
  public int getAcceptState() {
    return getStateIndex(acceptState) ;
  }
  
  
  /**
   * Return the numbre of referenced states.
   * 
   * @return The state count.
   */
  public int getStateCount() {
    
    return states.size() ;
    
  }
}
