/*
 * TransducerBddBuilder.java
 */
package bdd;

import automata.* ;

/**
 * Creates a BDD for the transducer rules. 
 * 
 * @author Pierre Salagnac
 *
 * changes made by Adam Rogalewicz
 */
public class TransducerBddBuilder extends BddBuilder<TransducerRule> {
  
  
  /**
   * Creates a new instance.
   * 
   * @param nodeWidth The max node width, used for copying.
   */
  public TransducerBddBuilder(int nodeWidth) {
    
    super(nodeWidth) ;
    
  }
  
  
  /**
   * Build the BDD nodes.
   * 
   * @param rules The automata rules.
   */
  void buildNodes(TransducerRule[] rules) throws BddException {
    
    int nodeWidth = getNodeWidth() ;
    StateIndexer states = getStateIndexer() ;
    
    // error node
    BddNode error = new BddNode(0) ;
    error.addState(states.getErrorState()) ;
    addNode(error) ;
    
    for (TransducerRule rule : rules) {
    
      String left   = rule.getLeftState() ;
      String right  = rule.getRightState() ; 
      String input  = rule.getInput() ;
      String output = rule.getOutput() ;
      String state  = rule.getState() ;
      
      if (input.length() > nodeWidth)
        throw new BddException("rule input too large: "+rule) ;
      if (output != null  &&  output.length() > nodeWidth)
        throw new BddException("rule output too large: "+rule) ;
      
      BddTree tree  = getTree(left,right) ;
      
      int stateIndex = states.getStateIndex(state) ;
      buildNodes(tree,input,output,stateIndex,0,null) ;
    }
  }
  
  
  /**
   * Build the nodes for the specified rule.
   * 
   * @param root The root node.
   * @param input The binary input.
   * @param output The binary output.
   * @param state The output state.
   * @param index The bit index.
   * @param paralel The pointer to the older paralel sequence of BDD nodes.
   */
  private void buildNodes(BddNode root,String input,String output,int state,int index, BddNode paralel)
  throws BddException {
	  /* There can be older sequence of BDD nodes starting from the sam BDD root. This sequence 
	   * is tracked by the pointer paralel. All undefined BDD branches are set to the old sequence.
	   * If we reach a leaf, we take all states from the paralel leaf and add them to the actual
	   * one
	   */
    
    // end
    if (index == getNodeWidth()) {
          root.addState(state) ;
	  if (paralel!=null) {
	  	root.add_list_of_states(paralel.get_all_states());
	}
      return ;
    }
    
    // we copy if the output is two short
    if (output == null  ||  output.length() == 0)
      output = input ;
    
    char inC ;
    char outC ;
    if (index < input.length()) {
      inC  = input.charAt(index) ;
      outC = output.charAt(index) ;
    }
    else {
      inC  = 'X' ;
      outC = 'X' ;
    }
   /* now we have 3 cases
    * 1: paralel=null + both childs of root are "null" -> no paralel sequence
    * 2: paralel=null + some of the childs are not null -> We need to track the paralel sequence
    * 3: paralel!=null -> paralel sequence is already tracked by the pointer paralel
    */
   BddNode falseSon = root.getChild(false);
   BddNode trueSon = root.getChild(true);
   if (paralel!=null) {
   	if (falseSon!=null || trueSon!=null)
		// This line is unreachable in correct run of this method
		throw new BddException("TransducerBddBuilder: buildNodes: 1: This situation may not happend ...") ;
	falseSon = paralel.getChild(false);
	trueSon = paralel.getChild(true);
   }
   // ---------------------------------------------
   // case 1
   if (falseSon==null && trueSon==null) {
	    BddNode inFalseNode = new BddNode(root.getBitIndex()+1) ;
	    BddNode inTrueNode  = new BddNode(root.getBitIndex()+1) ;
	    BddNode outNode     = new BddNode(root.getBitIndex()+2) ;
	    
	    // left
	    if (inC == '0'  ||  inC == 'X') {
	      root.setChild(false,inFalseNode) ;
	      if (outC == '0'  ||  outC == 'X')
	        inFalseNode.setChild(false,outNode) ;
	      else
	        inFalseNode.setChild(true,outNode) ;
	    }
	    
	    // right
	    if (inC == '1'  ||  inC == 'X') {
	      root.setChild(true,inTrueNode) ;
	      if (outC == '1'  ||  outC == 'X')
	        inTrueNode.setChild(true,outNode) ;
	      else
	        inTrueNode.setChild(false,outNode) ;
	    }
	    
	    buildNodes(outNode,input,output,state,index+1,null) ;
   } else {
   // end of case 1
   // ---------------------------------------------
   // case 2 and 3
   // In case 2, we set paralel=root and then continue as in case 3
	   if (paralel==null)
	   	paralel=root;
	
	   // falseS1 - node reachable by in:'0' out: 'outC' in the paralel sequence (if exists)
	   // trueS1 - node reachable by in:'1' out: 'outC' in the paralel sequence (if exists)
	   BddNode falseS1;
	   BddNode trueS1;
	   if (falseSon!=null) {
	   	boolean side = (outC == '1'); // '0' and 'X' has equal behaviour
		falseS1=falseSon.getChild(side);
	   } else
	   	falseS1=null;
	   if (trueSon!=null) {
	   	boolean side = (outC == '1' || outC == 'X'); // '1' and 'X' have equal behaviour
		trueS1=trueSon.getChild(side);
	   } else
	   	trueS1=null;
	   
	    // All undefined edges will be connected to the parralel sequnce
	    BddNode inFalseNode = new BddNode(root.getBitIndex()+1) ;
	    BddNode inTrueNode  = new BddNode(root.getBitIndex()+1) ;
	    BddNode outNode_f;
	    BddNode outNode_t;
	    if (trueS1==falseS1) {
		// Both branches continue in equal direction in paralel sequence
	    	outNode_f     = new BddNode(root.getBitIndex()+2) ;
		outNode_t=outNode_f;
	    } else {
		// The branches in paralel sequence leads in different directions 
		// -> we need also to branch
	    	outNode_f     = new BddNode(root.getBitIndex()+2) ;
	    	outNode_t     = new BddNode(root.getBitIndex()+2) ;
	    }
	    boolean l=false;
	    boolean r=false;
	    // left
	    if (inC == '0'  ||  inC == 'X') {
	      l=true; // mark that this part was executed
	      root.setChild(false,inFalseNode) ;
	      if (outC == '0'  ||  outC == 'X') {
	        inFalseNode.setChild(false,outNode_f) ;
	        if (falseSon!=null) 
	        	inFalseNode.setChild(true,falseSon.getChild(true)) ;
	      } else {
	        inFalseNode.setChild(true,outNode_f) ;
	        if (falseSon!=null) 
	        	inFalseNode.setChild(false,falseSon.getChild(false)) ;
	      }
	    }
	    
	    // right
	    if (inC == '1'  ||  inC == 'X') {
	      r=true; // mark that this part was executed
	      root.setChild(true,inTrueNode) ;
	      if (outC == '1'  ||  outC == 'X') {
	        inTrueNode.setChild(true,outNode_t) ;
	        if (trueSon!=null) 
			inTrueNode.setChild(false,trueSon.getChild(false)) ;
	      } else {
	        inTrueNode.setChild(false,outNode_t) ;
	        if (trueSon!=null) 
	        	inTrueNode.setChild(true,trueSon.getChild(true)) ;
	      }
	    }
	    if (l && (!r)) 
		// true branch is no in actual sequence - set it according the paralel sequence    
	    	root.setChild(true,trueSon);
	    if ((!l) && r)
		// false branch is no in actual sequence - set it according the paralel sequence    
	    	root.setChild(false,falseSon);
	    if ((!l) && (!r)) 
		// the actual sequence does not continue any direction -> something is really odd.    
		throw new BddException("TransducerBddBuilder: buildNodes: 2: This situation may not happend ...") ;
	    if (trueS1==falseS1)
	    	buildNodes(outNode_t,input,output,state,index+1,falseS1) ;
	    else { //paralel sequence is branching => we need to branch here also
	    	if (l) 
			buildNodes(outNode_f,input,output,state,index+1,falseS1) ;
	    	if (r) 
			buildNodes(outNode_t,input,output,state,index+1,trueS1) ;
	    }
	   // end of case 3
	   // ---------------------------------------------
   }
  }
  
}
