/*
 * MonaBuilder.java
 */
package mona ;

import java.io.* ;

import bdd.* ;


/**
 * Generates the Mona output.
 * 
 * @author Pierre Salagnac
 */
public class MonaBuilder {
  
  // all automata states
  private StateIndexer states ;
  
  
  /**
   * Creates a new instance.
   * 
   * @param states The automata states.
   */
  public MonaBuilder(StateIndexer states) {
    
    this.states = states ;
    
  }
  
  
  /**
   * Writes a complete BDD.
   * 
   * @param bdd The BDD to write.
   * @param writer The output.
   * 
   * @exception IOException Error during writing the output.
   */
  public void writeBdd(Bdd bdd,Writer writer)
  throws IOException {
    
    // file header
    writeHeader(writer) ;
    writer.write('\n') ;
    
    // state nodes
    writeStates(writer) ;
    writer.write('\n') ;
    
    // bdd content
    writeAutomata(writer,bdd) ;
    writer.write('\n') ;
    
    // other space
    writeEnd(writer) ;
    writer.write('\n') ;
    
    writer.write("end\n") ;
    
    writer.flush() ;
  }
  
  
  /**
   * Writes the header of the file.
   * 
   * @param writer The output.
   */
  private void writeHeader(Writer writer) throws IOException {
    
    writer.write("MONA GTA\n") ;
    
    writer.write("number of variables: 2\n") ;
    writer.write("state spaces: 3\n") ;
    writer.write("universes: 2\n") ;
    
    //state space sizes
    writer.write("state space sizes: ") ;
    writer.write(String.valueOf(states.getStateCount())) ;
    writer.write(' ') ;
    writer.write(String.valueOf(states.getStateCount())) ;
    writer.write(' ') ;
    writer.write("1\n") ;
    
    // final states
    writer.write("final: ") ;
    writer.write("-1 1 ") ;
    for (int i=2;i<states.getStateCount();i++)
      writer.write("0 ") ;
    writer.write('\n') ;
    
    // guide
    writer.write("guide:\n") ;
    writer.write(" <hat> 1 2\n") ;
    writer.write(" <univ> 1 1\n") ;
    writer.write(" <dummy> 2 2\n") ;
    
    //types
    writer.write("types: 0\n") ;
    
    // universes
    writer.write("universes:\n") ;
    writer.write(" <univ> 0\n") ;
    writer.write(" <dummy> 1\n") ;
    
    // variable orders and state spaces
    writer.write("variable orders and state spaces:\n") ;
    writer.write("P 2: 1\n") ;
    writer.write("Q 2: 1\n") ;
  }
  
  
  /**
   * Writes the mona content for the states.
   * 
   * @param writer The output.
   */
  private void writeStates(Writer writer) throws IOException {
    
    writer.write("state space 0:\n") ;
    writer.write(" initial state: 0\n") ;
    writer.write(" bdd nodes: 2\n") ;
    
    // matrix
    writer.write(" behaviour:\n") ;
    for (int i=0;i<states.getStateCount();i++) {
      writer.write("  ") ;
      if (i == states.getAcceptState())
        writer.write('1') ;
      else
        writer.write('0') ;
      writer.write('\n') ;
    }
    
    // nodes
    writer.write(" bdd:\n") ;
    writer.write("  -1 0 0\n") ;
    writer.write("  -1 1 0\n") ;
    
  }
  
  
  /**
   * Writes the 3rd space.
   * 
   * @param writer The output.
   */
  private void writeEnd(Writer writer) throws IOException {
    
    writer.write("state space 2:\n") ;
    writer.write(" initial state: 0\n") ;
    writer.write(" bdd nodes: 1\n") ;
    
    // matrix
    writer.write(" behaviour:\n") ;
    writer.write("  0\n") ;
    
    // nodes
    writer.write(" bdd:\n") ;
    writer.write("  -1 0 0\n") ;
    
  }
  
  
////////////////////////////////////////////////////////////////////////////////
//                          AUTOMATA                                          //
////////////////////////////////////////////////////////////////////////////////
  
  /**
   * Writes the MONA content for the BDD nodes.
   * 
   * @param writer The output.
   * @param bdd The BDD to write.
   */
  private void writeAutomata(Writer writer,Bdd bdd)
  throws IOException {
    
    writer.write("state space 1:\n") ;
    
    // bottom state
    writer.write("initial state: ") ;
    writer.write(String.valueOf(states.getBottomState())) ;
    writer.write('\n') ;
    
    // node count
    writer.write(" bdd nodes: ") ;
    writer.write(String.valueOf(bdd.getNodeCount())) ;
    writer.write('\n') ;
    
    // initialisation matrix
    writeMatrix(writer,bdd) ;
    
    // node list
    writeNodes(writer,bdd) ;
    
  }
  
  
  /**
   * Writes the matrix of initial nodes.
   * 
   * @param writer The output.
   * @param bdd The BDD to write.
   */
  private void writeMatrix(Writer writer,Bdd bdd)
  throws IOException {
    
    writer.write(" behaviour:\n") ;
    
    int count = states.getStateCount() ;
    
    for (int i=0;i<count;i++) {
      writer.write("  ") ;
      
      for (int j=0;j<count;j++) {
        
        String left  = states.getState(i) ;
        String right = states.getState(j) ;
        BddTree tree = bdd.getTree(left,right) ;
        
        if (tree == null)
          writer.write('0') ;
        else
          writer.write(String.valueOf(tree.getIndex())) ;
        
        writer.write('\t') ;
      }
      
      writer.write('\n') ;
    }
    
  }
  
  
  /**
   * Writes the BDD nodes.
   * 
   * @param writer The output.
   * @param bdd The BDD to write.
   */
  private void writeNodes(Writer writer,Bdd bdd)
  throws IOException {
    
    writer.write(" bdd:\n") ; 
    
    for (int i=0;i<bdd.getNodeCount();i++) {
      
      BddNode node = bdd.getNode(i) ;
      
      writeNode(writer,node) ;
    }
    
  }
  
  
  /**
   * Writes a single node in the output.
   * 
   * @param writer The output.
   * @param node The node to write.
   */
  private void writeNode(Writer writer,BddNode node) throws IOException {
    
    BddNode falseChild = node.getChild(false) ;
    BddNode trueChild  = node.getChild(true) ;
    
    int bit = node.getBitIndex() ;
    
    writer.write(' ') ;
    writer.write(' ') ;
    
    // terminal node
    if (node.getState() != -1) {
      writer.write("-1\t") ;
      writer.write(String.valueOf(node.getState())) ;
      writer.write("\t0\n") ;
      return ;
    }
    
    // input index
    writer.write(String.valueOf(bit)) ;
    writer.write('\t') ;
    
    // false
    if (falseChild != null) {
      int index = falseChild.getIndex() ;
      writer.write(String.valueOf(index)) ;
    }
    else
      writer.write('0') ;
    writer.write('\t') ;
    
    // true
    if (trueChild != null) {
      int index = trueChild.getIndex() ;
      writer.write(String.valueOf(index)) ;
    }
    else
      writer.write('0') ;
    
    writer.write('\n') ;
  }
  
  
  
}
