/*
 * CompoundPath.java
 */
package parser;

import java.util.* ;


/**
 * A path with multiple children.
 * 
 * @author Pierre Salagnac
 */
public abstract class CompoundPath extends Path {

  // children
  private List<Path> children ;
  
  
  /**
   * Create a new instance.
   */
  CompoundPath() {
    
    children = new ArrayList<Path>() ;
    
  }
  
  
  /**
   * Add a subpath to this compound path. The specified path is removed from
   * the previous parent, if exists.
   * 
   * @param path The path to add.
   */
  public void addChild(Path path) {
    
    if (path.equals(this)) throw new Error();
    path.removeFromParent() ;
    
    children.add(path) ;
    path.parent = this ;
    
  }
  
  
  /**
   * Remove the specified subpath. This method has no effect if the path parent
   * is not the current instance.
   * 
   * @param path The path to remove.
   */
  void removeChild(Path path) {
    
    boolean found = children.remove(path) ;
    
    if (found)
      path.parent = null ;
    
  }
  
  
  /**
   * Return the number of children.
   * 
   * @return The children count.
   */
  public int getChildrenCount() {
    return children.size();
  }
  
  
  /**
   * Return a child.
   * 
   * @param index The child index.
   */
  public Path getChild(int index) {
    return children.get(index) ;
  }
  
  /**
   * Return enclosed children.
   * 
   * @return All children.
   */
  Path[] getChildren() {
    
    Path[] array = new Path[children.size()] ;
    
    children.toArray(array) ;
    
    return array ;
    
  }
  
  
  /**
   * Return an equivalent path with the minimum size. The path content
   * can be modified.
   * 
   * @return The minimized path.
   * 
   * @exception PathException The path is invalid, and cannot be minimized.
   */
  abstract Path minimize() throws PathException ;
  
  
  /**
   * Return if this compound path can be finished by a leaf point. A path
   * with disjunction can be finished by a leaf or not.
   * 
   * @return true if the path finish with a leaf.
   */
  public abstract boolean canLeafEnd() ;
  
  /**
   * Return if this compound path can be empty, e.g. no move.
   * 
   * @return true If this path can be empty.
   */
  abstract boolean canBeEmpty() ; 
  
  
  /**
   * Merge absolute points.
   * 
   * @param target A path that should include this one.
   * 
   * @exception PathException There are differents points, that cannot be
   * merged.
   */
  void mergePoints(Path target) throws PathException {
    
    if (target == null)
      return ;
    
    String start       = getStartPoint() ;
    String end         = getEndPoint() ;
    String targetStart = target.getStartPoint() ;
    String targetEnd   = target.getEndPoint() ;
    
    // multiplicity
    if (isMultiple())
      target.setMultiple(true) ;
    
    // start
    if (targetStart == null)
      target.setStartPoint(start) ;
    else if (start != null  &&  !start.equals(targetStart))
      throw new PathException("different start points: "+start+","+targetStart) ;
      
    // end
    if (targetEnd == null)
      target.setEndPoint(end) ;
    else if (end != null  &&  !end.equals(targetEnd))
      throw new PathException("different end points: "+end+","+targetEnd) ;
      
  }
  
  
  /**
   * Return a string describing this path.
   * 
   * @return A string representation of the object.
   */
  @Override
  public String toString() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    buffer.append(getClass().getName()) ;
    buffer.append('[') ;
    
    if (isMultiple())
      buffer.append("multiple,") ;
    
    buffer.append("children={") ;
    for (int i=0;i<children.size();i++) {
      if (i != 0)
        buffer.append(',') ;
      buffer.append(children.get(i)) ;
    }
    buffer.append('}') ;
    
    buffer.append(']') ;
    
    return buffer.toString() ;
  }
  
  
}
