/*
 * Disjunction.java 
 */
package parser;


/**
 * Multiple paths, separeted by a OR.
 * 
 * @author Pierre Salagnac
 */
public class Disjunction extends CompoundPath {
  
  /**
   * Create a new instance.
   */
  public Disjunction() {
    
  }
  
  
  /**
   * Return an equivalent path with the minimum size. The path content
   * can be modified.
   * 
   * @return The minimized path.
   * 
   * @exception PathException The path is invalid, and cannot be minimized.
   */
  Path minimize() throws PathException {
    
    Path[] children = getChildren() ;
    
    // empty
    if (children.length == 0) {
      if (getEndPoint() == null)
        return null ;
      else
        return new Move(null,null,false,false,getEndPoint()) ;
    }
    
    // useless
    else if (children.length == 1) {
      Path unique = children[0] ;
      if (unique instanceof CompoundPath)
        unique = ((CompoundPath)unique).minimize() ;
      mergePoints(unique) ;
      return unique ;
    }
    
    // minimization
    else {
      minimizeDisjunction(this) ;
      if (getChildrenCount() == 0)
        return null ;
      else if (getChildrenCount() == 1)
        return minimize() ;
      else
        return this ;
    }
    
  }
  
  
  /**
   * Do minimization for a disjunction child.
   * 
   * @exception PathException If a minimization fail.
   */
  private void minimizeDisjunction(Disjunction disjunction)
  throws PathException {
    
    if (disjunction.getEndPoint() != null  ||  disjunction.isMultiple()) {
      if (!equals(disjunction)) {
        addChild(disjunction.minimize()) ;
        return ;
      }
    }
    
    Path[] children = disjunction.getChildren() ;
    
    for (int i=0;i<children.length;i++) {
      
      disjunction.removeChild(children[i]) ;
      
      if (children[i] instanceof Disjunction)
        minimizeDisjunction((Disjunction)children[i]) ;
      
      else if (children[i] instanceof CompoundPath) {
        Path min = ((CompoundPath)children[i]).minimize() ;
        if (min != null) {
          if (min instanceof Disjunction)
            minimizeDisjunction((Disjunction)min) ;
          else
            addChild(min) ;
        }
      }
      
      else
        addChild(children[i]) ;
    }
  }
  
  /**
   * Return if this compound path can be finished by a leaf point. A path
   * with disjunction can be finished by a leaf or not.
   * 
   * @return true if the path finish with a leaf.
   */
  public boolean canLeafEnd() {
    
    // disjunction end
    String end = getEndPoint() ;
    if (end != null) {
      if (end.equals("leaf"))
        return true ;
      else
        return false ;
    }
    
    // children ends
    Path[] children = getChildren() ;
    for (int i=children.length-1;i>=0;i--) {
      boolean childEnd ;
      if (children[i] instanceof CompoundPath)
        childEnd = ((CompoundPath)children[i]).canLeafEnd() ;
      else
        childEnd = "leaf".equals(children[i].getEndPoint()) ;
      
      if (childEnd)
        return true ;
    }
    
    return false ;
    
  }
  
  
  /**
   * Return if this compound path can be empty, e.g. no move.
   * 
   * @return true If this path can be empty.
   */
  boolean canBeEmpty() {
    
    if (isMultiple())
      return true ;
    
    boolean empty = false ;
    Path[] children = getChildren() ;
    
    for (Path child : children) {
      
      // compound
      if (child instanceof CompoundPath) {
        CompoundPath compound = (CompoundPath)child ;
        empty = empty | compound.canBeEmpty() ;
      }
      
      // move
      else {
        if (child.isMultiple())
          empty = true ;
      }
    }
    
    return empty ;
    
  }
  
  
  /**
   * Test if this disjunction is between a path and the null value.
   * A disjunction with only null members is not a null disjunction.
   * 
   * <p><strong>Warning: </strong> a minimization must be called before this
   * method.
   * 
   * @return true if the disjunction contains the null route and a normal
   * path.
   */
  public boolean isNullDisjunction() {
    
    boolean hasNull    = false ;
    boolean hasNotNull = false ;
    
    Path[] children = getChildren() ;
    
    for (Path child : children)
      if (child instanceof NullPath)
        hasNull = true ;
      else
        hasNotNull = true ;
    
    return hasNull & hasNotNull ;
    
  }
  
  
  /**
   * Return a disjunction without any null member.
   * 
   * <p><strong>Warning: </strong> a minimization must be called before this
   * method.
   * 
   * @return The path, without the null route.
   */
  public Path removeNull() {
    
    Path[] children = getChildren() ;
    
    for (Path child : children) {
      if (child instanceof NullPath)
        child.removeFromParent() ;
    }
    
    if (getChildrenCount() == 0)
      return null ;
    else if (getChildrenCount() == 1)
      return getChild(0) ;
    else
      return this ;
    
  }
  
  
  /**
   * Clone this disjunction. All subpaths are cloned too.
   * 
   * @return An equivalent disjunction.
   */
  @Override
  public Disjunction clone() {
    
    Disjunction disjunction = new Disjunction() ;
    
    Path[] children = getChildren() ;
    
    for (Path child : children)
      disjunction.addChild(child.clone()) ;
    
    return disjunction ;
  }
  
}
