/*
 * MemberList.java
 */
package parser;

import java.util.* ;


/**
 * All members of the structure.
 * 
 * @author Pierre Salagnac
 */
public class MemberList implements Cloneable {

  // data member
  private String data ;
  
  // members types and routes
  private Map<String,String> members ;
  private Map<String,Path> paths ;
  
  
  /**
   * Create a new instance.
   */
  MemberList() {
    
    members = new HashMap<String,String>() ;
    paths   = new HashMap<String,Path>() ;
    
  }
  
  
  /**
   * Return the data value of this content.
   * 
   * @return The data member, or null if not specified.
   */
  public String getData() {
    return data ;
  }
  
  /**
   * Return all member names.
   * 
   * @return The declared names.
   */
  public String[] getNames() {
    
    String[] array = new String[members.size()] ;
    
    members.keySet().toArray(array) ;
    
    return array ;
    
  }
  
  
  /**
   * Return if the member is primitive.
   * 
   * @param name A member name.
   * 
   * @return true if the member is primitive.
   */
  public boolean isPrimitive(String name) {
    
    Path path = paths.get(name) ;
    
    return (path == null) ;
  }
  
  
  /**
   * Set the data value of this structure content.
   * 
   * @param value The binary data.
   */
  void addData(String value) {
    
    data = value ;
    
  }
  
  
  /**
   * Add a link member to the structure content.
   * 
   * @param type Pointer type.
   * @param name The member name.
   * @param path Le routing path, or null if the member is primitive.
   */
  void addMember(String type,String name,Path path) {
    
    members.put(name,type) ;
    
    if (path != null)
      paths.put(name,path) ;
  }
  
  
  /**
   * Return the type name for the specified member.
   * 
   * @param name A member name.
   * 
   * @return The type of the member.
   */
  public String getType(String name) {
    
    return members.get(name) ;
    
  }
  
  
  /**
   * Put a path for an existing member.
   * 
   * @param member The member name.
   * @param path The path to add to the member.
   */
  public void setPath(String member,Path path) {
    
    if (!members.containsKey(member))
      throw new IllegalArgumentException("unknown member: "+member) ;
    
    if (path == null)
      paths.remove(member) ;
    else
      paths.put(member,path) ;
  }
  
  
  /**
   * Return the path for the specified member.
   * 
   * @param name A member name.
   * 
   * @return The path, or null if the member is primitive.
   */
  public Path getPath(String name) {
    
    return paths.get(name) ;
    
  }
  
  
  /**
   * Clone this member list. All fields are not cloned, but references are
   * kept.
   * 
   * @return A clone of the structure content.
   */
  @Override
  public MemberList clone() {
    
    MemberList clone = new MemberList() ;
    
    clone.data = this.data ;
    clone.members.putAll(this.members) ;
    clone.paths.putAll(this.paths) ;
    
    return clone ;
  }
  
  
  /**
   * Return a string describing this structure content.
   * 
   * @return A string representation of the object.
   */
  @Override
  public String toString() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    buffer.append(getClass().getName()) ;
    buffer.append('[') ;
    
    buffer.append("data=") ;
    buffer.append(data) ;
    buffer.append(',') ;
    
    buffer.append("direct=") ;
    buffer.append(members.size() - paths.size()) ;
    buffer.append(",") ;
    
    buffer.append("path=") ;
    buffer.append(paths.size()) ;
    
    buffer.append(']') ;
    
    return buffer.toString() ;
  }

  
}
