/*
 * ParseHandler.java
 */
package parser ;

import java.util.* ;


/**
 * This class handles events from the parser.
 * 
 * @author Pierre Salagnac
 */
class ParseHandler {

  // list of types and current type
  private List<Typedef> types ;
  private Typedef typedef ;
  private MemberList members ;
  
  // pending member
  private String pendingName ;
  private String pendingType ;
  private Stack<Sequence> pendingPaths ;
  private Stack<Sequence> subPaths ;
  
  // last absolute point
  private String lastPoint ;
  
  
  /**
   * Create a new instance.
   */
  ParseHandler() {
    
    types = new ArrayList<Typedef>() ;
    
    pendingPaths = new Stack<Sequence>() ;
    subPaths     = new Stack<Sequence>() ;
  }
  
  
  /**
   * Return the list of parsed types.
   * 
   * @return The parsed typedefs.
   */
  Typedef[] getTypedefs() {
    
    Typedef[] array = new Typedef[types.size()] ;
    
    types.toArray(array) ;
    
    return array ;
  }
  
  
  /**
   * Open a new typedef declaration.
   * 
   * @param name The name of the structure.
   */
  void addTypedef(String name) {
    
    typedef = null ;
    for (Typedef t : types) {
      if (t.getName().equals(name)) {
        typedef = t ;
        break ;
      }
    }
    
    if (typedef == null) {
      typedef = new Typedef(name) ;
      types.add(typedef) ;
    }
    
    // members
    members = new MemberList() ;
    typedef.addContent(members) ;
    
  }
  
  
  /**
   * Add a variable pointing to the root for the current typedef.
   * 
   * @param variable The name of the variable.
   */
  void addRootVariable(String variable) {
    
    typedef.addRootVariable(variable) ;
    
  }
  
  
  /**
   * Set the data member of the current typedef.
   * 
   * @param value The binary string equals to the default data.
   */
  void addData(String value) {
    
    members.addData(value) ;
    
  }
  
  
  /**
   * Add a link member to the current typedef.
   * 
   * @param type The type of the pointed value.
   * @param name The member name.
   * 
   * @exception PathException The ended path is invalid.
   */
  void addMember(String type,String name) throws PathException {
    
    if (pendingName != null) {
      if (pendingPaths.isEmpty())
        members.addMember(pendingType,pendingName,null) ;
      else {
        Path path = pendingPaths.get(0).minimize() ;
        members.addMember(pendingType,pendingName,path) ;
      }
    }
    
    pendingType = type ;
    pendingName = name ;
    pendingPaths.clear() ;
    subPaths.clear() ;
  }
  
  
////////////////////////////////////////////////////////////////////////////////
//                               ROUTING                                      //
////////////////////////////////////////////////////////////////////////////////
  
  /**
   * Ready the handler to build a new routing expression.
   */
  void startRouting() {
    
    pendingPaths.push(new Sequence()) ;
    
    lastPoint = null ;
    
  }
  
  
  /**
   * Stop the handling of the routing expression. The routing expression
   * is added to the last declared member.
   */
  void endRouting() {
    
    
    
  }
  
  
  /**
   * Add the null route to the expression.
   */
  void addNullPath() {
    
    Sequence parent = pendingPaths.pop() ;
       
    Disjunction disjunction = new Disjunction() ;
    disjunction.addChild(new NullPath()) ;
    disjunction.addChild(parent) ;
    
    if (pendingPaths.isEmpty()) {
      Sequence root = new Sequence() ;
      root.addChild(disjunction) ;
      pendingPaths.push(root) ;
    }
    
    pendingPaths.push(parent) ;
  }
  
  
  /**
   * Add a disjonction in the path. The current path is splitted
   */
  void addDisjunction() {
    
    Sequence parent      = pendingPaths.pop();
    CompoundPath grandParent = parent.getParent() ;
    
    Sequence sequence = new Sequence() ;
    
    Disjunction disjunction = new Disjunction() ;
    disjunction.addChild(parent) ;
    disjunction.addChild(sequence) ;
    
    if (grandParent != null)
      grandParent.addChild(disjunction) ;
    
    if (pendingPaths.isEmpty()) {
      Sequence root = new Sequence() ;
      root.addChild(disjunction) ;
      pendingPaths.push(root) ;
    }
    
    pendingPaths.push(sequence) ;
    
    lastPoint = null ;
  }
  
  
  /**
   * Add a single move to the current path.
   * 
   * @param start A starting position, or null to put no constraint.
   * @param reverse The primitive link is followed to the up.
   * @param name The primitive link.
   * @param multiplicity Multiplicity of the move.
   * @param end An ending position, or null to put no constraint.
   * 
   * @exception PathException If the path is invalid.
   */
  void addMove(String start,boolean reverse,String name,String multiplicity,
      String end)
  throws PathException {
    
    // check absolute points
    if (lastPoint == null)
      lastPoint = start ;
    else if (start != null  &&  !start.equals(lastPoint))
      throw new PathException("different absolute points") ;
    
    if (name != null)
      lastPoint = null ;
    
    if (lastPoint == null)
      lastPoint = end ;
    else if (end != null  &&  !end.equals(lastPoint))
      throw new PathException("different absolute points") ;
    
    
    CompoundPath parent = pendingPaths.peek() ;
    
    // single move
    if (multiplicity == null) {
      Move move = new Move(start,name,reverse,false,end) ;
      parent.addChild(move) ;
    }
    
    // star
    else if (multiplicity.equals("*")) {
      Move move = new Move(start,name,reverse,true,end) ;
      parent.addChild(move) ;
    }
    
    // integer
    else {
      int count = Integer.parseInt(multiplicity) ;
      if (count == 0)
        return ;
      else if (count == 1) {
        Move move = new Move(start,name,reverse,false,end) ;
        parent.addChild(move) ;
      }
      else {
        Move first = new Move(start,name,reverse,false,null) ;
        parent.addChild(first) ;
        for (int i=0;i<count-2;i++) {
          Move move = new Move(null,name,reverse,false,null) ;
          parent.addChild(move) ;
        }
        Move last = new Move(null,name,reverse,false,end) ;
        parent.addChild(last) ;
      }
    }
    
  }
  
  
  /**
   * Start a subpart of the path.
   */
  void startSubpath() {
    
    Sequence parent = pendingPaths.peek();
    
    Sequence subpath = new Sequence() ;
    Sequence content = new Sequence() ;
    
    subpath.addChild(content) ;
    
    parent.addChild(subpath) ;
    
    pendingPaths.push(subpath) ;
    pendingPaths.push(content) ;
    subPaths.push(subpath) ;
  }
  
  
  /**
   * End a subpath of the path.
   * 
   * @param multiplicity Multiplicity of the move.
   * @param end An ending position, or null to put no constraint.
   * 
   * @exception If the path is invalid.
   */
  void endSubpath(String multiplicity,String end)
  throws PathException {
    
    // check absolute points
    if (lastPoint == null)
      lastPoint = end ;
    else if (end != null  &&  !lastPoint.equals(end))
      throw new PathException("different absolute points") ;
    
    // single path
    if (multiplicity == null) {
      Sequence subpath = popSubpath() ;
      subpath.setEndPoint(end) ;
    }
    
    
    // star
    else if (multiplicity.equals("*")) {
      Sequence subpath = popSubpath() ;
      subpath.setMultiple(true) ;
      subpath.setEndPoint(end) ;
    }
    
    // integer
    else {
      int count = Integer.parseInt(multiplicity) ;
      if (count == 0) {
        Sequence subpath = popSubpath() ;
        subpath.removeFromParent() ;
      }
      else if (count == 1)
        endSubpath(null,end) ;
      else {
        Sequence subpath = popSubpath() ;
        Sequence parent  = pendingPaths.peek() ;
        for (int i=0;i<count-2;i++) {
          subpath = subpath.clone() ;
          parent.addChild(subpath) ;
        }
        subpath = subpath.clone() ;
        subpath.setEndPoint(end) ;
        parent.addChild(subpath) ;
      }
    }
    
  }
  
  
  /**
   * Pop the stack at the end of a subpath.
   */
  private Sequence popSubpath() {
    
    Sequence subpath = subPaths.pop() ;
    
    Sequence pending ;
    do {
      pending = pendingPaths.pop() ;
    }
    while (!pending.equals(subpath)) ;
    
    return subpath ;
  }
  
}
