/*
 * Path.java
 */
package parser;


/**
 * Abstract class for a subpart of a routing expression.
 * 
 * @author Pierre Salagnac
 */
public abstract class Path implements Cloneable {

  // the parent path
  CompoundPath parent ;
  
  // starting and ending points
  private String start,end ;
  
  // multiple move
  private boolean multiple ;
  
  
  /**
   * Return the parent path, enclosing this path.
   * 
   * @return The parent path.
   */
  public CompoundPath getParent() {
    
    return parent ;
    
  }
  
  
  /**
   * Remove this path from its parent. This method has no effect if the path
   * has no parent.
   */
  void removeFromParent() {
    
    if (parent == null)
      return ;
    
    parent.removeChild(this) ;
    
    parent = null ;
  }
  
  
  /**
   * Set the start point.
   * 
   * @param start The new start point.
   */
  void setStartPoint(String start) {
    this.start = start ;
  }
  
  /**
   * Return the start point of this move.
   * 
   * @return The start point, or null.
   */
  public String getStartPoint() {
    return start ;
  }
  
  /**
   * Set the end point.
   * 
   * @param end The new end point.
   */
  void setEndPoint(String end) {
    this.end = end ;
  }
  
  /**
   * Return the end point of this move.
   * 
   * @return The end point, or null.
   */
  public String getEndPoint() {
    return end ;
  }
  
  /**
   * Set the multiplicity.
   * 
   * @param multiple true if the path is multiple.
   */
  public void setMultiple(boolean multiple) {
    this.multiple = multiple ;
  }
  
  /**
   * Return if the path is multiple.
   * 
   * @return true if the move is multiple.
   */
  public boolean isMultiple() {
    return multiple ;
  }
  
  
  /**
   * Clone this path. All subpaths has to be cloned too.  The parent path
   * sould be ignored.
   */
  @Override
  public abstract Path clone() ;
  
  
  /**
   * Return a string describing this path.
   * 
   * @return A string representation of the object.
   */
  @Override
  public String toString() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    buffer.append(getClass().getName()) ;
    buffer.append('[') ;
    
    if (multiple)
      buffer.append("multiple") ;
    
    if (start != null) {
      if (multiple)
        buffer.append(',') ;
      buffer.append("start=") ;
      buffer.append(start) ;
    }
    
    if (end != null) {
      if (multiple  ||  start != null)
        buffer.append(',') ;
      buffer.append("end=") ;
      buffer.append(end) ;
    }
    
    buffer.append(']') ;
    
    return buffer.toString() ;
  } 
  
  
////////////////////////////////////////////////////////////////////////////////
//                                FACTORY                                     //
////////////////////////////////////////////////////////////////////////////////
  
  
  /**
   * Create a new path for a primitive link.
   * 
   * @param name The member name.
   */
  public static Path createPrimitivePath(String name) {
    
    return new Move(null,name,false,false,null) ;
    
  }
  
}
