/*
 * Sequence.java 
 */
package parser ;


/**
 * This class is a sequence of one or more path.
 * 
 * @author Pierre Salagnac
 */
public class Sequence extends CompoundPath {

  /**
   * Creates a new instance.
   */
  public Sequence() {
    
  }
  
  
  /**
   * Returns an equivalent path with the minimum size. The path content
   * can be modified.
   * 
   * @return The minimized path.
   * 
   * @exception PathException The path is invalid, and cannot be minimized.
   */
  Path minimize() throws PathException {
    
    Path[] children = getChildren() ;
    
    // empty
    if (children.length == 0) {
      if (getEndPoint() == null)
        return null ;
      else
        return new Move(null,null,false,false,getEndPoint()) ;
    }
    
    // useless
    else if (children.length == 1) {
      Path unique = children[0] ;
      if (unique instanceof CompoundPath) 
        unique = ((CompoundPath)unique).minimize() ;
      mergePoints(unique) ;
      return unique ;
    }
    
    // minimization
    else {
      minimizeSequence(this) ;
      if (getChildrenCount() == 0)
        return null ;
      else if (getChildrenCount() == 1)
        return minimize() ;
      else
        return this ;
    }
    
  }
  
  
  /**
   * Performs minimization for a sequence child.
   * 
   * @exception PathException If a minimization fail.
   */
  private void minimizeSequence(Sequence sequence)
  throws PathException {
    
    if (sequence.getEndPoint() != null  ||  sequence.isMultiple()) {
      if (!equals(sequence)) {
        addChild(sequence.minimize()) ;
        return ;
      }
    }
    
    Path[] children = sequence.getChildren() ;
    
    for (int i=0;i<children.length;i++) {
      
      sequence.removeChild(children[i]) ;
      
      if (children[i] instanceof Sequence)
        minimizeSequence((Sequence)children[i]) ;
      else if (children[i] instanceof CompoundPath){
        Path min = ((CompoundPath)children[i]).minimize() ;
        if (min != null)
          addChild(min) ;
      }
      else
        addChild(children[i]) ;
    }
  }
  
  
  /**
   * Returns if this compound path can be finished by a leaf point. A path
   * with disjunctions can be finished by a leaf or not.
   * 
   * @return true if the path finish with a leaf.
   */
  public boolean canLeafEnd() {
    
    // sequence end
    String end = getEndPoint() ;
    if (end != null) {
      if (end.equals("leaf"))
        return true ;
      else
        return false ;
    }
    
    
    // children ends
    Path[] children = getChildren() ;
    for (int i=children.length-1;i>=0;i--) {
      boolean childEnd ;
      if (children[i] instanceof CompoundPath)
        childEnd = ((CompoundPath)children[i]).canLeafEnd() ;
      else
        childEnd = "leaf".equals(children[i].getEndPoint()) ;
      
      if (childEnd)
        return true ;
      else if (children[i] instanceof CompoundPath) {
        if (!((CompoundPath)children[i]).canBeEmpty())
        return false ;
      }
      else if (!children[i].isMultiple())
        return false ;
    }
    
    return false ;
  }
  
  
  /**
   * Returns if this compound path can be empty, e.g. no move.
   * 
   * @return true If this path can be empty.
   */
  boolean canBeEmpty() {
    
    if (isMultiple())
      return true ;
    
    boolean empty = true ;
    Path[] children = getChildren() ;
    
    for (Path child : children) {
      
      // compound
      if (child instanceof CompoundPath) {
        CompoundPath compound = (CompoundPath)child ;
        empty = empty & compound.canBeEmpty() ;
      }
      
      // move
      else {
        if (!child.isMultiple())
          empty = false ;
      }
    }
    
    return empty ;
  }
  
  
  /**
   * Clones this sequence. All subpaths are cloned too.
   * 
   * @return An equivalent sequence.
   */
  @Override
  public Sequence clone() {
    
    Sequence sequence = new Sequence() ;
    
    Path[] children = getChildren() ;
    
    for (Path child : children)
      sequence.addChild(child.clone()) ;
    
    return sequence ;
  }
  
}
