/*
 * Pattern.java
 */
package tree;

import java.util.* ;


/**
 * This is a type pattern from the type graph.
 * 
 * @author Pierre Salagnac
 */
public class Pattern {
  
  // output name
  private String output ;
  
  // inputs names
  private Input[] inputs ;
  
  // initial data value
  private String data ;
  
  // variables pointing the root
  private String[] rootVariables ;
  
  // F marks
  private List<Integer> marks ;
  
  
  /**
   * Create a new instance.
   * 
   * @param output The name of the output of this patern.
   * @param inputCount
   * @param data The initial data value.
   */
  public Pattern(String output,int inputCount,String data,String[] rootVariables) {
    
    this.output        = output ;
    this.data          = data ;
    this.rootVariables = rootVariables ;
    
    inputs = new Input[inputCount] ;
    
    marks = new ArrayList<Integer>() ;
  }
  
  
  /**
   * Return the output of this pattern.
   * 
   * @return The output type.
   */
  public String getOutput() {
    return output ;
  }
  
  
  /**
   * Return the data of this pattern.
   * 
   * @return The data, or null if the data is not set.
   */
  public String getData() {
    return data ;
  }
  
  
  /**
   * Return the root variables pointing on the type of this pattern.
   * 
   * @return The root variables.
   */
  public String[] getRootVariables() {
    return rootVariables ;
  }
  
  
  /**
   * Return if the pattern has a root variable.
   * 
   * @return true if there is a least a root variable.
   */
  public boolean hasRootVariable() {
    return rootVariables != null  &&  rootVariables.length > 0 ;
  }
  
  
  /**
   * Return if the pattern contains the specified root variable.
   * 
   * @param name The variable name.
   * 
   * @return true if the pattern type is pointed by the variable.
   */
  public boolean containsRootVariable(String name) {
    
    boolean found = false ;
    
    if (rootVariables == null)
      return false ;
    
    for (String root : rootVariables) {
      if (root.equals(name))
        found = true ;
    }
    
    return found ;
    
  }
  
  
  /**
   * Return all inputs from this pattern.
   * 
   * @return The inputs.
   */
  public Input[] getInputs() {
    return inputs.clone();
  }
  
  
  /**
   * Set an input at the specified index.
   * 
   * @param index The input position.
   * @param input The input to add.
   */
  public void setInput(int index,Input input) {
    
    inputs[index] = input ;
    
  }
  
  
  /**
   * Return if this pattern is a leaf.
   * 
   * @return True if this pattern is a leaf.
   */
  boolean isLeaf() {
    
    boolean leaf = true ;
    
    for (Input input : inputs) {
      if (input.isPrimitive())
        leaf = false ;
    }
    
    return leaf ;
  }
  
  
  /**
   * Add a F mark to the pattern.
   * 
   * @param mark The id of the mark.
   */
  void addMark(int mark) {
    
    Integer integer = Integer.valueOf(mark) ;
    
    marks.add(integer) ;
  }
  
  
  /**
   * Return all F mark put on the pattern.
   * 
   * @return The pattern marks.
   */
  public int[] getMarks() {
    
    int[] array = new int[marks.size()] ;
    
    for (int i=0;i<array.length;i++)
      array[i] = marks.get(i).intValue() ;
    
    return array ;
  }
  
  
  /**
   * Return is the specified F mark is specified on this pattern.
   * 
   * @param mark The F mark to search.
   * 
   * @return true if the pattern has the mark.
   */
  public boolean containsMark(int mark) {
    
    Integer integer = Integer.valueOf(mark) ;
    
    return marks.contains(integer) ;
    
  }
  
  
  /**
   * Return a string describing this pattern.
   * 
   * @return A string representation of the object.
   */
  @Override
  public String toString() {
    
    StringBuffer buffer = new StringBuffer() ;
    
    buffer.append(getClass().getName()) ;
    buffer.append('[') ;
    
    buffer.append("data=") ;
    buffer.append(data) ;
    
    buffer.append(",output=") ;
    buffer.append(output) ;
    
    buffer.append(",marks={") ;
    for (int i=0;i<marks.size();i++) {
      if (i != 0)
        buffer.append(',') ;
      buffer.append(marks.get(i)) ;
    }
    buffer.append('}') ;
    
    buffer.append(",roots={") ;
    for (int i=0;i<rootVariables.length;i++) {
      if (i != 0)
        buffer.append(',') ;
      buffer.append(rootVariables[i]) ;
    }
    buffer.append('}') ;
    
    buffer.append(']') ;
    
    return buffer.toString() ;
  }
  
}
