/*
 * PatternBuilder.java 
 */
package tree;

import java.io.* ;
import java.util.* ;

import parser.* ;


/**
 * Build a pattern node from a parsed typedef declaration.
 * 
 * @author Pierre Salagnac
 */
public class PatternBuilder {
  
  // input order
  private ProgramParser program ;
  
  // all inputs
  private String[] inputNames ;
  private boolean[] primitives ;
  
  // all routes
  private RouteIndexer routeIndexer ;
  
  // build patterns
  private List<Pattern> patterns ;
  
  
  /**
   * Create a new instance.
   * 
   * @param program The result of program parse. 
   */
  public PatternBuilder(ProgramParser program) {
    
    this.program = program ;
    
    routeIndexer = new RouteIndexer() ;
    
    patterns = new ArrayList<Pattern>() ;
  }
  
  
  /**
   * Construct the collection of patterns.
   * 
   * @param typedefs All the typedefs.
   * @param filter Filter input names with member list.
   * 
   * @exception PathException If the typedefs can't be converted to patterns.
   */
  public void build(Typedef[] typedefs,boolean filter) throws PathException {
    
    // we split contents with (null+route) path.
    splitNullPaths(typedefs,filter) ;
    
    buildInputNames(typedefs,filter) ;
    buildPrimitiveRoutes() ;
    
    for (Typedef typedef : typedefs) {
      MemberList[] contents = typedef.getContents() ;
      
      for (MemberList content : contents) {
      
        Pattern pattern = new Pattern(
            typedef.getName(),
            filter ? program.getMemberCount() : inputNames.length,
            content.getData(),
            typedef.getRootVariables()
        ) ;
      
        buildInputs(pattern,content,filter) ;
      
        patterns.add(pattern) ;
      }
    }
    
    buildMarks() ;
    
    
    // check of attribute desc_num
    if (filter) {
      int descCount  = program.getDescriptorCount() ;
      int routeCount = routeIndexer.getRouteCount() ;
      int markCount  = getMarkCount() ;
      
      if (routeCount > descCount)
        throw new PathException("desc_num:"+descCount+", routes:"+routeCount) ;
    
      if (markCount  > descCount)  
        throw new PathException("desc_num:"+descCount+" , F-marks:"+markCount) ;
    }    
  }
  
  
  /**
   * Return the built patterns.
   * 
   * @return All built patterns.
   */
  public Pattern[] getPatterns() {
    
    Pattern[] array = new Pattern[patterns.size()] ;
    patterns.toArray(array) ;
    
    return array ;
  }
  
  
  /**
   * Return the indexer with all routes used by the patterns.
   * 
   * @return The route indexer.
   */
  public RouteIndexer getRouteIndexer() {
    return routeIndexer ;
  }
  
  
  /**
   * Return all referenced input names
   * 
   * @return the input names.
   */
  public String[] getInputNames() {
    return inputNames.clone();
  }
  
  
////////////////////////////////////////////////////////////////////////////////
//                         NULL DISJUNCTION                                   //
////////////////////////////////////////////////////////////////////////////////
  /**
   * Split contents with (null or route) paths.
   * 
   * @param typedefs The list of parsed typedefs.
   * @param filter Filter input names.
   */
  private void splitNullPaths(Typedef[] typedefs,boolean filter) {
    
    for (Typedef typedef : typedefs) {
      
      MemberList[] contents = typedef.getContents() ;
      for (MemberList content : contents) {
        
        MemberList[] newContents = splitNullPaths(content,filter) ;
        if (newContents != null) {
          for (MemberList newContent : newContents) {
            if (newContent != content)
            typedef.addContent(newContent) ;
          }
        }
      }
    }
    
  }
  
  
  /**
   * Split contents for a single definition.
   * 
   * @param content The structure content to split.
   * @param filter Filter input names.
   * 
   * @return The contents to add to the typedef.
   */
  private MemberList[] splitNullPaths(MemberList content,boolean filter) {
    
    List<String> nulls = new ArrayList<String>() ;          
    
    // search fo null disjunctions
    String[] members = content.getNames() ;
    
    for (String member : members) {
      if (filter  &&  !program.containsMember(member))
        continue ;
      
      Path path = content.getPath(member) ;
      
      if (path instanceof Disjunction) {
        Disjunction disjunction = (Disjunction)path ;
        if (disjunction.isNullDisjunction()) {
          nulls.add(member) ;
          Path newPath = disjunction.removeNull() ;
          content.setPath(member,newPath) ;
        }
      }
    }
    
    // convertion to array
    String[] array = new String[nulls.size()] ;
    nulls.toArray(array) ;
    
    // duplication
    if (array.length == 0)
      return null ;
    else
      return duplicateNullPaths(0,array,content) ;
    
  }
  
  
  /**
   * Duplicate inputs with recursivity.
   * 
   * @param index The switching member.
   * @param nulls All null members.
   * @param fullContent The base content.  
   */
  private MemberList[] duplicateNullPaths(int index,String[] nulls,MemberList fullContent) {
    
    String member = nulls[index] ;
    
    MemberList nullContent = fullContent.clone() ;
    
    nullContent.setPath(member,new NullPath()) ;
    
    // recursive call
    index++ ;
    if (index < nulls.length) {
      MemberList[] fullArray = duplicateNullPaths(index,nulls,fullContent) ;
      MemberList[] nullArray = duplicateNullPaths(index,nulls,nullContent) ;
      MemberList[] array = new MemberList[fullArray.length+nullArray.length] ;
      System.arraycopy(fullArray,0,array,0,fullArray.length) ;
      System.arraycopy(nullArray,0,array,fullArray.length,nullArray.length) ;
      return array ;
    }
    
    // terminal
    else {
      if (index == 1)
        return new MemberList[]{nullContent} ;
      else
        return new MemberList[]{fullContent,nullContent} ;
    }
    
  }
  
  
////////////////////////////////////////////////////////////////////////////////
//                            INPUTS                                          //
////////////////////////////////////////////////////////////////////////////////
  
  /**
   * Construct an array with all members names found in the structures.
   * 
   * @param typedefs All parsed the typedefs.
   */
  private void buildInputNames(Typedef[] typedefs,boolean filter)
  throws PathException {
    
    Map<String,Boolean> inputs = new TreeMap<String,Boolean>() ;
    
    for (Typedef typedef : typedefs) {
      MemberList[] contents = typedef.getContents() ;
      
      for (MemberList content : contents) {
        String[] members = content.getNames() ;
        
        for (String member : members) {
          
          // input with filter
          if (filter) {
            if (program.containsMember(member)) {
              Boolean primitive = inputs.get(member) ;
              if (primitive == null)
                inputs.put(member,content.isPrimitive(member)) ;
              else
                inputs.put(member,content.isPrimitive(member) | primitive) ;
            }
            else if (content.isPrimitive(member))
              throw new PathException("primitive input not declared: "+member) ;
          }
          
          // input without filter
          else {
            Boolean primitive = inputs.get(member) ;
            if (primitive == null)
              inputs.put(member,content.isPrimitive(member)) ;
            else
              inputs.put(member,content.isPrimitive(member) | primitive) ;
          }
        }
      }
    }
    
    // convertion to array
    inputNames = new String[inputs.size()] ;
    inputs.keySet().toArray(inputNames) ;
    
    primitives = new boolean[inputs.size()] ;
    for (int i=0;i<primitives.length;i++)
      primitives[i] = inputs.get(inputNames[i]).booleanValue() ;    
  }
  
  
  /**
   * Add the primitives routes to the indexer.
   */
  private void buildPrimitiveRoutes() {
    
    int index = 0 ;
    
    for (int i=0;i<inputNames.length;i++) {
      
      if (primitives[i]) {
        
        Path move = Path.createPrimitivePath(inputNames[i]) ;
        
        Route route = new Route(inputNames[i],move) ;
        routeIndexer.setRoute(index,route) ;
        index++ ;
      }
    }
    
  }
  
  
  /**
   * Fill inputs in the specified pattern.
   * 
   * @param pattern The pattern to fill.
   * @param content The parsed content of the pattern.
   * @param boolean filter Filter input names.
   */
  private void buildInputs(Pattern pattern,MemberList content,boolean filter) {
    
    // undefined everywhere
    if (filter) {
      for (int i=0;i<program.getMemberCount();i++) {
        String name = program.getMember(i) ;
        pattern.setInput(i,Input.createUndefinedInput(name)) ;
      }
    }
    else {
      for (int i=0;i<inputNames.length;i++) {
        pattern.setInput(i,Input.createUndefinedInput(inputNames[i])) ;
      }
    }
    
    // defined inputs
    String[] members = content.getNames() ;
    for (int i=0;i<members.length;i++) {
      String member = members[i] ;
      
      Input input ;
      Path path = content.getPath(member) ;
      
      // primitive
      if (path == null) {
        String type = content.getType(member) ;
        int route = routeIndexer.getPrimitiveIndex(member) ;
        input = Input.createPrimitiveInput(member,type,route) ;
      }
      
      // routed
      else {
        String type = content.getType(member) ;
        if (path instanceof NullPath)
          input = Input.createNullInput(member,type) ;
        else {
          int index = routeIndexer.getAvailableIndex() ;
          Route route = new Route(member,path) ;
          routeIndexer.setRoute(index,route) ;
          input = Input.createRoutedInput(member,type,index) ;
        }
      }
      
      // we ignore undeclared inputs (keep undefined)
      if (filter) {
        int index = program.getMemberIndex(member) ;
        if (index >= 0)
          pattern.setInput(index,input) ;
      }
      else
        pattern.setInput(i,input) ;
    }
    
  }
  
////////////////////////////////////////////////////////////////////////////////
//                              MARKS                                         //
////////////////////////////////////////////////////////////////////////////////
  
  // next mark to be used
  private int nextMark ;
  
  
  /**
   * Build the F marks for all routes.
   */
  private void buildMarks() {
    
    nextMark = 0 ;
    
    buildPrimitiveMarks() ;
    
    buildLeafMarks() ;
    
    buildTypeMarks() ;
  }
  
  
  /**
   * Build the primitive marks.
   */
  private void buildPrimitiveMarks() {
    
    // mark end of primitives routes
    for (Pattern pattern : patterns) {
      Input[] inputs = pattern.getInputs() ;
      for (Input input : inputs) {
        if (input.isPrimitive()) {
          Route route = routeIndexer.getRoute(input.getRoute()) ;
          route.addMark(nextMark) ;
        }
      }
    }
    
    // put primitive mark everywhere
    for (Pattern pattern : patterns)
      pattern.addMark(nextMark) ;
    
    nextMark++ ;
  }
  
  
  /**
   * Build marks for route ending with a leaf.
   */
  private void buildLeafMarks() {
    
    // put leaf mark at leaf-ending routes
    for (Pattern pattern : patterns) {
      
      Input[] inputs = pattern.getInputs() ;
      for (Input input : inputs) {
        int index = input.getRoute() ;
        
        if (index < 0)
          continue ;
        Route route = routeIndexer.getRoute(index) ;
        Path path = route.getPath() ;
        
        if (path instanceof CompoundPath) {
          CompoundPath compound = (CompoundPath)path ;
          if (compound.canLeafEnd())
            route.addMark(nextMark) ;
        }
        else {
          String end = path.getEndPoint() ;
          if (end != null  &&  end.equals("leaf"))
            route.addMark(nextMark) ;
        }
        
      }
    }
    
    // put leaf mark on each leaf pattern
    for (Pattern pattern : patterns) {
      if (pattern.isLeaf())
        pattern.addMark(nextMark) ;
    }
    
    nextMark++ ;
  }
  
  
  /**
   * Build marks based on types.
   */
  private void buildTypeMarks() {
    
    // creating list of all types
    Set<String> types = new HashSet<String>() ;
    for (Pattern pattern : patterns)
      types.add(pattern.getOutput()) ;
    
    for (String type : types) {
      boolean markPatterns = false ;
      
      // mark routes ending with the type
      for (Pattern pattern : patterns) {
        Input[] inputs = pattern.getInputs() ;
        for (Input input : inputs) {
          int index = input.getRoute() ;
          if (index < 0  ||  !type.equals(input.getType()))
            continue ;
          Route route = routeIndexer.getRoute(index) ;
          route.addMark(nextMark) ;
          markPatterns = true ;
        }
      }
      
      // mark patterns if needed
      if (markPatterns) {
        for (Pattern pattern : patterns) {
          if (type.equals(pattern.getOutput()))
            pattern.addMark(nextMark) ;
        }
        nextMark++ ;
      }
    }
    
  }
  
  
  /**
   * Return the number of different F marks used in the patterns and in the
   * routes.
   * 
   * @return The F mark count.
   */
  private int getMarkCount() {
    
    Set<Integer> set = new TreeSet<Integer>() ;
    
    for (Pattern pattern : patterns) {
      
      int[] marks = pattern.getMarks() ;
      for (int mark : marks)
        set.add(Integer.valueOf(mark)) ;
    }
    
    return set.size() ;
  }
  
}
