/*
 * TestTransducerBuilder.java
 */
package automata;

import junit.framework.* ;

import tree.* ;
import parser.* ;


/**
 * Test the route transformation.
 * 
 * @author Pierre Salagnac
 */
public class TestTransducerBuilder extends TestCase {
  
  
  /**
   * Tests that the specified rule is good.
   * 
   * @param left The expected left state.
   * @param right The expected right state.
   * @param state The expected out state.
   */
  private void assertRule(TransducerRule rule,String left,String right,String state) {
    
    String ruleLeft  = rule.getLeftState() ;
    String ruleRight = rule.getRightState() ;
    String ruleState = rule.getState() ;
    
    assertEquals("left",left,ruleLeft) ;
    assertEquals("right",right,ruleRight) ;
    assertEquals("state",state,ruleState) ;
    
  }
  
  
  /**
   * Tests that the specified rule is good for a type 1.
   * 
   * @param rule The rule to test.
   */
  private void assertRule1(TransducerRule rule) {
    
    String input  = rule.getInput() ;
    String output = rule.getOutput() ;
    
    assertEquals('0',input.charAt(0)) ;
    assertEquals('0',input.charAt(0)) ;
    
    assertNull(output) ;
  }
  
  /**
   * Tests that the specified rule is good for a type 3.
   * 
   * @param rule The rule to test.
   * @param left The expected left state.
   * @param right The expected right state.
   * @param state The expected out state.
   */
  private void assertRule3(TransducerRule rule,String left,String right,String state) {
    
    String input = rule.getInput() ;
    assertEquals('1',input.charAt(0)) ;
    
    assertRule(rule,left,right,state) ;
  }
  
  
  /**
   * Tests that the specified rule is good for a type 3.
   * 
   * @param rule The rule to test.
   * @param left The expected left state.
   * @param right The expected right state.
   * @param state The expected out state.
   */
  private void assertRule4(TransducerRule rule,String left,String right,String state) {
    
    String input = rule.getInput() ;
    assertEquals('0',input.charAt(0)) ;
    
    assertRule(rule,left,right,state) ;
  }
  
  
  /**
   * Tests the starting type 3 rule.
   * 
   * @param rule The rule to test.
   * @param left The expected left state.
   * @param right The expected right state.
   * @param state The expected out state.
   */
  private void assertRuleS(TransducerRule rule,String left,String right,String state) {
    
    boolean found = false ;
    String input  = rule.getInput() ;
    String output = rule.getOutput() ;
    
    for (int i=0;i<input.length();i++) {
      char cin  = input.charAt(i) ;
      char cout = output.charAt(i) ;
      if (cin == '1'  &&  cout =='0') {
        found = true ;
        break ;
      }
    }
    
    if (!found)
      fail("start token not found") ;
    
    assertRule3(rule,left,right,state) ;
    
  }
  
  
  /**
   * Tests the endinf type 3 rule.
   * 
   * @param rule The rule to test.
   * @param left The expected left state.
   * @param right The expected right state.
   * @param state The expected out state.
   */
  private void assertRuleE(TransducerRule rule,String left,String right,String state) {
    
    boolean found = false ;
    String input  = rule.getInput() ;
    String output = rule.getOutput() ;
    
    for (int i=0;i<input.length();i++) {
      char cin  = input.charAt(i) ;
      char cout = output.charAt(i) ;
      if (cin == '0'  &&  cout =='1') {
        found = true ;
        break ;
      }
    }
    
    if (!found)
      fail("start token not found") ;
    
    assertRule3(rule,left,right,state) ;
    
  }
  
  
  /**
   * Transforms an unique going down move.
   */
  public void testUnique() throws Exception {
    
    Move move = new Move(null,"move",false,false,null) ;
    Route route = new Route("move",move) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","move:0","x:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A2",builder.getAcceptState()) ;
    assertEquals(8,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleE(rules[3],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4],"A1"             ,Rule.BOTTOM_STATE,"A3") ;
    assertRuleS(rules[5],"A3"             ,Rule.BOTTOM_STATE,"A2") ;
    assertRule (rules[6],Rule.BOTTOM_STATE,"A2"             ,"A2") ;
    assertRule (rules[7],"A2"             ,Rule.BOTTOM_STATE,"A2") ;
  }
  
  
  /**
   * Transform an unique going up move.
   */
  public void testReversedUnique() throws Exception {
    
    Move move = new Move(null,"move",true,false,null) ;
    Route route = new Route("move",move) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","x:1 move:0","x:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A2",builder.getAcceptState()) ;
    assertEquals(8,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleS(rules[3],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4],"A1"             ,Rule.BOTTOM_STATE,"A3") ;
    assertRuleE(rules[5],"A3"             ,Rule.BOTTOM_STATE,"A2") ;
    assertRule (rules[6],Rule.BOTTOM_STATE,"A2"             ,"A2") ;
    assertRule (rules[7],"A2"             ,Rule.BOTTOM_STATE,"A2") ;
  }
  
  
  /**
   * Transforms an unique going down move with a star.
   */
  public void testMultipleUnique() throws Exception {
    
    Move move = new Move(null,"move",false,true,null) ;
    Route route = new Route("move",move) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","x:1 move:0","x:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A1",builder.getAcceptState()) ;
    assertEquals(8,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleE(rules[3],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4],"A1"             ,Rule.BOTTOM_STATE,"A2") ;
    assertRuleS(rules[5],"A2"             ,Rule.BOTTOM_STATE,"A1") ;
    assertRule (rules[6],Rule.BOTTOM_STATE,"A1"             ,"A1") ;
    assertRule (rules[7],"A1"             ,Rule.BOTTOM_STATE,"A1") ;
  }
   
  
  /**
   * Transforms an unique going down move with a star.
   */
  public void testMultipleReversedUnique() throws Exception {
    
    Move move = new Move(null,"move",true,true,null) ;
    Route route = new Route("move",move) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","x:1 move:0","x:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A1",builder.getAcceptState()) ;
    assertEquals(8,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleS(rules[3],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4],"A1"             ,Rule.BOTTOM_STATE,"A2") ;
    assertRuleE(rules[5],"A2"             ,Rule.BOTTOM_STATE,"A1") ;
    assertRule (rules[6],Rule.BOTTOM_STATE,"A1"             ,"A1") ;
    assertRule (rules[7],"A1"             ,Rule.BOTTOM_STATE,"A1") ;
  }
  
  
  /**
   * Transforms two moves going down.
   */
  public void testDouble() throws Exception {
    
    Sequence sequence = new Sequence() ;
    sequence.addChild(Path.createPrimitivePath("left")) ;
    sequence.addChild(Path.createPrimitivePath("right")) ;
        
    Route route = new Route("sequence",sequence) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","left:0 other:1 right:2","x:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A6",builder.getAcceptState()) ;
    assertEquals(12,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleE(rules[3] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4] ,"A1"             ,Rule.BOTTOM_STATE,"A3") ;
    assertRule4(rules[5] ,Rule.BOTTOM_STATE,"A3"             ,"A4") ;
    assertRule4(rules[6] ,Rule.BOTTOM_STATE,"A4"             ,"A2") ;
    assertRule3(rules[7] ,"A2"             ,Rule.BOTTOM_STATE,"A5") ;
    assertRule4(rules[8] ,"A5"             ,Rule.BOTTOM_STATE,"A7") ;
    assertRuleS(rules[9] ,"A7"             ,Rule.BOTTOM_STATE,"A6") ;
    assertRule (rules[10],Rule.BOTTOM_STATE,"A6"             ,"A6") ;
    assertRule (rules[11],"A6"             ,Rule.BOTTOM_STATE,"A6") ;
  }
  
  
  /**
   * Transforms two moves going up.
   */
  public void testReversedDouble() throws Exception {
    
    Sequence sequence = new Sequence() ;
    sequence.addChild(new Move(null,"left",true,false,null)) ;
    sequence.addChild(new Move(null,"right",true,false,null)) ;
        
    Route route = new Route("sequence",sequence) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","right:0 other:1 left:2","x:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A6",builder.getAcceptState()) ;
    assertEquals(12,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleS(rules[3] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4] ,"A1"             ,Rule.BOTTOM_STATE,"A3") ;
    assertRule4(rules[5] ,Rule.BOTTOM_STATE,"A3"             ,"A4") ;
    assertRule4(rules[6] ,Rule.BOTTOM_STATE,"A4"             ,"A2") ;
    assertRule3(rules[7] ,"A2"             ,Rule.BOTTOM_STATE,"A5") ;
    assertRule4(rules[8] ,"A5"             ,Rule.BOTTOM_STATE,"A7") ;
    assertRuleE(rules[9] ,"A7"             ,Rule.BOTTOM_STATE,"A6") ;
    assertRule (rules[10],Rule.BOTTOM_STATE,"A6"             ,"A6") ;
    assertRule (rules[11],"A6"             ,Rule.BOTTOM_STATE,"A6") ;
  }
  
  
  /**
   * Transforms two moves going down with a star on each.
   */
/*  public void testMultipleDouble() throws Exception {
    
    Sequence sequence = new Sequence() ;
    sequence.addChild(new Move(null,"left",false,true,null)) ;
    sequence.addChild(new Move(null,"right",false,true,null)) ;
    
    Route route = new Route("sequence",sequence) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","right:0 left:1","x:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    //assertEquals("TODO",builder.getAcceptState()) ;
    //assertEquals(9,rules.length) ;
    assertRule1(rules[0]) ;
    
  }
  */
  
  /**
   * Transforms two moves going up and down.
   */
  public void testUpDown() throws Exception {
    
    Sequence sequence = new Sequence() ;
    sequence.addChild(new Move(null,"left",true,false,null)) ;
    sequence.addChild(new Move(null,"right",false,false,null)) ;
        
    Route route = new Route("sequence",sequence) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","left:1 other:0 right:2","x:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
        
    assertEquals("A4",builder.getAcceptState()) ;
    assertEquals(11,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleS(rules[3] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRuleE(rules[4] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A3") ;
    assertRule4(rules[5] ,"A3"             ,Rule.BOTTOM_STATE,"A6") ;
    assertRule4(rules[6] ,"A1"             ,"A6"             ,"A7") ;
    assertRule4(rules[7] ,Rule.BOTTOM_STATE,"A7"             ,"A5") ;
    assertRule3(rules[8] ,"A5"             ,Rule.BOTTOM_STATE,"A4") ;
    assertRule (rules[9] ,Rule.BOTTOM_STATE,"A4"             ,"A4") ;
    assertRule (rules[10],"A4"             ,Rule.BOTTOM_STATE,"A4") ;
  }
  
  
  /**
   * Transforms a sequence, top and down.
   */
  public void testSequence() throws Exception {
    
    Sequence sequence = new Sequence() ;
    sequence.addChild(new Move(null,"right",true,false,null)) ;
    sequence.addChild(new Move(null,"left",true,false,null)) ;
    sequence.addChild(Path.createPrimitivePath("right")) ;
    sequence.addChild(Path.createPrimitivePath("left")) ;
        
    Route route = new Route("sequence",sequence) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","u:0 left:1 t:2 right:3","y:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A12",builder.getAcceptState()) ;
    assertEquals(20,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleS(rules[3] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4] ,"A1"             ,Rule.BOTTOM_STATE,"A3") ;
    assertRule4(rules[5] ,Rule.BOTTOM_STATE,"A3"             ,"A4") ;
    assertRule4(rules[6] ,Rule.BOTTOM_STATE,"A4"             ,"A5") ;
    assertRule4(rules[7] ,Rule.BOTTOM_STATE,"A5"             ,"A2") ;
    assertRule3(rules[8] ,"A2"             ,Rule.BOTTOM_STATE,"A6") ;
    assertRuleE(rules[9] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A8") ;
    assertRule4(rules[10] ,"A8"             ,Rule.BOTTOM_STATE,"A10") ;
    assertRule4(rules[11],Rule.BOTTOM_STATE,"A10"             ,"A9") ;
    assertRule3(rules[12],"A9"             ,Rule.BOTTOM_STATE,"A11") ;
    assertRule4(rules[13],"A11"            ,Rule.BOTTOM_STATE,"A14") ;
    assertRule4(rules[14],Rule.BOTTOM_STATE,"A14"            ,"A15") ;
    assertRule4(rules[15],"A6"             ,"A15"            ,"A16") ;
    assertRule4(rules[16],Rule.BOTTOM_STATE,"A16"            ,"A13") ;
    assertRule3(rules[17],"A13"            ,Rule.BOTTOM_STATE,"A12");
    assertRule (rules[18] ,Rule.BOTTOM_STATE,"A12"             ,"A12") ;
    assertRule (rules[19],"A12"             ,Rule.BOTTOM_STATE,"A12") ;
  }
  
  
  /**
   * Test a sequence going down with a star.
   */
  public void testMultipleSequence() throws Exception {
    
    Sequence sequence = new Sequence() ;
    sequence.addChild(new Move(null,"right",false,false,null)) ;
    sequence.addChild(new Move(null,"left",false,false,null)) ;
    sequence.setMultiple(true) ;
    Route route = new Route("sequence",sequence) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","left:0 right:1","y:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A1",builder.getAcceptState()) ;
    assertEquals(11,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleE(rules[3] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4] ,"A1"             ,Rule.BOTTOM_STATE,"A2") ;
    assertRule3(rules[5] ,"A2"             ,Rule.BOTTOM_STATE,"A3") ;
    assertRule4(rules[6] ,"A3"             ,Rule.BOTTOM_STATE,"A5") ;
    assertRule4(rules[7] ,Rule.BOTTOM_STATE,"A5"             ,"A4") ;
    assertRuleS(rules[8] ,"A4"             ,Rule.BOTTOM_STATE,"A1") ;
    assertRule (rules[9] ,Rule.BOTTOM_STATE,"A1"             ,"A1") ;
    assertRule (rules[10],"A1"             ,Rule.BOTTOM_STATE,"A1") ;
  }
  
  
  /**
   * Transforms a simple disjunction, going down.
   */
  public void testDisjunction() throws Exception {
    
    Disjunction disjunction = new Disjunction() ;
    disjunction.addChild(new Move(null,"left",false,false,null)) ;
    disjunction.addChild(new Move(null,"right",false,false,null)) ;
    
    Route route = new Route("disjunction",disjunction) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","left:0 right:1","y:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A3",builder.getAcceptState()) ;
    assertEquals(9,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleE(rules[3],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4],"A1"             ,Rule.BOTTOM_STATE,"A5") ;
    assertRule4(rules[5],"A1"             ,"A5"             ,"A4") ;
    assertRuleS(rules[6],"A4"             ,Rule.BOTTOM_STATE,"A3") ;
    assertRule (rules[7],Rule.BOTTOM_STATE,"A3"             ,"A3") ;
    assertRule (rules[8],"A3"             ,Rule.BOTTOM_STATE,"A3") ;
    
  }
  
  
  /**
   * Transforms a simple disjunction, going up.
   */
  public void testRevertedDisjunction() throws Exception {
    
    Disjunction disjunction = new Disjunction() ;
    disjunction.addChild(new Move(null,"left",true,false,null)) ;
    disjunction.addChild(new Move(null,"right",true,false,null)) ;
    
    Route route = new Route("disjunction",disjunction) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","left:0 right:1","y:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A3",builder.getAcceptState()) ;
    assertEquals(9,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleS(rules[3],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4],"A1"             ,Rule.BOTTOM_STATE,"A5") ;
    assertRule4(rules[5],"A1"             ,"A5"             ,"A4") ;
    assertRuleE(rules[6],"A4"             ,Rule.BOTTOM_STATE,"A3") ;
    assertRule (rules[7],Rule.BOTTOM_STATE,"A3"             ,"A3") ;
    assertRule (rules[8],"A3"             ,Rule.BOTTOM_STATE,"A3") ;
  }
  
  
  /**
   * Transforms a going down disjunction with a star.
   */
  public void testMultipleDisjnuction() throws Exception {
    
    Disjunction disjunction = new Disjunction() ;
    disjunction.addChild(new Move(null,"left",false,false,null)) ;
    disjunction.addChild(new Move(null,"right",false,false,null)) ;
    disjunction.setMultiple(true) ;
    
    Route route = new Route("disjunction",disjunction) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","left:0 right:1","y:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
        
    assertEquals("A1",builder.getAcceptState()) ;
    assertEquals(9,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleE(rules[3],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4],"A1"             ,Rule.BOTTOM_STATE,"A4") ;
    assertRule4(rules[5],"A1"             ,"A4"             ,"A3") ;
    assertRuleS(rules[6],"A3"             ,Rule.BOTTOM_STATE,"A1") ;
    assertRule (rules[7],Rule.BOTTOM_STATE,"A1"             ,"A1") ;
    assertRule (rules[8],"A1"             ,Rule.BOTTOM_STATE,"A1") ;
  }
  
  
  /**
   * Transforms a sequence with two disjunctions, going down.
   */
  public void testDisjunctionSequence() throws Exception {
    
    Disjunction disjunction1 = new Disjunction() ;
    disjunction1.addChild(new Move(null,"left",false,false,null)) ;
    disjunction1.addChild(new Move(null,"right",false,false,null)) ;
    
    Disjunction disjunction2 = new Disjunction() ;
    disjunction2.addChild(new Move(null,"right",false,false,null)) ;
    disjunction2.addChild(new Move(null,"left",false,false,null)) ;
    
    Sequence sequence = new Sequence() ;
    sequence.addChild(disjunction1) ;
    sequence.addChild(disjunction2) ;
    
    Route route = new Route("sequence",sequence) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","left:0 right:1","y:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A6",builder.getAcceptState()) ;
    assertEquals(13,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleE(rules[3],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4],"A1"             ,Rule.BOTTOM_STATE,"A3") ;
    assertRule4(rules[5],Rule.BOTTOM_STATE,"A3"             ,"A2") ;
    assertRule4(rules[6],"A1"             ,Rule.BOTTOM_STATE,"A2") ;
    assertRule3(rules[7],"A2"             ,Rule.BOTTOM_STATE,"A4") ;
    assertRule4(rules[8],"A4"             ,Rule.BOTTOM_STATE,"A8") ;
    assertRule4(rules[9],"A4"             ,"A8"             ,"A7") ;
    assertRuleS(rules[10],"A7"             ,Rule.BOTTOM_STATE,"A6") ;
    assertRule (rules[11],Rule.BOTTOM_STATE,"A6"             ,"A6") ;
    assertRule (rules[12],"A6"             ,Rule.BOTTOM_STATE,"A6") ;
    
  }
  
  /**
   * Transforms a sequence with two disjunctions, going up.
   */
  public void testRevertedDisjunctionSequence() throws Exception {
    
    Disjunction disjunction1 = new Disjunction() ;
    disjunction1.addChild(new Move(null,"left",true,false,null)) ;
    disjunction1.addChild(new Move(null,"right",true,false,null)) ;
    
    Disjunction disjunction2 = new Disjunction() ;
    disjunction2.addChild(new Move(null,"right",true,false,null)) ;
    disjunction2.addChild(new Move(null,"left",true,false,null)) ;
    
    Sequence sequence = new Sequence() ;
    sequence.addChild(disjunction1) ;
    sequence.addChild(disjunction2) ;
    
    Route route = new Route("sequence",sequence) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","left:0 right:1","y:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A6",builder.getAcceptState()) ;
    assertEquals(13,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleS(rules[3],Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4],"A1"             ,Rule.BOTTOM_STATE,"A2") ;
    assertRule4(rules[5],"A1"             ,Rule.BOTTOM_STATE,"A3") ;
    assertRule4(rules[6],Rule.BOTTOM_STATE,"A3"             ,"A2") ;
    assertRule3(rules[7],"A2"             ,Rule.BOTTOM_STATE,"A4") ;
    assertRule4(rules[8],"A4"             ,Rule.BOTTOM_STATE,"A8") ;
    assertRule4(rules[9],"A4"             ,"A8"             ,"A7") ;
    assertRuleE(rules[10],"A7"             ,Rule.BOTTOM_STATE,"A6") ;
    assertRule (rules[11],Rule.BOTTOM_STATE,"A6"             ,"A6") ;
    assertRule (rules[12],"A6"             ,Rule.BOTTOM_STATE,"A6") ;
    
  }
  
  
  /**
   * Transforms a sequence with two disjunctions, going up and down.
   */
  public void testUpDownDisjunctionSequence() throws Exception {
    
    Disjunction disjunction1 = new Disjunction() ;
    disjunction1.addChild(new Move(null,"left",true,false,null)) ;
    disjunction1.addChild(new Move(null,"right",true,false,null)) ;
    
    Disjunction disjunction2 = new Disjunction() ;
    disjunction2.addChild(new Move(null,"right",false,false,null)) ;
    disjunction2.addChild(new Move(null,"left",false,false,null)) ;
    
    Sequence sequence = new Sequence() ;
    sequence.addChild(disjunction1) ;
    sequence.addChild(disjunction2) ;
    
    Route route = new Route("sequence",sequence) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","left:0 right:1","y:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A6",builder.getAcceptState()) ;
    assertEquals(12,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleS(rules[3] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRuleE(rules[4] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A4") ;
    assertRule4(rules[5] ,"A1"             ,Rule.BOTTOM_STATE,"A8") ;
    assertRule4(rules[6] ,"A1"             ,"A8"             ,"A7") ;
    assertRule4(rules[7] ,"A4"             ,Rule.BOTTOM_STATE,"A9") ;
    assertRule4(rules[8] ,"A4"             ,"A9"             ,"A7") ;
    assertRule3(rules[9] ,"A7"             ,Rule.BOTTOM_STATE,"A6") ;
    assertRule (rules[10],Rule.BOTTOM_STATE,"A6"             ,"A6") ;
    assertRule (rules[11],"A6"             ,Rule.BOTTOM_STATE,"A6") ;
    
  }
  
  
  /**
   * Transforms a disjunction, with sequences inside, going down
   */
  public void testSequenceDisjunction() throws Exception {
    
    Sequence sequence1 = new Sequence() ;
    sequence1.addChild(new Move(null,"left",false,false,null)) ;
    sequence1.addChild(new Move(null,"right",false,false,null)) ;
    
    Sequence sequence2 = new Sequence() ;
    sequence2.addChild(new Move(null,"right",false,false,null)) ;
    sequence2.addChild(new Move(null,"left",false,false,null)) ;
    
    Disjunction disjunction = new Disjunction() ;
    disjunction.addChild(sequence1) ;
    disjunction.addChild(sequence2) ;
    
    Route route = new Route("disjunction",disjunction) ;
    
    ProgramParser program = new MockProgramParser(10,2,"101","left:0 right:1","y:1") ;
    
    TransducerBuilder builder = new TransducerBuilder(program) ;
    builder.build(route) ;
    
    TransducerRule[] rules = builder.getRules() ;
    
    assertEquals("A8",builder.getAcceptState()) ;
    assertEquals(14,rules.length) ;
    assertRule1(rules[0]) ;
    assertRule3(rules[1] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRule4(rules[2] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE) ;
    assertRuleE(rules[3] ,Rule.BOTTOM_STATE,Rule.BOTTOM_STATE,"A1") ;
    assertRule4(rules[4] ,"A1"             ,Rule.BOTTOM_STATE,"A4") ;
    assertRule4(rules[5] ,Rule.BOTTOM_STATE,"A4"             ,"A3") ;
    assertRule3(rules[6] ,"A3",Rule.BOTTOM_STATE,"A5") ;
    assertRule4(rules[7] ,"A1"                  ,Rule.BOTTOM_STATE,"A6") ;
    assertRule3(rules[8] ,"A6",Rule.BOTTOM_STATE,"A7") ;
    assertRule4(rules[9] ,"A7"                  ,Rule.BOTTOM_STATE,"A10") ;
    assertRule4(rules[10] ,"A5"             ,"A10"             ,"A9") ;
    assertRuleS(rules[11],"A9"             ,Rule.BOTTOM_STATE,"A8") ;
    assertRule (rules[12],Rule.BOTTOM_STATE,"A8"             ,"A8") ;
    assertRule (rules[13],"A8"             ,Rule.BOTTOM_STATE,"A8") ;
    
  }
  
  
}
