/*
 * TestBddNode.java
 */
package bdd;

import junit.framework.* ;

/**
 * Test expansion of nodes
 * 
 * @author Pierre Salagnac
 */
public class TestBddNode extends TestCase {

  
  /**
   * Test move of not-leave states.
   */
  public void testExpandNotLeaf() throws BddException {
    
    BddNode node = new BddNode(0) ;
    node.addState(0) ;
    
    node.setChild(false,new BddNode(1)) ;
    
    node.expandNotLeaf(5) ;
    
    BddNode falseChild = node.getChild(false) ;
    BddNode trueChild  = node.getChild(true) ;
    
    assertNotNull(falseChild) ;
    assertNotNull(trueChild) ;
    assertTrue(falseChild.isLeaf()) ;
    assertTrue(trueChild.isLeaf()) ;
    
    assertEquals(-1,node.getState()) ;
    assertEquals(0,falseChild.getState()) ;
    assertEquals(0,trueChild.getState()) ;
    
    assertEquals(0,node.getBitIndex()) ;
    assertEquals(1,falseChild.getBitIndex()) ;
    assertEquals(5,trueChild.getBitIndex()) ;
  }
  
  
  /**
   * Test the recursive move to leaf.
   */
  public void testRecursiveExpandNotLeaf() throws BddException {
    
    BddNode node = new BddNode(0) ;
    node.addState(0) ;
    node.addState(1) ;
    
    BddNode falseChild = new BddNode(1) ;
    falseChild.addState(2) ;
    node.setChild(false,falseChild) ;
    
    BddNode falseFalseChild = new BddNode(2) ;
    falseFalseChild.addState(3) ;
    falseChild.setChild(false,falseFalseChild) ;
    
    node.expandNotLeaf(4) ;
    
    BddNode trueChild       = node.getChild(true) ;
            falseFalseChild = falseChild.getChild(false) ; 
    BddNode falseTrueChild  = falseChild.getChild(true) ;
      
    // root
    assertEquals(-1,node.getState()) ;
    
    // left
    assertEquals(-1,falseChild.getState()) ;
    assertNotNull(falseFalseChild) ;
    assertNotNull(falseTrueChild) ;
    assertTrue(falseFalseChild.isLeaf()) ;
    assertTrue(falseTrueChild.isLeaf()) ;
    
    // right
    assertNotNull(trueChild) ;
    assertTrue(trueChild.isLeaf()) ;
    
    // bit indexes
    assertEquals(0,node.getBitIndex()) ;
    assertEquals(1,falseChild.getBitIndex()) ;
    assertEquals(2,falseFalseChild.getBitIndex()) ;
    assertEquals(4,trueChild.getBitIndex()) ;
    assertEquals(4,falseTrueChild.getBitIndex()) ;
  }
  
  
  /**
   * Test multiple states expansion.
   * 
   * Node with 2 states.
   */
  public void testExpandMultiple_2() throws BddException {
    
    BddNode node = new BddNode(0) ;
    node.addState(1) ;
    node.addState(2) ;
    
    assertTrue(node.isLeaf()) ;
    node.expandMultiple(0) ;
    assertFalse(node.isLeaf()) ;
    
    BddNode falseChild = node.getChild(false) ;
    BddNode trueChild  = node.getChild(true) ;
    
    assertNotNull(falseChild) ;
    assertNotNull(trueChild) ;
    assertTrue(falseChild.isLeaf()) ;
    assertTrue(trueChild.isLeaf()) ;
    
    assertEquals(1,falseChild.getState()) ;
    assertEquals(2,trueChild.getState()) ;
  }
  
  
  /**
   * Test multiple states expansion.
   * 
   * Node with 5 states.
   */
  public void testExpandMultiple_5() throws BddException {
    
    BddNode node = new BddNode(0) ;
    node.addState(1) ;
    node.addState(2) ;
    node.addState(3) ;
    node.addState(4) ;
    node.addState(5) ;
    
    assertTrue(node.isLeaf()) ;
    node.expandMultiple(3) ;
    assertFalse(node.isLeaf()) ;
    assertEquals(-1,node.getState()) ;
    assertEquals(3,node.getBitIndex()) ;
    
    BddNode falseChild = node.getChild(false) ; 
    BddNode trueChild  = node.getChild(true) ;
    
    assertNotNull(falseChild) ;
    assertNotNull(trueChild) ;
    assertFalse(falseChild.isLeaf()) ;
    assertFalse(trueChild.isLeaf()) ;
    assertEquals(4,falseChild.getBitIndex()) ;
    assertEquals(4,trueChild.getBitIndex()) ;
    
    BddNode falseFalseChild = falseChild.getChild(false) ;
    BddNode falseTrueChild  = falseChild.getChild(true) ;
    BddNode trueFalseChild  = trueChild.getChild(false) ;
    BddNode trueTrueChild   = trueChild.getChild(true) ;
    
    assertNotNull(falseFalseChild) ;
    assertNotNull(falseTrueChild) ;
    assertNotNull(trueFalseChild) ;
    assertNotNull(trueTrueChild) ;
    assertTrue(falseFalseChild.isLeaf()) ;
    assertTrue(falseTrueChild.isLeaf()) ;
    assertTrue(trueFalseChild.isLeaf()) ;
    assertFalse(trueTrueChild.isLeaf()) ;
    assertEquals(5,falseFalseChild.getBitIndex()) ;
    assertEquals(5,falseTrueChild.getBitIndex()) ;
    assertEquals(5,trueFalseChild.getBitIndex()) ;
    assertEquals(5,trueTrueChild.getBitIndex()) ;
    
    BddNode trueTrueFalseChild = trueTrueChild.getChild(false) ;
    BddNode trueTrueTrueChild  = trueTrueChild.getChild(true) ;
    
    assertNotNull(trueTrueFalseChild) ;
    assertNotNull(trueTrueTrueChild) ;
    assertTrue(trueTrueFalseChild.isLeaf()) ;
    assertTrue(trueTrueTrueChild.isLeaf()) ;
    assertEquals(6,trueTrueFalseChild.getBitIndex()) ;
    assertEquals(6,trueTrueTrueChild.getBitIndex()) ;
    
    assertEquals(1,falseFalseChild.getState()) ;
    assertEquals(2,falseTrueChild.getState()) ;
    assertEquals(3,trueFalseChild.getState()) ;
    assertEquals(4,trueTrueFalseChild.getState()) ;
    assertEquals(5,trueTrueTrueChild.getState()) ;
  }
  
  
  /**
   * Test witch combines the two expansions.
   */
  public void testExpand() throws BddException {
    
    BddTree tree = new BddTree("left","right") ;
    tree.addState(1) ;
    
    BddNode falseNode = new BddNode(1) ;
    BddNode trueNode  = new BddNode(1) ;
    tree.setChild(false,falseNode) ;
    tree.setChild(true,trueNode) ;
    
    falseNode.addState(2) ;
    trueNode.addState(3) ;
    
    tree.expand(5) ;
    
    BddNode falseFalseNode = falseNode.getChild(false) ;
    BddNode falseTrueNode  = falseNode.getChild(true) ;
    BddNode trueFalseNode  = trueNode.getChild(false) ;
    BddNode trueTrueNode   = trueNode.getChild(true) ;
    
    assertEquals(0,tree.getBitIndex()) ;
    assertEquals(5,falseNode.getBitIndex()) ;
    assertEquals(5,trueNode.getBitIndex()) ;
    assertEquals(6,falseFalseNode.getBitIndex()) ;
    assertEquals(6,falseTrueNode.getBitIndex()) ;
    assertEquals(6,trueFalseNode.getBitIndex()) ;
    assertEquals(6,falseFalseNode.getBitIndex()) ;
    
    assertNotNull(falseFalseNode) ;
    assertNotNull(falseTrueNode) ;
    assertNotNull(trueFalseNode) ;
    assertNotNull(trueTrueNode) ;
    
    assertFalse(falseNode.isLeaf()) ;
    assertFalse(trueNode.isLeaf()) ;
    assertTrue(falseFalseNode.isLeaf()) ;
    assertTrue(falseTrueNode.isLeaf()) ;
    assertTrue(trueFalseNode.isLeaf()) ;
    assertTrue(trueTrueNode.isLeaf()) ;
    
    assertEquals(-1,falseNode.getState()) ;
    assertEquals(-1,trueNode.getState()) ;
    assertEquals(2,falseFalseNode.getState()) ;
    assertEquals(1,falseTrueNode.getState()) ;
    assertEquals(3,trueFalseNode.getState()) ;
    assertEquals(1,trueTrueNode.getState()) ;
  }
  
}
