/*
 * TestDisjunction.java
 */
package parser;

import junit.framework.* ;;


/**
 * Test the compound disjunction.
 * 
 * @author Pierre Salagnac
 */
public class TestDisjunction extends TestCase {
  

  /**
   * Test minimisation for subsdisjunctions.
   */
  public void testMinimize() throws Exception {
    
    Move move1 = new Move(null,"move1",false,false,null) ;
    Move move2 = new Move(null,"move2",false,false,null) ;
    Move move3 = new Move(null,"move3",false,false,null) ;
    Move move4 = new Move(null,"move4",false,false,null) ;
    
    Disjunction subDisjunction1 = new Disjunction() ;
    Disjunction subDisjunction2 = new Disjunction() ;
    
    subDisjunction1.addChild(move1) ;
    subDisjunction1.addChild(move2) ;
    subDisjunction2.addChild(move3) ;
    subDisjunction2.addChild(move4) ;
    
    
    Disjunction disjunction = new Disjunction() ;
    disjunction.addChild(subDisjunction1) ;
    disjunction.addChild(subDisjunction2) ;
    
    disjunction.minimize() ;
    
    assertEquals("children count",4,disjunction.getChildrenCount()) ;
    
    assertEquals("child 1",move1,disjunction.getChild(0)) ;
    assertEquals("child 2",move2,disjunction.getChild(1)) ;
    assertEquals("child 3",move3,disjunction.getChild(2)) ;
    assertEquals("child 4",move4,disjunction.getChild(3)) ;
    
  }
  
  /**
   * Test minimisation for an empty disjunction.
   */
  public void testEmptyMinimize() throws Exception {
    
    Disjunction disjunction = new Disjunction() ;
    
    assertNull("empty not null",disjunction.minimize()) ;
    
    Disjunction subDisjunction1 = new Disjunction() ;
    Disjunction subDisjunction2 = new Disjunction() ;
    
    disjunction.addChild(subDisjunction1) ;
    disjunction.addChild(subDisjunction2) ;
    
    assertNull("empty not null",disjunction.minimize()) ;
    
  }
  
  
  /**
   * Test degenerated empty sequence.
   */
  public void testDegeneratedEmptyMinimize() throws Exception {
    
    Disjunction disjunction = new Disjunction() ;
    Disjunction subDisjunction = new Disjunction() ;
    Sequence subSequence = new Sequence() ;
    
    subDisjunction.addChild(subSequence) ;
    disjunction.addChild(subDisjunction) ;
    
    assertNull(disjunction.minimize()) ;
  }
  
  
  /**
   * Test minimisation for a large empty disjunction, with empty sequences.
   */
  public void testLargeEmptyMinimize() throws Exception {
    
    Disjunction disjunction = new Disjunction() ;
    
    CompoundPath sub1 = new Disjunction() ;
    CompoundPath sub2 = new Sequence() ;
    
    sub1.addChild(new Sequence()) ;
    sub2.addChild(new Disjunction()) ;
    
    disjunction.addChild(sub1) ;
    disjunction.addChild(sub2) ;
    
    assertNull("empty not null",disjunction.minimize()) ;
    
  }
  
  /**
   * Test a multiple dusjunction with a single move.
   */
  public void testStarMinimize() throws Exception {
    
    Move move = new Move(null,"move",false,false,null) ;
    
    Disjunction disjunction = new Disjunction() ;
    disjunction.addChild(move) ;
    disjunction.setMultiple(true) ;
    
    Path path = disjunction.minimize() ;
    
    assertEquals(move,path) ;
    assertTrue(path.isMultiple()) ;
  }
  
  
  /**
   * Minimize a multpile disjunction composed with useless sequences.
   */
  public void testMultipleMinimize() throws Exception {
    
    Disjunction disjunction = new Disjunction() ;
    
    Sequence subSequence1 = new Sequence() ;
    Sequence subSequence2 = new Sequence() ;
    
    Move move1 = new Move(null,"left",false,false,null) ;
    Move move2 = new Move(null,"right",false,false,null) ;
    
    subSequence1.addChild(move1) ;
    subSequence2.addChild(move2) ;
    
    disjunction.addChild(subSequence1) ;
    disjunction.addChild(subSequence2) ;
    
    disjunction.setMultiple(true) ;
    disjunction.minimize() ;
    
    assertEquals(2,disjunction.getChildrenCount()) ;
    
    assertEquals("move 1",move1,disjunction.getChild(0)) ;
    assertEquals("move 2",move2,disjunction.getChild(1)) ;
  }
  
  
  /**
   * Test a disjunction between a path and the null value.
   */
  public void testNullDisjunction() throws Exception {
    
    Disjunction disjunction = new Disjunction() ;
    
    // no exception with an empty disjunction
    assertFalse(disjunction.isNullDisjunction()) ;
    
    NullPath nullPath = new NullPath() ;
    Move move = new Move(null,"name",false,false,null) ;
    
    // move + null
    disjunction.addChild(move) ;
    disjunction.addChild(nullPath) ;
    assertTrue(disjunction.isNullDisjunction()) ;
    
    // null + path
    disjunction = new Disjunction() ;
    disjunction.addChild(nullPath) ;
    disjunction.addChild(move) ;
    assertTrue(disjunction.isNullDisjunction()) ;
    
    // null
    disjunction = new Disjunction() ;
    disjunction.addChild(nullPath) ;
    assertFalse(disjunction.isNullDisjunction()) ;
  }
  
  
  /**
   * Test the romoving of null part.
   */
  public void testRemoveNull() throws Exception {
    
    // empty
    Disjunction disjunction = new Disjunction() ;
    assertNull(disjunction.removeNull()) ;
    
    // null
    disjunction = new Disjunction() ;
    disjunction.addChild(new NullPath()) ;
    assertNull(disjunction.removeNull()) ;
    
    // single move
    Move move = new Move(null,"move",false,false,null) ;
    disjunction = new Disjunction() ;
    disjunction.addChild(move) ;
    assertEquals(move,disjunction.removeNull()) ;
    
    // move + null
    move = new Move(null,"move",false,false,null) ;
    disjunction = new Disjunction() ;
    disjunction.addChild(move) ;
    disjunction.addChild(new NullPath()) ;
    assertEquals(move,disjunction.removeNull()) ;
    
    // move1 + null + move2
    Move move1 = new Move(null,"move1",false,false,null) ;
    Move move2 = new Move(null,"move2",false,false,null) ;
    disjunction = new Disjunction() ;
    disjunction.addChild(move1) ;
    disjunction.addChild(new NullPath()) ;
    disjunction.addChild(move2) ;
    
    Path path = disjunction.removeNull() ;
    assertEquals(Disjunction.class,path.getClass()) ;
    assertEquals(2,((Disjunction)path).getChildrenCount()) ;
    assertEquals(move1,((Disjunction)path).getChild(0)) ;
    assertEquals(move2,((Disjunction)path).getChild(1)) ;
  }
  
}
