/*
 * TestOrderParser.java
 */
package parser ;

import java.io.* ;

import junit.framework.* ;


/**
 * Unit test for the input list parser.
 * 
 * @author Pierre Salagnac
 */
public class TestOrderParser extends TestCase {

  /**
   * Parse a classical value.
   */
  public void testParse() throws Exception {
    
    String order = "left:0  right:2  \t\r\n\f  up:1" ; 
    StringReader reader = new StringReader(order) ;
    
    OrderParser parser = new OrderParser(reader) ;
    parser.parse() ;
    
    assertEquals(3,parser.getInputCount()) ;
    
    assertEquals("left" ,0,parser.getInputIndex("left")) ;
    assertEquals("up"   ,1,parser.getInputIndex("up")) ;
    assertEquals("right",2,parser.getInputIndex("right")) ;
    
    assertFalse(parser.containsInput("down")) ;
  }
  
  
  /**
   * Parse a string with a hole : the index 1 is missing.
   */
  public void testHole() {
    
    String order = "left:0  right:2" ;
    StringReader reader = new StringReader(order) ;
    
    OrderParser parser = new OrderParser(reader) ;
    
    try {
      parser.parse() ;
      fail("hole accepted") ;
    }
    catch (ParseException e) {
    }
    
  }
  
  
  /**
   * Parse a string with a duplicated name and a duplicate index.
   */
  public void testDuplicate() {
    
    // duplicated name
    try {
      String order = "left:0 left:1" ;
      StringReader reader = new StringReader(order) ;
      
      OrderParser parser = new OrderParser(reader) ;
      parser.parse() ;
      fail("duplicated name accepted") ;
    }
    catch (ParseException e) {
    }
    
    // duplicated index
    try {
      String order = "left:0 right:0" ;
      StringReader reader = new StringReader(order) ;
      
      OrderParser parser = new OrderParser(reader) ;
      parser.parse() ;
      fail("duplicated index accepted") ;
    }
    catch (ParseException e) {
    }
    
  }
  
  
  /**
   * Parse an order without the 0 index.
   */
  public void testZero() throws Exception {
    
    String order = "left:1 right:2" ;
    StringReader reader = new StringReader(order) ;
    
    OrderParser parser = new OrderParser(reader) ;
    parser.parse() ;
    
    assertEquals(2,parser.getInputCount()) ;
    
    assertEquals(1,parser.getInputIndex("left")) ;
    assertEquals(2,parser.getInputIndex("right")) ;
    
    assertFalse(parser.containsInput("down")) ;
  }
  
  
  /**
   * Parse with the 'other' syntax.
   */
  public void testSyntax() throws Exception {
    
    String order = "left=0, right:1 up=2" ;
    StringReader reader = new StringReader(order) ;
    
    OrderParser parser = new OrderParser(reader) ;
    parser.parse() ;
    
  }
  
  
  /**
   * Add a positive offset to the indexes.
   */
  public void testPositiveOffset() throws Exception {
    
    String order = "left=0 right=1 up=2" ;
    StringReader reader = new StringReader(order) ;
    
    OrderParser parser = new OrderParser(reader) ;
    parser.parse(2) ;
    
    assertEquals(2,parser.getInputIndex("left")) ;
    assertEquals(3,parser.getInputIndex("right")) ;
    assertEquals(4,parser.getInputIndex("up")) ;
    assertEquals(-1,parser.getInputIndex("other")) ;
    
    assertEquals("left",parser.getInput(2)) ;
    assertEquals("right",parser.getInput(3)) ;
    assertEquals("up",parser.getInput(4)) ;
  }
  
  
  /**
   * Add a negative offset to the indexes.
   */
  public void testNegativeOffset() throws Exception {
    
    String order = "left=1 right=2 up=3" ;
    StringReader reader = new StringReader(order) ;
    
    OrderParser parser = new OrderParser(reader) ;
    parser.parse(-1) ;
    
    assertEquals(0,parser.getInputIndex("left")) ;
    assertEquals(1,parser.getInputIndex("right")) ;
    assertEquals(2,parser.getInputIndex("up")) ;
    assertEquals(-1,parser.getInputIndex("other")) ;
    
    assertEquals("left",parser.getInput(0)) ;
    assertEquals("right",parser.getInput(1)) ;
    assertEquals("up",parser.getInput(2)) ;
  }
  
  
  /**
   * Add a negative offset to the indexes.
   */
  public void testWrongOffset() throws Exception {
    
    String order = "left=0 right=1 up=2" ;
    StringReader reader = new StringReader(order) ;
    
    OrderParser parser = new OrderParser(reader) ;
    
    try {
      parser.parse(-1) ;
      fail("negative index accepted") ;
    }
    catch (ParseException e) {}
    
  }
}
