/*
 * TestSequence.java
 */
package parser;

import junit.framework.* ;


/**
 * Test the compound sequence.
 * 
 * @author Pierre Salagnac
 */
public class TestSequence extends TestCase {
  
  
  /**
   * Test minimisation for subsequences.
   */
  public void testMinimize() throws Exception {
    
    Move move1 = new Move(null,"move1",false,false,null) ;
    Move move2 = new Move(null,"move2",false,false,null) ;
    Move move3 = new Move(null,"move3",false,false,null) ;
    Move move4 = new Move(null,"move4",false,false,null) ;
    
    Sequence subSequence1 = new Sequence() ;
    Sequence subSequence2 = new Sequence() ;
    
    subSequence1.addChild(move1) ;
    subSequence1.addChild(move2) ;
    subSequence2.addChild(move3) ;
    subSequence2.addChild(move4) ;
    
    
    Sequence sequence = new Sequence() ;
    sequence.addChild(subSequence1) ;
    sequence.addChild(subSequence2) ;
    
    sequence.minimize() ;
    
    assertEquals("children count",4,sequence.getChildrenCount()) ;
    
    assertEquals("child 1",move1,sequence.getChild(0)) ;
    assertEquals("child 2",move2,sequence.getChild(1)) ;
    assertEquals("child 3",move3,sequence.getChild(2)) ;
    assertEquals("child 4",move4,sequence.getChild(3)) ;
    
  }
  
  
  /**
   * Test minimisation for an empty sequence.
   */
  public void testEmptyMinimize() throws Exception {
    
    Sequence sequence = new Sequence() ;
    
    assertNull("empty not null",sequence.minimize()) ;
    
    Sequence subSequence1 = new Sequence() ;
    Sequence subSequence2 = new Sequence() ;
    
    sequence.addChild(subSequence1) ;
    sequence.addChild(subSequence2) ;
    
    assertNull("empty not null",sequence.minimize()) ;
    
  }
  
  
  /**
   * Test degenerated empty sequence.
   */
  public void testDegeneratedEmptyMinimize() throws Exception {
    
    Sequence sequence = new Sequence() ;
    Sequence subSequence = new Sequence() ;
    Disjunction subDisjunction = new Disjunction() ;
    
    subSequence.addChild(subDisjunction) ;
    sequence.addChild(subSequence) ;
    
    assertNull(sequence.minimize()) ;
  }
  
  
  /**
   * Test minimisation for a large empty sequence, with empty disjunctions.
   */
  public void testLargeEmptyMinimize() throws Exception {
    
    Sequence sequence = new Sequence() ;
    
    CompoundPath sub1 = new Sequence() ;
    CompoundPath sub2 = new Disjunction() ;
    
    sub1.addChild(new Disjunction()) ;
    sub2.addChild(new Sequence()) ;
    
    sequence.addChild(sub1) ;
    sequence.addChild(sub2) ;
    
    assertNull("empty not null",sequence.minimize()) ;
    
  }
  
  
  /**
   * Test a multiple sequence with a single move.
   * 
   * The not-multiple move becomes multiple.
   */
  public void testStarMinimize() throws Exception {
    
    Move move = new Move(null,"move",false,false,null) ;
    
    Sequence sequence = new Sequence() ;
    sequence.addChild(move) ;
    sequence.setMultiple(true) ;
    
    Path path = sequence.minimize() ;
    
    assertEquals(move,path) ;
    assertTrue(path.isMultiple()) ;
  }
  
  
  /**
   * Minimize a useldess multiple sequence.
   */
  public void testTripleStarMinimize() throws Exception {
    
    Move move = new Move(null,"move",false,true,null) ;
    
    Sequence sequence = new Sequence() ;
    Sequence subSequence = new Sequence() ;
    sequence.setMultiple(true) ;
    subSequence.setMultiple(true) ;
    
    subSequence.addChild(move) ;
    sequence.addChild(subSequence) ;
    
    Path path = sequence.minimize() ;
    
    assertEquals(move,path) ;
    assertTrue(path.isMultiple()) ;
  }
  
  
  /**
   * Minimize a multpile disjunction composed with useless sequences.
   */
  public void testMultipleMinimize() throws Exception {
    
    Sequence sequence = new Sequence() ;
    
    Sequence subSequence1 = new Sequence() ;
    Sequence subSequence2 = new Sequence() ;
    
    Move move1 = new Move(null,"left",false,false,null) ;
    Move move2 = new Move(null,"right",false,false,null) ;
    
    subSequence1.addChild(move1) ;
    subSequence2.addChild(move2) ;
    
    sequence.addChild(subSequence1) ;
    sequence.addChild(subSequence2) ;
    
    sequence.setMultiple(true) ;
    sequence.minimize() ;
    
    assertEquals(2,sequence.getChildrenCount()) ;
    
    assertEquals("move 1",move1,sequence.getChild(0)) ;
    assertEquals("move 2",move2,sequence.getChild(1)) ;
  }
  
}
