/*
 * TestTypedefParser.java
 */
package parser ;

import java.io.* ;
import java.util.* ;

import junit.framework.*;


/**
 * Unit test for the typedef parser.
 * 
 * @author Pierre Salagnac
 */
public class TestTypedefParser extends TestCase {
	
	
	/**
	 * Parse a test file 
	 */
  private Typedef[] parse(String name) throws Exception {
  	
    String basePath = System.getProperty("syntax") ;
    
    File baseDir = new File(basePath) ;
    File file = new File(basePath,name) ;
    
    Reader reader = new FileReader(file) ;
    
    try {
      TypedefParser parser = new TypedefParser(reader) ;
      parser.parse() ;
      reader.close() ;
      
      return parser.getTypedefs() ;
    }
    catch (ParseException e) {
      e.printStackTrace() ;
      fail("parse exception: "+name) ;
    }
    catch(TokenMgrError e) {
      e.printStackTrace() ;
      fail("token error: "+name) ;
    }
    
    return null ;
  }
  
  
  /**
   * Display the types.
   */
  private void trace(Typedef[] types) {
    
    for (int i=0;i<types.length;i++) {
      System.err.println(types[i]) ;
      
      MemberList[] contents = types[i].getContents() ;
      for (int j=0;j<contents.length;j++) {
        System.err.println("  "+contents[j]) ;
        
        String[] names = contents[j].getNames() ;
        for (int k=0;k<names.length;k++) {
          String type = contents[j].getType(names[k]) ;
          Path   path = contents[j].getPath(names[k]) ;
          System.err.println("    "+type+" * "+names[k]) ;
          if (path != null)
            System.err.println("      "+path) ;
        }
      }
    }
    
    
  }
  
  
  /**
   * Parse strange files. A successful parse is a success.
   */
  public void testParse() throws Exception {
    
    System.err.println("\n---------- route") ;
    trace(parse("route")) ;
    
    System.err.println("\n---------- subpath") ;
    trace(parse("subpath")) ;
  }
  
  
  /**
   * Test a structure with lots of members.
   */
  public void testStructure() throws Exception {
    
    Typedef[] typedefs = parse("structure") ;
    
    assertEquals(3,typedefs.length) ;
    assertEquals("T1",typedefs[0].getName()) ;
    assertEquals("T2",typedefs[1].getName()) ;
    assertEquals("T3",typedefs[2].getName()) ;
    
    MemberList[] contentsT1 = typedefs[0].getContents() ;
    MemberList[] contentsT2 = typedefs[1].getContents() ;
    MemberList[] contentsT3 = typedefs[2].getContents() ;
    
    assertEquals(1,contentsT1.length) ;
    assertEquals(2,contentsT2.length) ;
    assertEquals(3,contentsT3.length) ;
    
  }
  
  
  /**
   * Test primitive and routed links.
   */
  public void testPrimitiveRouted() throws Exception {
    
    Typedef[] typedefs = parse("primitive-routed") ;
    assertEquals(1,typedefs.length) ;
    
    MemberList[] contents = typedefs[0].getContents() ;
    assertEquals(1,contents.length) ;
    
    assertTrue(contents[0].isPrimitive("left")) ;
    assertTrue(contents[0].isPrimitive("right")) ;
    assertFalse(contents[0].isPrimitive("next")) ;
  }
  
  
  /**
   * Test the data member.
   */
  public void testData() throws Exception {
    
    Typedef[] typedefs = parse("data") ;
    assertEquals(2,typedefs.length) ;
    
    MemberList[] contentsT1 = typedefs[0].getContents() ;
    MemberList[] contentsT2 = typedefs[1].getContents() ;
    assertEquals(1,contentsT1.length) ;
    assertEquals(1,contentsT2.length) ;
    
    assertNull(contentsT1[0].getData()) ;
    
    assertEquals("01001101",contentsT2[0].getData()) ;
    
  }
  
  
  /**
   * Test path multiplicities
   */
  public void testMultiplicity() throws Exception {
    
    Typedef[] typedefs = parse("multiplicity") ;
    assertEquals(1,typedefs.length) ;
    Typedef typedef = typedefs[0] ;
    
    MemberList[] contents = typedef.getContents() ;
    assertEquals(1,contents.length) ;
    MemberList content = contents[0] ;
    
    // left
    Path left = content.getPath("left") ;
    assertEquals(Move.class,left.getClass()) ;
    assertFalse(left.isMultiple()) ;
    
    // left 0
    Path left0 = content.getPath("left0") ;
    assertNull(left0) ;
    
    // left 1
    Path left1 = content.getPath("left1") ;
    assertEquals(Move.class,left1.getClass()) ;
    assertFalse(left1.isMultiple()) ;
    
    // left 2
    Path left2 = content.getPath("left2") ;
    assertEquals(Sequence.class,left2.getClass()) ;
    assertEquals(2,((Sequence)left2).getChildrenCount()) ;
    
    // left star
    Path leftS = content.getPath("leftS") ;
    assertEquals(Move.class,leftS.getClass()) ;
    assertTrue(leftS.isMultiple()) ;
    
    // sequence
    Path sequence = content.getPath("sequence") ;
    assertEquals(Sequence.class,sequence.getClass()) ;
    assertEquals(4,((Sequence)sequence).getChildrenCount()) ;
    assertTrue(((Sequence)sequence).getChild(3).isMultiple()) ;
    assertFalse(sequence.isMultiple()) ;
    
    // sequence 0
    Path sequence0 = content.getPath("sequence0") ;
    assertNull(sequence0) ;
    
    // sequence 1
    Path sequence1 = content.getPath("sequence1") ;
    assertEquals(Sequence.class,sequence1.getClass()) ;
    assertEquals(2,((Sequence)sequence1).getChildrenCount()) ;
    
    // sequence 2
    Path sequence2 = content.getPath("sequence2") ;
    assertEquals(Sequence.class,sequence2.getClass()) ;
    assertEquals(4,((Sequence)sequence2).getChildrenCount()) ;
    assertFalse(sequence2.isMultiple()) ;
    
    // sequence star
    Path sequenceS = content.getPath("sequenceS") ;
    assertEquals(Sequence.class,sequenceS.getClass()) ;
    assertEquals(2,((Sequence)sequenceS).getChildrenCount()) ;
    assertTrue(sequenceS.isMultiple()) ;
    
    // sequence 2*3
    Path sequence2x3 = content.getPath("sequence2x3") ;
    assertEquals(Sequence.class,sequence2x3.getClass()) ;
    assertEquals(12,((Sequence)sequence2x3).getChildrenCount()) ;
    
    // disjunction
    Path disjunction = content.getPath("disjunction") ;
    assertEquals(Disjunction.class,disjunction.getClass()) ;
    assertEquals(2,((Disjunction)disjunction).getChildrenCount()) ;
    assertFalse(disjunction.isMultiple()) ;
    
    // disjunction 0
    Path disjunction0 = content.getPath("disjunction0") ;
    assertNull(disjunction0) ;
    
    // disjunction 1
    Path disjunction1 = content.getPath("disjunction1") ;
    assertEquals(Disjunction.class,disjunction1.getClass()) ;
    assertEquals(2,((Disjunction)disjunction1).getChildrenCount()) ;
    
    // disjunction 2
    Path disjunction2 = content.getPath("disjunction2") ;
    assertEquals(Sequence.class,disjunction2.getClass()) ;
    assertEquals(2,((Sequence)disjunction2).getChildrenCount()) ;
    assertEquals(Disjunction.class,((Sequence)disjunction2).getChild(0).getClass()) ;
    assertEquals(Disjunction.class,((Sequence)disjunction2).getChild(1).getClass()) ;
    
    // disjunction star
    Path disjunctionS = content.getPath("disjunctionS") ;
    assertEquals(Disjunction.class,disjunctionS.getClass()) ;
    assertEquals(2,((Disjunction)disjunctionS).getChildrenCount()) ;
    assertTrue(disjunctionS.isMultiple()) ;
    
    // star
    Path star = content.getPath("star") ;
    assertEquals(Move.class,star.getClass()) ;
    assertTrue(star.isMultiple()) ;
    
    // double star
    Path doublestar = content.getPath("doublestar") ;
    assertEquals(Move.class,doublestar.getClass()) ;
    assertTrue(doublestar.isMultiple()) ;
  }
  
  
  /**
   * Test the null route.
   */
  public void testNull() throws Exception {
    
    Typedef[] typedefs = parse("null") ;
    assertEquals(1,typedefs.length) ;
    Typedef typedef = typedefs[0] ;
    
    MemberList[] contents = typedef.getContents() ;
    assertEquals(1,contents.length) ;
    MemberList content = contents[0] ;
    
    // ptr
    Path ptr = content.getPath("ptr") ;
    assertEquals(Disjunction.class,ptr.getClass()) ;
    assertEquals(2,((Disjunction)ptr).getChildrenCount()) ;
    
    assertEquals(NullPath.class,((Disjunction)ptr).getChild(0).getClass()) ;
    assertEquals(Move.class,((Disjunction)ptr).getChild(1).getClass()) ;
    
    // nn
    Path nn = content.getPath("nn") ;
    assertEquals(NullPath.class,nn.getClass()) ;
    
  }
  
  
  /**
   * Test absolute points in route.
   */
  public void testEndPoint() throws Exception {
    
    Typedef[] typedefs = parse("endpoint") ;
    assertEquals(1,typedefs.length) ;
    Typedef typedef = typedefs[0] ;
    
    MemberList[] contents = typedef.getContents() ;
    assertEquals(1,contents.length) ;
    MemberList content = contents[0] ;
    
    // only root
    Path onlyroot = content.getPath("onlyroot") ;
    assertEquals(Move.class,onlyroot.getClass()) ;
    assertNull(((Move)onlyroot).getName()) ;
    assertEquals("root",onlyroot.getStartPoint()) ;
    
    // sequence
    Path sequence = content.getPath("sequence") ;
    assertEquals(Sequence.class,sequence.getClass()) ;
    assertEquals("leaf",sequence.getEndPoint()) ;
    
    // disjunction
    Path disjunction = content.getPath("disjunction") ;
    assertEquals(Disjunction.class,disjunction.getClass()) ;
    assertEquals("leaf",disjunction.getEndPoint()) ;
  }
  
  
  /**
   * Parse an empty structure.
   */
  public void testEmpty() throws Exception {
    
    Typedef[] typedefs = parse("empty") ;
    assertEquals(1,typedefs.length) ;
    
    MemberList[] contents = typedefs[0].getContents() ;
    assertEquals(2,contents.length) ;
    
    assertEquals(0,contents[0].getNames().length) ;
    assertEquals(0,contents[1].getNames().length) ;
  }
  
  
  /**
   * Test root pointers.
   */
  public void testRootList() throws Exception {
    
    Typedef[] typedefs = parse("rootlist") ;
    assertEquals(1,typedefs.length) ;
    
    assertEquals(1,typedefs[0].getContents().length) ;
    
    String[] names = typedefs[0].getRootVariables() ;
    
    Arrays.sort(names) ;
    assertTrue(Arrays.binarySearch(names,"x")  >= 0) ;
    assertTrue(Arrays.binarySearch(names,"y")  >= 0) ;
    assertTrue(Arrays.binarySearch(names,"z")  >= 0) ;
    assertFalse(Arrays.binarySearch(names,"t") >= 0) ;
  }
  
  
  /**
   * Test that two different absolute points cannot be on the same 
   * possition.
   */
  public void testDifferentPoints() throws Exception {
    
    try {
      Typedef[] typedefs = parse("differentpoints") ;
      fail("two different absolute points accepted") ;
    }
    catch (PathException e) {
      
    }
  }
  
  
  /**
   * Test if compound paths end with a leaf.
   */
  public void testLeafEnding() throws Exception {
    
    Typedef[] typedefs = parse("leafend") ;
    assertEquals(1,typedefs.length) ;
    
    assertEquals(1,typedefs[0].getContents().length) ;
    MemberList content = typedefs[0].getContents()[0] ;
    
    // yes
    for (int i=0; ;i++) {
      String name = "yes"+i ;
      Path path = content.getPath(name) ;
      if (path == null)
        break ;
      if (path instanceof CompoundPath)
        assertTrue(name,((CompoundPath)path).canLeafEnd()) ;
      else
        assertEquals(name,"leaf",path.getEndPoint()) ;
    }
    
    // no
    for (int i=0; ;i++) {
      String name = "no"+i ;
      Path path = content.getPath(name) ;
      if (path == null)
        break ;
      if (path instanceof CompoundPath)
        assertFalse(name,((CompoundPath)path).canLeafEnd()) ;
      else
        assertFalse(name,"leaf".equals(path.getEndPoint())) ;
    }
  }
  
  
  /**
   * Test the possibility to add comments.
   */
  public void testComment() throws Exception {
    
    Typedef[] typedefs = parse("comment") ;
    
    assertNotNull(typedefs) ;
    
  }
  
}
