/*
 * TestPatternBuilder.java
 */
package tree;

import java.io.* ;
import java.util.* ;

import parser.* ;

import junit.framework.* ;


/**
 * Test construction a tree patterns. 
 * 
 * @author Pierre Salagnac
 */
public class TestPatternBuilder extends TestCase {
  
  
  /**
   * Parse a typedef file, and build the patterns list.
   * 
   * @param file The file to parse.
   * @param inputs The input order.
   * 
   * @return All parsed patterns.
   */
  private Pattern[] parsePatterns(File file,String inputs)
  throws Exception {
    
    FileReader reader = new FileReader(file) ;
    
    TypedefParser parser = new TypedefParser(reader) ;
    parser.parse() ;
    reader.close() ;
    
    Typedef[] typedefs = parser.getTypedefs() ;
    
    ProgramParser program = new MockProgramParser(10,10,"",inputs,"x:1") ;
    
    PatternBuilder builder = new PatternBuilder(program) ;
    builder.build(typedefs,true) ;
    
    return builder.getPatterns() ;
  }
  
  
  /**
   * Test the list of names.
   */
  public void testInputNames() throws Exception {
    
    String order = "left:0 right:1 up:2" ;
    
    File file = new File(System.getProperty("syntax"),"names") ;
    FileReader reader = new FileReader(file) ;
    
    TypedefParser parser = new TypedefParser(reader) ;
    parser.parse() ;
    reader.close() ;
    
    Typedef[] typedefs = parser.getTypedefs() ;
    
    ProgramParser program = new MockProgramParser(10,10,"",order,"x:1") ;
    PatternBuilder builder = new PatternBuilder(program) ;
    builder.build(typedefs,true) ;
    
    String[] names = builder.getInputNames() ;
    Arrays.sort(names) ;
    
    assertTrue(Arrays.binarySearch(names,"left")  >= 0) ;
    assertTrue(Arrays.binarySearch(names,"right") >= 0) ;
    assertTrue(Arrays.binarySearch(names,"up")    >= 0) ;
    assertFalse(Arrays.binarySearch(names,"next") >= 0) ;
    
  }
  
  
  /**
   * Test the inputs.
   */
  public void testInputs() throws Exception {
    
    File file = new File(System.getProperty("syntax"),"inputs") ;
    
    Pattern[] patterns = parsePatterns(file,"left:0 right:1 null_:2 other:3") ;
    assertEquals(1,patterns.length) ;
    
    Pattern pattern = patterns[0] ;
    assertFalse(pattern.isLeaf()) ;
    assertEquals("T",pattern.getOutput()) ;
    
    Input[] inputs = pattern.getInputs() ;
    assertEquals(4,inputs.length) ;
    
    // left
    assertTrue("primitive",inputs[0].isPrimitive()) ;
    
    // right
    assertFalse("routed",inputs[1].isPrimitive()) ;
    
    // null
    assertTrue("null",inputs[2].isNull()) ;
    
    // other
    assertTrue("undefined",inputs[3].isUndefined()) ;
  }
  
  
  /**
   * Test a typedef with a null route. The pattern has to be splitted.
   */
  public void testNull() throws Exception {
    
    File file = new File(System.getProperty("syntax"),"nullpattern") ;
    // 4 inputs
    
    Pattern[] patterns = parsePatterns(file,"left:0 right:1 next:2 prev:3") ;
    assertEquals(4,patterns.length) ;
    
    // 3 inputs
    patterns = parsePatterns(file,"left:0 right:1 next:2") ;
    assertEquals(2,patterns.length) ;
    
    // 2 inputs
    patterns = parsePatterns(file,"left:0 right:1") ;
    assertEquals(1,patterns.length) ;
    
  }
  
  
  /**
   * Test the F marks.
   */
  public void testMarks() throws Exception {
    
    File file = new File(System.getProperty("syntax"),"marks") ;
    
    Pattern[] patterns = parsePatterns(file,"primitive:0 routed:1 leaf:2") ;
    
    assertEquals(2,patterns.length) ;
    
    assertEquals(3,patterns[0].getInputs().length) ;
    assertEquals(3,patterns[1].getInputs().length) ;
    
    // non leaf
    assertEquals("Tleaf",patterns[0].getOutput()) ;
    assertTrue(patterns[0].containsMark(0)) ;
    assertFalse(patterns[0].containsMark(1)) ;
    assertTrue(patterns[0].containsMark(2)) ;
    
    // leaf
    assertEquals("Tnode",patterns[1].getOutput()) ;
    assertTrue(patterns[1].containsMark(0)) ;
    assertTrue(patterns[1].containsMark(1)) ;
    assertFalse(patterns[1].containsMark(2)) ;
    
  }
  
  
  /**
   * Test a input with is routed and primitive.
   */
  public void testPrimitiveAndRouted() throws Exception {
    
    File file = new File(System.getProperty("syntax"),"primitive-and-routed") ;
    
    Pattern[] patterns = parsePatterns(file,"left:0 right:1") ;
    
    assertEquals(2,patterns.length) ;
    
    Input[] inputs0 = patterns[0].getInputs() ;
    Input[] inputs1 = patterns[1].getInputs() ;
    
    assertEquals(2,inputs0.length) ;
    assertEquals(2,inputs1.length) ;
    
    assertEquals("left", inputs0[0].getName()) ;
    assertEquals("right",inputs0[1].getName()) ;
    assertTrue(inputs0[1].isPrimitive()) ;
    
    assertEquals("left", inputs1[0].getName()) ;
    assertEquals("right",inputs1[1].getName()) ;
    assertFalse(inputs1[1].isPrimitive()) ;
  }
  
}
